/*
    pixilang_vm.cpp
    This file is part of the Pixilang programming language.
    
    [ MIT license ]

    Copyright (c) 2006 - 2012, Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to 
    deal in the Software without restriction, including without limitation the 
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is 
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in 
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.
*/

#include "pixilang.h"

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

//#define SHOW_DEBUG_MESSAGES

#ifdef SHOW_DEBUG_MESSAGES
#define DPRINT( fmt, ARGS... ) dprint( fmt, ## ARGS )
#else
#define DPRINT( fmt, ARGS... ) {}
#endif

//#define VM_DEBUGGER

#define FIRST_CODE_PAGE	    1024
#define NEXT_CODE_PAGE	    1024

const UTF8_CHAR* g_pix_container_type_names[] =
{
    "INT8",
    "INT16",
    "INT32",
    "INT64",
    "FLOAT32",
    "FLOAT64",
    "DYNAMIC"
};

const int g_pix_container_type_sizes[] =
{
    1,
    2,
    4,
    8,
    4,
    8,
    sizeof( PIX_VAL )
};

void pix_vm_fill_input_buffer( sound_struct* ss, sound_slot* slot, pix_vm* vm, int chnum, int frames )
{
    int src_offset = 0;
    int src_step = ( slot->frames << 15 ) / frames;
    switch( vm->audio_format )
    {
        default: break;
	case PIX_CONTAINER_TYPE_INT8:
	    {
		signed char* dest_ptr = (signed char*)vm->audio_input_buffers[ chnum ];
		switch( ss->in_type )
		{
		    case sound_buffer_int16:
			{
			    signed short* src_ptr = (signed short*)slot->in_buffer;
			    src_ptr += chnum;
			    for( int i = 0; i < frames; i++ )
			    {
				dest_ptr[ i ] = src_ptr[ ( src_offset >> 15 ) * ss->in_channels ] >> 8;
				src_offset += src_step;
			    }
			}
			break;
		    case sound_buffer_float32:
			{
			    float* src_ptr = (float*)slot->in_buffer;
			    src_ptr += chnum;
			    for( int i = 0; i < frames; i++ )
			    {
				float v = src_ptr[ ( src_offset >> 15 ) * ss->in_channels ] * 128;
				if( v > 127 ) v = 127;
				if( v < -127 ) v = -127;
				dest_ptr[ i ] = v;
				src_offset += src_step;
			    }
			}
			break;
		}
	    }
	    break;
	case PIX_CONTAINER_TYPE_INT16:
	    {
		signed short* dest_ptr = (signed short*)vm->audio_input_buffers[ chnum ];
		switch( ss->in_type )
		{
		    case sound_buffer_int16:
			{
			    signed short* src_ptr = (signed short*)slot->in_buffer;
			    src_ptr += chnum;
			    for( int i = 0; i < frames; i++ )
			    {
				dest_ptr[ i ] = src_ptr[ ( src_offset >> 15 ) * ss->in_channels ];
				src_offset += src_step;
			    }
			}
			break;
		    case sound_buffer_float32:
			{
			    float* src_ptr = (float*)slot->in_buffer;
			    src_ptr += chnum;
			    for( int i = 0; i < frames; i++ )
			    {
				float v = src_ptr[ ( src_offset >> 15 ) * ss->in_channels ] * 32768;
				if( v > 32767 ) v = 32767;
				if( v < -32767 ) v = -32767;
				dest_ptr[ i ] = v;
				src_offset += src_step;
			    }
			}
			break;
		}
	    }
	    break;
	case PIX_CONTAINER_TYPE_INT32:
	    {
		int* dest_ptr = (int*)vm->audio_input_buffers[ chnum ];
		switch( ss->in_type )
		{
		    case sound_buffer_int16:
			{
			    signed short* src_ptr = (signed short*)slot->in_buffer;
			    src_ptr += chnum;
			    for( int i = 0; i < frames; i++ )
			    {
				dest_ptr[ i ] = src_ptr[ ( src_offset >> 15 ) * ss->in_channels ];
				src_offset += src_step;
			    }
			}
			break;
		    case sound_buffer_float32:
			{
			    float* src_ptr = (float*)slot->in_buffer;
			    src_ptr += chnum;
			    for( int i = 0; i < frames; i++ )
			    {
				float v = src_ptr[ ( src_offset >> 15 ) * ss->in_channels ] * 32768;
				if( v > 32767 ) v = 32767;
				if( v < -32767 ) v = -32767;
				dest_ptr[ i ] = v;
				src_offset += src_step;
			    }
			}
			break;
		}
	    }
	    break;
	case PIX_CONTAINER_TYPE_FLOAT32:
	    {
		float* dest_ptr = (float*)vm->audio_input_buffers[ chnum ];
		switch( ss->in_type )
		{
		    case sound_buffer_int16:
			{
			    signed short* src_ptr = (signed short*)slot->in_buffer;
			    src_ptr += chnum;
			    for( int i = 0; i < frames; i++ )
			    {
				dest_ptr[ i ] = (float)src_ptr[ ( src_offset >> 15 ) * ss->in_channels ] / 32768;
				src_offset += src_step;
			    }
			}
			break;
		    case sound_buffer_float32:
			{
			    float* src_ptr = (float*)slot->in_buffer;
			    src_ptr += chnum;
			    for( int i = 0; i < frames; i++ )
			    {
				dest_ptr[ i ] = src_ptr[ ( src_offset >> 15 ) * ss->in_channels ];
				src_offset += src_step;
			    }
			}
			break;
		}
	    }
	    break;
#ifdef PIX_FLOAT64_ENABLED
	case PIX_CONTAINER_TYPE_FLOAT64:
	    {
		double* dest_ptr = (double*)vm->audio_input_buffers[ chnum ];
		switch( ss->in_type )
		{
		    case sound_buffer_int16:
			{
			    signed short* src_ptr = (signed short*)slot->in_buffer;
			    src_ptr += chnum;
			    for( int i = 0; i < frames; i++ )
			    {
				dest_ptr[ i ] = (double)src_ptr[ ( src_offset >> 15 ) * ss->in_channels ] / 32768;
				src_offset += src_step;
			    }
			}
			break;
		    case sound_buffer_float32:
			{
			    float* src_ptr = (float*)slot->in_buffer;
			    src_ptr += chnum;
			    for( int i = 0; i < frames; i++ )
			    {
				dest_ptr[ i ] = src_ptr[ ( src_offset >> 15 ) * ss->in_channels ];
				src_offset += src_step;
			    }
			}
			break;
		}
	    }
	    break;
#endif
    }
}

int pix_vm_render_piece_of_sound( sound_struct* ss, int slot_num )
{
    bool handled = 0;
    
    if( ss )
    {
	sound_slot* slot = &ss->slots[ slot_num ];
	int frames = slot->frames;
	
	pix_vm* vm = (pix_vm*)slot->user_data;
	if( vm->audio_callback < 0 ) return 0;

	uint freq_ratio = ( vm->audio_freq << PIX_AUDIO_PTR_PREC ) / ss->freq;
	
	vm->audio_src_ptr += frames * freq_ratio;
	uint interpolation_frames = 1;
	uint src_frames = ( ( ( vm->audio_src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 + interpolation_frames ) - vm->audio_src_rendered ) & ((1<<(32-PIX_AUDIO_PTR_PREC))-1);
	vm->audio_src_rendered += src_frames;
	
	if( src_frames ) 
	{
	    //Need a new piece of sound:
	    uint src_buffer_offset;
	    if( vm->audio_src_buffers[ 0 ] == 0 )
	    {
		//Buffers are empty. Let create it:
		for( int i = 0; i < vm->audio_channels; i++ )
		{
		    mem_free( vm->audio_src_buffers[ i ] );
		    vm->audio_src_buffers[ i ] = MEM_NEW2( g_pix_container_type_sizes[ vm->audio_format ] * src_frames );
		}
		vm->audio_src_buffer_size = src_frames;
		vm->audio_src_buffer_ptr = 0;
		src_buffer_offset = 0;
	    }
	    else
	    {
		//Buffer already has some data. We need to save last 16 frames from it - for interpolation:
		if( vm->audio_src_buffer_size > 16 )
		{
		    for( int i = 0; i < vm->audio_channels; i++ )
			mem_copy( vm->audio_src_buffers[ i ], (char*)vm->audio_src_buffers[ i ] + ( vm->audio_src_buffer_size - 16 ) * g_pix_container_type_sizes[ vm->audio_format ], 16 * g_pix_container_type_sizes[ vm->audio_format ] );
		    vm->audio_src_buffer_ptr -= ( vm->audio_src_buffer_size - 16 ) << PIX_AUDIO_PTR_PREC;
		    vm->audio_src_buffer_size = 16;
		}
		src_buffer_offset = vm->audio_src_buffer_size;
		vm->audio_src_buffer_size += src_frames;
		size_t new_buffer_bytes = vm->audio_src_buffer_size * g_pix_container_type_sizes[ vm->audio_format ];
		if( new_buffer_bytes > mem_get_size( vm->audio_src_buffers[ 0 ] ) )
		    for( int i = 0; i < vm->audio_channels; i++ )
			vm->audio_src_buffers[ i ] = mem_resize( vm->audio_src_buffers[ i ], new_buffer_bytes );
	    }
	    //Render a new piece of sound:
	    {
		//Output:
		pix_vm_container* c = vm->c[ vm->audio_channels_cont ];
		int* cdata = (int*)c->data;
		for( int i = 0; i < PIX_VM_AUDIO_CHANNELS; i++ )
		{
		    if( i < vm->audio_channels )
			cdata[ i ] = vm->audio_buffers_conts[ i ];
		    else
			cdata[ i ] = -1;
		}
		for( int i = 0; i < vm->audio_channels; i++ )
		{
		    c = vm->c[ vm->audio_buffers_conts[ i ] ];
		    c->type = vm->audio_format;
		    c->data = (void*)( (char*)vm->audio_src_buffers[ i ] + src_buffer_offset * g_pix_container_type_sizes[ vm->audio_format ] );
		    c->xsize = src_frames;
		    c->size = src_frames;
		}
		
		//Input:
		if( vm->audio_input_enabled && slot->in_buffer )
		{
		    c = vm->c[ vm->audio_input_channels_cont ];
		    cdata = (int*)c->data;
		    for( int i = 0; i < PIX_VM_AUDIO_CHANNELS; i++ )
		    {
			if( i < vm->audio_channels )
			    cdata[ i ] = vm->audio_input_buffers_conts[ i ];
			else
			    cdata[ i ] = -1;
		    }
		    for( int i = 0; i < vm->audio_channels; i++ )
		    {
			if( mem_get_size( vm->audio_input_buffers[ i ] ) < src_frames * g_pix_container_type_sizes[ vm->audio_format ] )
			    vm->audio_input_buffers[ i ] = mem_resize( vm->audio_input_buffers[ i ], src_frames * g_pix_container_type_sizes[ vm->audio_format ] );
			c = vm->c[ vm->audio_input_buffers_conts[ i ] ];
			c->type = vm->audio_format;
			if( i < ss->in_channels )
			{
			    if( mem_get_size( vm->audio_input_buffers[ i ] ) < src_frames * g_pix_container_type_sizes[ vm->audio_format ] )
				vm->audio_input_buffers[ i ] = mem_resize( vm->audio_input_buffers[ i ], src_frames * g_pix_container_type_sizes[ vm->audio_format ] );
			    c->data = vm->audio_input_buffers[ i ];
			    pix_vm_fill_input_buffer( ss, slot, vm, i, src_frames );
			}
			else
			{
			    c->data = vm->audio_input_buffers[ 0 ];
			}
			c->xsize = src_frames;
			c->size = src_frames;
		    }
		}
	    }
	    pix_vm_function fun;
	    PIX_VAL pp[ 7 ];
	    char pp_types[ 7 ];
	    fun.p = pp;
	    fun.p_types = pp_types;
	    fun.addr = vm->audio_callback;
	    fun.p[ 0 ].i = 0;
	    fun.p_types[ 0 ] = 0;
	    fun.p[ 1 ] = vm->audio_userdata;
	    fun.p_types[ 1 ] = vm->audio_userdata_type;
	    fun.p[ 2 ].i = vm->audio_channels_cont;
	    fun.p_types[ 2 ] = 0;
	    fun.p[ 3 ].i = src_frames;
	    fun.p_types[ 3 ] = 0;
	    fun.p[ 4 ].i = slot->time;
	    fun.p_types[ 4 ] = 0;
	    if( vm->audio_input_enabled && slot->in_buffer )
		fun.p[ 5 ].i = vm->audio_input_channels_cont;
	    else
		fun.p[ 5 ].i = -1;
	    fun.p_types[ 5 ] = 0;
	    fun.p[ 6 ].i = ( ss->latency * freq_ratio ) >> PIX_AUDIO_PTR_PREC;
	    fun.p_types[ 6 ] = 0;
	    fun.p_num = 7;
	    pix_vm_run( PIX_VM_THREADS - 1, 0, &fun, PIX_VM_CALL_FUNCTION, vm );
	    PIX_VAL retval;
	    char retval_type;
	    pix_vm_get_thread_retval( PIX_VM_THREADS - 1, vm, &retval, &retval_type );
	    if( retval_type == 0 )
		handled = retval.i;
	    else 
		handled = retval.f;
	    {
		for( int i = 0; i < PIX_VM_AUDIO_CHANNELS; i++ )
		{
		    pix_vm_container* c = vm->c[ vm->audio_buffers_conts[ i ] ];
		    c->data = 0;
		    c->xsize = 0;
		    c->size = 0;
		}
	    }
	}
	
	//Final interpolation:
	if( handled )
	{
	    uint src_ptr = vm->audio_src_buffer_ptr;
	    if( ss->mode & SOUND_MODE_INT16 )
	    {
		signed short* dest = (signed short*)slot->buffer;
		switch( vm->audio_format )
		{
		    case PIX_CONTAINER_TYPE_INT8:
			{
			    signed char* src1 = (signed char*)vm->audio_src_buffers[ 0 ];
			    signed char* src2 = (signed char*)vm->audio_src_buffers[ 1 ];
			    if( src2 == 0 ) src2 = src1;
			    if( vm->audio_flags & PIX_AUDIO_FLAG_INTERP2 )
			    {
				for( int i = 0; i < frames; i++ )
				{
				    int c1 = src_ptr & ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 );
				    int c2 = ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 ) - c1;
				    int v1 = ( (int)src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + (int)src1[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1 ) >> PIX_AUDIO_PTR_PREC;
				    int v2 = ( (int)src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + (int)src2[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1 ) >> PIX_AUDIO_PTR_PREC;
				    dest[ i * 2 + 0 ] = (signed short)v1 << 8;
				    dest[ i * 2 + 1 ] = (signed short)v2 << 8;
				    src_ptr += freq_ratio;
				}
			    }
			    else
			    {
				for( int i = 0; i < frames; i++ )
				{
				    dest[ i * 2 + 0 ] = (signed short)src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] << 8;
				    dest[ i * 2 + 1 ] = (signed short)src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] << 8;
				    src_ptr += freq_ratio;
				}
			    }
			    vm->audio_src_buffer_ptr = src_ptr;
			}
			break;
		    case PIX_CONTAINER_TYPE_INT16:
			{
			    signed short* src1 = (signed short*)vm->audio_src_buffers[ 0 ];
			    signed short* src2 = (signed short*)vm->audio_src_buffers[ 1 ];
			    if( src2 == 0 ) src2 = src1;
			    if( vm->audio_flags & PIX_AUDIO_FLAG_INTERP2 )
			    {
				for( int i = 0; i < frames; i++ )
				{
				    int c1 = src_ptr & ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 );
				    int c2 = ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 ) - c1;
				    int v1 = ( (int)src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + (int)src1[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1 ) >> PIX_AUDIO_PTR_PREC;
				    int v2 = ( (int)src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + (int)src2[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1 ) >> PIX_AUDIO_PTR_PREC;
				    dest[ i * 2 + 0 ] = (signed short)v1;
				    dest[ i * 2 + 1 ] = (signed short)v2;
				    src_ptr += freq_ratio;
				}
			    }
			    else
			    {
				for( int i = 0; i < frames; i++ )
				{
				    dest[ i * 2 + 0 ] = src1[ src_ptr >> PIX_AUDIO_PTR_PREC ];
				    dest[ i * 2 + 1 ] = src2[ src_ptr >> PIX_AUDIO_PTR_PREC ];
				    src_ptr += freq_ratio;
				}
			    }
			    vm->audio_src_buffer_ptr = src_ptr;
			}
			break;
		    case PIX_CONTAINER_TYPE_INT32:
			{
			    signed int* src1 = (signed int*)vm->audio_src_buffers[ 0 ];
			    signed int* src2 = (signed int*)vm->audio_src_buffers[ 1 ];
			    if( src2 == 0 ) src2 = src1;
			    if( vm->audio_flags & PIX_AUDIO_FLAG_INTERP2 )
			    {
				for( int i = 0; i < frames; i++ )
				{
				    int c1 = src_ptr & ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 );
				    int c2 = ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 ) - c1;
				    int v_1 = src1[ src_ptr >> PIX_AUDIO_PTR_PREC ];
				    if( v_1 > 32767 ) v_1 = 32767;
				    if( v_1 < -32767 ) v_1 = -32767;
				    int v_2 = src1[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ];
				    if( v_2 > 32767 ) v_2 = 32767;
				    if( v_2 < -32767 ) v_2 = -32767;
				    int v1 = ( v_1 * c2 + v_2 * c1 ) >> PIX_AUDIO_PTR_PREC;
				    v_1 = src2[ src_ptr >> PIX_AUDIO_PTR_PREC ];
				    if( v_1 > 32767 ) v_1 = 32767;
				    if( v_1 < -32767 ) v_1 = -32767;
				    v_2 = src2[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ];
				    if( v_2 > 32767 ) v_2 = 32767;
				    if( v_2 < -32767 ) v_2 = -32767;
				    int v2 = ( v_1 * c2 + v_2 * c1 ) >> PIX_AUDIO_PTR_PREC;
				    dest[ i * 2 + 0 ] = (signed short)v1;
				    dest[ i * 2 + 1 ] = (signed short)v2;
				    src_ptr += freq_ratio;
				}
			    }
			    else
			    {
				for( int i = 0; i < frames; i++ )
				{
				    int v = src1[ src_ptr >> PIX_AUDIO_PTR_PREC ];
				    if( v > 32767 ) v = 32767;
				    if( v < -32767 ) v = -32767;
				    dest[ i * 2 + 0 ] = (signed short)v;
				    v = src2[ src_ptr >> PIX_AUDIO_PTR_PREC ];
				    if( v > 32767 ) v = 32767;
				    if( v < -32767 ) v = -32767;
				    dest[ i * 2 + 1 ] = (signed short)v;
				    src_ptr += freq_ratio;
				}
			    }
			    vm->audio_src_buffer_ptr = src_ptr;
			}
			break;
		    case PIX_CONTAINER_TYPE_FLOAT32:
			{
			    float* src1 = (float*)vm->audio_src_buffers[ 0 ];
			    float* src2 = (float*)vm->audio_src_buffers[ 1 ];
			    if( src2 == 0 ) src2 = src1;
			    if( vm->audio_flags & PIX_AUDIO_FLAG_INTERP2 )
			    {
				for( int i = 0; i < frames; i++ )
				{
				    float c1 = (float)( src_ptr & ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 ) ) / (float)( 1 << PIX_AUDIO_PTR_PREC );
				    float c2 = 1 - c1;
				    float v1 = src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + src1[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1;
				    float v2 = src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + src2[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1;
				    v1 *= 32768.0;
				    v2 *= 32768.0;
				    if( v1 > 32767 ) v1 = 32767;
				    if( v1 < -32767 ) v1 = -32767;
				    if( v2 > 32767 ) v2 = 32767;
				    if( v2 < -32767 ) v2 = -32767;
				    dest[ i * 2 + 0 ] = (signed short)v1;
				    dest[ i * 2 + 1 ] = (signed short)v2;
				    src_ptr += freq_ratio;
				}
			    }
			    else
			    {
				for( int i = 0; i < frames; i++ )
				{
				    int v = (int)( src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] * 32767.0F );
				    if( v > 32767 ) v = 32767;
				    if( v < -32767 ) v = -32767;
				    dest[ i * 2 + 0 ] = (signed short)v;
				    v = (int)( src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] * 32767.0F );
				    if( v > 32767 ) v = 32767;
				    if( v < -32767 ) v = -32767;
				    dest[ i * 2 + 1 ] = (signed short)v;
				    src_ptr += freq_ratio;
				}
			    }
			    vm->audio_src_buffer_ptr = src_ptr;
			}
			break;
#ifdef PIX_FLOAT64_ENABLED
		    case PIX_CONTAINER_TYPE_FLOAT64:
			{
			    double* src1 = (double*)vm->audio_src_buffers[ 0 ];
			    double* src2 = (double*)vm->audio_src_buffers[ 1 ];
			    if( src2 == 0 ) src2 = src1;
			    if( vm->audio_flags & PIX_AUDIO_FLAG_INTERP2 )
			    {
				for( int i = 0; i < frames; i++ )
				{
				    double c1 = (double)( src_ptr & ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 ) ) / (double)( 1 << PIX_AUDIO_PTR_PREC );
				    double c2 = 1 - c1;
				    double v1 = src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + src1[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1;
				    double v2 = src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + src2[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1;
				    v1 *= 32768.0;
				    v2 *= 32768.0;
				    if( v1 > 32767 ) v1 = 32767;
				    if( v1 < -32767 ) v1 = -32767;
				    if( v2 > 32767 ) v2 = 32767;
				    if( v2 < -32767 ) v2 = -32767;
				    dest[ i * 2 + 0 ] = (signed short)v1;
				    dest[ i * 2 + 1 ] = (signed short)v2;
				    src_ptr += freq_ratio;
				}
			    }
			    else
			    {
				for( int i = 0; i < frames; i++ )
				{
				    int v = (int)( src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] * 32767.0 );
				    if( v > 32767 ) v = 32767;
				    if( v < -32767 ) v = -32767;
				    dest[ i * 2 + 0 ] = (signed short)v;
				    v = (int)( src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] * 32767.0 );
				    if( v > 32767 ) v = 32767;
				    if( v < -32767 ) v = -32767;
				    dest[ i * 2 + 1 ] = (signed short)v;
				    src_ptr += freq_ratio;
				}
			    }
			    vm->audio_src_buffer_ptr = src_ptr;
			}
			break;
#endif
		    default:
			break;
		}
	    }
	    if( ss->mode & SOUND_MODE_FLOAT32 )
	    {
		float* dest = (float*)slot->buffer;
		switch( vm->audio_format )
		{
		    case PIX_CONTAINER_TYPE_INT8:
			{
			    signed char* src1 = (signed char*)vm->audio_src_buffers[ 0 ];
			    signed char* src2 = (signed char*)vm->audio_src_buffers[ 1 ];
			    if( src2 == 0 ) src2 = src1;
			    if( vm->audio_flags & PIX_AUDIO_FLAG_INTERP2 )
			    {
				for( int i = 0; i < frames; i++ )
				{
				    int c1 = src_ptr & ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 );
				    int c2 = ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 ) - c1;
				    int v1 = ( (int)src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + (int)src1[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1 ) >> PIX_AUDIO_PTR_PREC;
				    int v2 = ( (int)src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + (int)src2[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1 ) >> PIX_AUDIO_PTR_PREC;
				    dest[ i * 2 + 0 ] = (float)v1 / 128.0F;
				    dest[ i * 2 + 1 ] = (float)v2 / 128.0F;
				    src_ptr += freq_ratio;
				}
			    }
			    else
			    {
				for( int i = 0; i < frames; i++ )
				{
				    dest[ i * 2 + 0 ] = (float)src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] / 128.0F;
				    dest[ i * 2 + 1 ] = (float)src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] / 128.0F;
				    src_ptr += freq_ratio;
				}
			    }
			    vm->audio_src_buffer_ptr = src_ptr;
			}
			break;
		    case PIX_CONTAINER_TYPE_INT16:
			{
			    signed short* src1 = (signed short*)vm->audio_src_buffers[ 0 ];
			    signed short* src2 = (signed short*)vm->audio_src_buffers[ 1 ];
			    if( src2 == 0 ) src2 = src1;
			    if( vm->audio_flags & PIX_AUDIO_FLAG_INTERP2 )
			    {
				for( int i = 0; i < frames; i++ )
				{
				    int c1 = src_ptr & ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 );
				    int c2 = ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 ) - c1;
				    int v1 = ( (int)src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + (int)src1[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1 ) >> PIX_AUDIO_PTR_PREC;
				    int v2 = ( (int)src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + (int)src2[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1 ) >> PIX_AUDIO_PTR_PREC;
				    dest[ i * 2 + 0 ] = (float)v1 / 32768.0F;
				    dest[ i * 2 + 1 ] = (float)v2 / 32768.0F;
				    src_ptr += freq_ratio;
				}
			    }
			    else
			    {
				for( int i = 0; i < frames; i++ )
				{
				    dest[ i * 2 + 0 ] = (float)src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] / 32768.0F;
				    dest[ i * 2 + 1 ] = (float)src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] / 32768.0F;
				    src_ptr += freq_ratio;
				}
			    }
			    vm->audio_src_buffer_ptr = src_ptr;
			}
			break;
		    case PIX_CONTAINER_TYPE_INT32:
			{
			    signed int* src1 = (signed int*)vm->audio_src_buffers[ 0 ];
			    signed int* src2 = (signed int*)vm->audio_src_buffers[ 1 ];
			    if( src2 == 0 ) src2 = src1;
			    if( vm->audio_flags & PIX_AUDIO_FLAG_INTERP2 )
			    {
				for( int i = 0; i < frames; i++ )
				{
				    float c1 = (float)( src_ptr & ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 ) ) / (float)( 1 << PIX_AUDIO_PTR_PREC );
				    float c2 = (float)( ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 ) - c1 ) / (float)( 1 << PIX_AUDIO_PTR_PREC );
				    float v1 = (float)src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + (float)src1[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1;
				    float v2 = (float)src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + (float)src2[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1;
				    dest[ i * 2 + 0 ] = v1 / 32768.0F;
				    dest[ i * 2 + 1 ] = v2 / 32768.0F;
				    src_ptr += freq_ratio;
				}
			    }
			    else
			    {
				for( int i = 0; i < frames; i++ )
				{
				    dest[ i * 2 + 0 ] = (float)src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] / 32768.0F;
				    dest[ i * 2 + 1 ] = (float)src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] / 32768.0F;
				    src_ptr += freq_ratio;
				}
			    }
			    vm->audio_src_buffer_ptr = src_ptr;
			}
			break;
		    case PIX_CONTAINER_TYPE_FLOAT32:
			{
			    float* src1 = (float*)vm->audio_src_buffers[ 0 ];
			    float* src2 = (float*)vm->audio_src_buffers[ 1 ];
			    if( src2 == 0 ) src2 = src1;
			    if( vm->audio_flags & PIX_AUDIO_FLAG_INTERP2 )
			    {
				for( int i = 0; i < frames; i++ )
				{
				    float c1 = (float)( src_ptr & ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 ) ) / (float)( 1 << PIX_AUDIO_PTR_PREC );
				    float c2 = 1 - c1;
				    float v1 = src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + src1[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1;
				    float v2 = src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + src2[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1;
				    dest[ i * 2 + 0 ] = v1;
				    dest[ i * 2 + 1 ] = v2;
				    src_ptr += freq_ratio;
				}
			    }
			    else
			    {
				for( int i = 0; i < frames; i++ )
				{
				    dest[ i * 2 + 0 ] = src1[ src_ptr >> PIX_AUDIO_PTR_PREC ];
				    dest[ i * 2 + 1 ] = src2[ src_ptr >> PIX_AUDIO_PTR_PREC ];
				    src_ptr += freq_ratio;
				}
			    }
			    vm->audio_src_buffer_ptr = src_ptr;
			}
			break;
#ifdef PIX_FLOAT64_ENABLED
		    case PIX_CONTAINER_TYPE_FLOAT64:
			{
			    double* src1 = (double*)vm->audio_src_buffers[ 0 ];
			    double* src2 = (double*)vm->audio_src_buffers[ 1 ];
			    if( src2 == 0 ) src2 = src1;
			    if( vm->audio_flags & PIX_AUDIO_FLAG_INTERP2 )
			    {
				for( int i = 0; i < frames; i++ )
				{
				    double c1 = (double)( src_ptr & ( ( 1 << PIX_AUDIO_PTR_PREC ) - 1 ) ) / (double)( 1 << PIX_AUDIO_PTR_PREC );
				    double c2 = 1 - c1;
				    double v1 = src1[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + src1[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1;
				    double v2 = src2[ src_ptr >> PIX_AUDIO_PTR_PREC ] * c2 + src2[ ( src_ptr >> PIX_AUDIO_PTR_PREC ) + 1 ] * c1;
				    dest[ i * 2 + 0 ] = v1;
				    dest[ i * 2 + 1 ] = v2;
				    src_ptr += freq_ratio;
				}
			    }
			    else
			    {
				for( int i = 0; i < frames; i++ )
				{
				    dest[ i * 2 + 0 ] = src1[ src_ptr >> PIX_AUDIO_PTR_PREC ];
				    dest[ i * 2 + 1 ] = src2[ src_ptr >> PIX_AUDIO_PTR_PREC ];
				    src_ptr += freq_ratio;
				}
			    }
			    vm->audio_src_buffer_ptr = src_ptr;
			}
			break;
#endif
		    default:
			break;
		}
	    }
	}
	else
	{
	    uint src_ptr = vm->audio_src_buffer_ptr;
	    src_ptr += freq_ratio * frames;
	    vm->audio_src_buffer_ptr = src_ptr;
	}
    }
    
    return handled;
}

int pix_vm_init( pix_vm* vm )
{
    int rv = 0;
    
    mem_set( vm, sizeof( pix_vm ), 0 );
    
    //Containers:
    vm->c = (pix_vm_container**)MEM_NEW2( sizeof( pix_vm_container* ) * PIX_VM_CONTAINERS );
    if( vm->c == 0 )
    {
	dprint( "Memory allocation error (containers)\n" );
	rv = 1;
	goto pix_vm_init_end;
    }
    mem_zero( vm->c );
    sundog_mutex_init( &vm->c_mutex, 0 );
    vm->c_counter = 0;
    
    //Events:
    sundog_mutex_init( &vm->events_mutex, 0 );
    
    //Threads:
    sundog_mutex_init( &vm->th_mutex, 0 );

    //Some initial values:
    vm->transp = 255;
    pix_vm_gfx_matrix_reset( vm );
    vm->gl_callback = -1;
    vm->audio_callback = -1;
    vm->quit_action = 1;
    vm->zbuf = -1;
    vm->audio_channels_cont = -1;
    vm->audio_input_channels_cont = -1;
    for( int i = 0; i < PIX_VM_AUDIO_CHANNELS; i++ )
    {
	vm->audio_buffers_conts[ i ] = -1;
	vm->audio_input_buffers_conts[ i ] = -1;
    }
    vm->event = -1;
    vm->current_path = -1;
    vm->user_path = -1;
    vm->temp_path = -1;
    vm->os_name = -1;
    vm->arch_name = -1;

#ifdef OPENGL
    sundog_mutex_init( &vm->gl_mutex, 0 );
#endif
        
    vm->ready = 1;
    
pix_vm_init_end:
    
    return rv;
}

int pix_vm_deinit( pix_vm* vm )
{
    int rv = 0;

    if( vm == 0 ) return 2;
    
    vm->ready = 0;
    
    //Stop separate threads created by Pixilang:
    int sleep_counter = 0;
    int sleep_counter_deadline = 500;
    for( int i = 0; i < PIX_VM_THREADS; i++ )
    {
	sundog_mutex_lock( &vm->th_mutex );
	if( vm->th[ i ] )
	{
	    if( vm->th[ i ]->thread_open )
	    {
		int step_ms = 25;
		bool closed = 0;
		while( sleep_counter < sleep_counter_deadline )
		{
		    int r = sundog_thread_destroy( &vm->th[ i ]->th, -step_ms );
		    sundog_mutex_unlock( &vm->th_mutex );
		    time_sleep( 10 );
		    sundog_mutex_lock( &vm->th_mutex );
		    if( r == 0 )
		    {
			closed = 1;
			break;
		    }
		    sleep_counter += step_ms;
		}
		if( closed == 0 )
		{
		    vm->th[ i ]->active = 0;
		    if( sundog_thread_destroy( &vm->th[ i ]->th, 100 ) )
		    {
			dprint( "Thread %d is not responding. Forcibly removed.\n", i );
		    }
		    else 
		    {
			dprint( "Thread %d is not responding. Forcibly removed (via halt).\n", i );
		    }
		}
	    }
	}
	sundog_mutex_unlock( &vm->th_mutex );
    }
    
    //Close audio stream:
    PIX_VAL u;
    u.i = 0;
    pix_vm_set_audio_callback( -1, u, 0, 0, PIX_CONTAINER_TYPE_INT16, 0, 0, vm );
    
    //Remove all threads data:
    for( int i = 0; i < PIX_VM_THREADS; i++ )
    {
	sundog_mutex_lock( &vm->th_mutex );
	if( vm->th[ i ] )
	{
	    mem_free( vm->th[ i ] );
	    vm->th[ i ] = 0;
	}
	sundog_mutex_unlock( &vm->th_mutex );
    }
    sundog_mutex_destroy( &vm->th_mutex );

    //Events:
    sundog_mutex_destroy( &vm->events_mutex );

    //Variables:
    if( vm->var_names )
    {
	for( size_t i = 0; i < vm->vars_num; i++ )
	{
	    mem_free( vm->var_names[ i ] );
	}
    }
    mem_free( vm->vars );
    mem_free( vm->var_types );
    mem_free( vm->var_names );
    vm->vars = 0;
    vm->var_types = 0;
    vm->var_names = 0;
    vm->vars_num = 0;
    
    //Containers:
    if( vm->c )
    {
	for( int i = 0; i < mem_get_size( vm->c ) / sizeof( pix_vm_container* ); i++ )
	{
	    pix_vm_remove_container( i, vm );
	}
	mem_free( vm->c );
	vm->c = 0;
    }
    sundog_mutex_destroy( &vm->c_mutex );
    
    //Text:
    mem_free( vm->text );
    mem_free( vm->text_lines );
    vm->text = 0;
    vm->text_lines = 0;
    
    //Effector:
    mem_free( vm->effector_colors_r );
    mem_free( vm->effector_colors_g );
    mem_free( vm->effector_colors_b );
    
    //Code:
    mem_free( vm->code );
    vm->code = 0;
    vm->code_ptr = 0;
    vm->code_size = 0;
    
    //Base path:
    mem_free( vm->base_path );
    vm->base_path = 0;

#ifdef OPENGL    
    pix_vm_empty_gl_trash( vm );
    mem_free( vm->gl_unused_textures );
    sundog_mutex_destroy( &vm->gl_mutex );
#endif    

    return rv;
}

int pix_vm_set_audio_callback( PIX_ADDR callback, PIX_VAL userdata, char userdata_type, int freq, pix_container_type format, int channels, int flags, pix_vm* vm )
{
    bool on = 0;
    if( callback == -1 && vm->audio_callback != -1 )
    {
	//Close SunDog sound stream:
	sound_stream_close();
    }
    if( callback != -1 && vm->audio_callback == -1 )
    {
	//Open SunDog sound stream:
	if( sound_stream_init( -1, -1, -1, 0 ) )
	{
	    profile_remove_key( KEY_AUDIODRIVER, 0 );
    	    profile_remove_key( KEY_AUDIODEVICE, 0 );
    	    profile_remove_key( KEY_AUDIODEVICE_IN, 0 );
    	    profile_remove_key( KEY_SOUNDBUFFER, 0 );
    	    profile_remove_key( KEY_FREQ, 0 );
    	    profile_save( 0 );
#ifdef IPHONE
	    debug_show_error_report();
#endif
	    return -1;
	}
	on = 1;
    }
    vm->audio_callback = callback;
    vm->audio_userdata = userdata;
    vm->audio_userdata_type = userdata_type;
    vm->audio_freq = freq;
    vm->audio_format = format;
    vm->audio_channels = channels;
    vm->audio_flags = flags;
    vm->audio_src_ptr = 0;
    vm->audio_src_rendered = 0;
    vm->audio_src_buffer_size = 0;
    vm->audio_src_buffer_ptr = 0;
    for( int i = 0; i < PIX_VM_AUDIO_CHANNELS; i++ )
    {
	if( vm->audio_src_buffers[ i ] )
	{
	    mem_free( vm->audio_src_buffers[ i ] );
	    vm->audio_src_buffers[ i ] = 0;
	}
	if( vm->audio_input_buffers[ i ] )
	{
	    mem_free( vm->audio_input_buffers[ i ] );
	    vm->audio_input_buffers[ i ] = 0;
	}
    }
    if( on )
    {
	if( vm->audio_channels_cont == -1 )
	{
	    //Create audio channels (container):
	    vm->audio_channels_cont = pix_vm_new_container( -1, PIX_VM_AUDIO_CHANNELS, 1, PIX_CONTAINER_TYPE_INT32, 0, 0, vm );
	    vm->audio_input_channels_cont = pix_vm_new_container( -1, PIX_VM_AUDIO_CHANNELS, 1, PIX_CONTAINER_TYPE_INT32, 0, 0, vm );
	    //Create audio buffers (containers):
	    for( int i = 0; i < PIX_VM_AUDIO_CHANNELS; i++ )
	    {
		vm->audio_buffers_conts[ i ] = pix_vm_new_container( -1, 1, 1, PIX_CONTAINER_TYPE_INT32, &vm->audio_channels_cont, 0, vm );
		vm->audio_input_buffers_conts[ i ] = pix_vm_new_container( -1, 1, 1, PIX_CONTAINER_TYPE_INT32, &vm->audio_input_channels_cont, 0, vm );
		pix_vm_set_container_flags( vm->audio_buffers_conts[ i ], pix_vm_get_container_flags( vm->audio_buffers_conts[ i ], vm ) | PIX_CONTAINER_FLAG_STATIC_DATA, vm );
		pix_vm_set_container_flags( vm->audio_input_buffers_conts[ i ], pix_vm_get_container_flags( vm->audio_input_buffers_conts[ i ], vm ) | PIX_CONTAINER_FLAG_STATIC_DATA, vm );
	    }
	}
    	sound_stream_set_slot_callback( 0, (void*)&pix_vm_render_piece_of_sound, vm );
	sound_stream_play( 0 );
    }
    
    return 0;
}

void pix_vm_put_opcode( PIX_OPCODE opcode, pix_vm* vm )
{
    if( vm->code == 0 )
    {
	vm->code = (PIX_OPCODE*)MEM_NEW2( FIRST_CODE_PAGE * sizeof( PIX_OPCODE ) );
	vm->code_size = 1;
    }
    
    if( vm->code_ptr >= vm->code_size )
    {
	vm->code_size = vm->code_ptr + 1;
	if( vm->code_size > mem_get_size( vm->code ) / sizeof( PIX_OPCODE ) )
	{
	    vm->code = (PIX_OPCODE*)mem_resize( vm->code, ( vm->code_size + NEXT_CODE_PAGE ) * sizeof( PIX_OPCODE ) );
	}
    }
    
    vm->code[ vm->code_ptr++ ] = opcode;
}

void pix_vm_put_int( PIX_INT v, pix_vm* vm )
{
    if( sizeof( PIX_OPCODE ) >= sizeof( PIX_INT ) )
    {
	pix_vm_put_opcode( (PIX_OPCODE)v, vm );
    }
    else 
    {
	for( int i = 0; i < sizeof( PIX_INT ); i += sizeof( PIX_OPCODE ) )
	{
	    PIX_OPCODE opcode;
	    int size = sizeof( PIX_INT ) - i;
	    if( size > sizeof( PIX_OPCODE ) ) size = sizeof( PIX_OPCODE );
	    mem_copy( &opcode, (char*)&v + i, size );
	    pix_vm_put_opcode( opcode, vm );
	}
    }
}

void pix_vm_put_float( PIX_FLOAT v, pix_vm* vm )
{
    if( sizeof( PIX_OPCODE ) >= sizeof( PIX_FLOAT ) )
    {
	volatile PIX_OPCODE opcode;
	volatile PIX_FLOAT* ptr = (PIX_FLOAT*)&opcode;
	*ptr = v;
	pix_vm_put_opcode( opcode, vm );
    }
    else 
    {
	for( int i = 0; i < sizeof( PIX_FLOAT ); i += sizeof( PIX_OPCODE ) )
	{
	    PIX_OPCODE opcode;
	    int size = sizeof( PIX_FLOAT ) - i;
	    if( size > sizeof( PIX_OPCODE ) ) size = sizeof( PIX_OPCODE );
	    mem_copy( &opcode, (char*)&v + i, size );
	    pix_vm_put_opcode( opcode, vm );
	}
    }
}

UTF8_CHAR* pix_vm_get_variable_name( pix_vm* vm, size_t vnum )
{
    UTF8_CHAR* name = 0;
    if( vm->var_names && vnum < vm->vars_num )
    {
	name = vm->var_names[ vnum ];
	if( name == 0 )
	{
	    if( vnum < 128 )
	    {
		vm->var_names[ vnum ] = (UTF8_CHAR*)MEM_NEW2( 2 );
		name = vm->var_names[ vnum ];
		name[ 0 ] = (UTF8_CHAR)vnum;
		name[ 1 ] = 0;
	    }
	}
    }
    return name;
}

void pix_vm_resize_variables( pix_vm* vm )
{
    if( vm->vars == 0 )
    {
	vm->vars = (PIX_VAL*)MEM_NEW2( vm->vars_num * sizeof( PIX_VAL ) );
	vm->var_types = (char*)MEM_NEW2( vm->vars_num * sizeof( char ) );
	vm->var_names = (UTF8_CHAR**)MEM_NEW2( vm->vars_num * sizeof( UTF8_CHAR* ) );
	mem_zero( vm->vars );
	mem_zero( vm->var_types );
	mem_zero( vm->var_names );
    }
    else
    {
	size_t prev_size = mem_get_size( vm->vars ) / sizeof( PIX_VAL );
	if( vm->vars_num > prev_size ) 
	{
	    size_t new_size = vm->vars_num + 64;
	    vm->vars = (PIX_VAL*)mem_resize( vm->vars, new_size * sizeof( PIX_VAL ) );
	    vm->var_types = (char*)mem_resize( vm->var_types, new_size * sizeof( char ) );
	    vm->var_names = (UTF8_CHAR**)mem_resize( vm->var_names, new_size * sizeof( UTF8_CHAR* ) );
	    mem_set( vm->vars + prev_size, ( new_size - prev_size ) * sizeof( PIX_VAL ), 0 );
	    mem_set( vm->var_types + prev_size, ( new_size - prev_size ) * sizeof( char ), 0 );
	    mem_set( vm->var_names + prev_size, ( new_size - prev_size ) * sizeof( UTF8_CHAR* ), 0 );
	}
    }
}

int pix_vm_send_event(
    int16 type,
    int16 flags,
    int16 x,
    int16 y,
    int16 key,
    int16 scancode,
    int16 pressure,
    UTF32_CHAR unicode,
    pix_vm* vm )
{
    int rv = 0;
    
    if( vm == 0 ) return 2;
    
    sundog_mutex_lock( &vm->events_mutex );
    
    if( vm->events_count + 1 <= PIX_VM_EVENTS )
    {
	//Get pointer to a new event:
	int new_ptr = ( vm->current_event_num + vm->events_count ) & ( PIX_VM_EVENTS - 1 );
	
	//Save new event to FIFO buffer:
	vm->events[ new_ptr ].type = (uint16)type;
	vm->events[ new_ptr ].time = time_ticks();
	vm->events[ new_ptr ].flags = (uint16)flags;
	vm->events[ new_ptr ].x = (int16)x;
	vm->events[ new_ptr ].y = (int16)y;
	vm->events[ new_ptr ].key = (uint16)key;
	vm->events[ new_ptr ].scancode = (uint16)scancode;
	vm->events[ new_ptr ].pressure = (uint16)pressure;
	vm->events[ new_ptr ].unicode = unicode;
	
	//Increment number of unhandled events:
	vm->events_count = vm->events_count + 1;
    }
    else
    {
	rv = 1;
    }
    
    sundog_mutex_unlock( &vm->events_mutex );
    
    return rv;
}

int pix_vm_get_event( pix_vm* vm )
{
    if( vm->events_count )
    {
	sundog_mutex_lock( &vm->events_mutex );
	
	//There are unhandled events:
	//Copy current event (prepare it for handling):
	pix_vm_event* evt = &vm->events[ vm->current_event_num ];
	int current_event = vm->event;
	if( (unsigned)current_event < (unsigned)PIX_VM_CONTAINERS )
	{
	    pix_vm_container* c = vm->c[ current_event ];
	    if( c && c->data && c->type == PIX_CONTAINER_TYPE_INT32 && c->size >= 16 )
	    {
		int* fields = (int*)c->data;
		fields[ 0 ] = evt->type;
		fields[ 1 ] = evt->flags;
		fields[ 2 ] = evt->time;
		fields[ 3 ] = evt->x;
		fields[ 4 ] = evt->y;
		fields[ 5 ] = evt->key;
		fields[ 6 ] = evt->scancode;
		fields[ 7 ] = evt->pressure;
		fields[ 8 ] = evt->unicode;
	    }
	}
	//This event will be handled. So decrement count of events:
	vm->events_count--;
	//And increment FIFO pointer:
	vm->current_event_num = ( vm->current_event_num + 1 ) & ( PIX_VM_EVENTS - 1 );
	
	sundog_mutex_unlock( &vm->events_mutex );
	
	return 1;
    }
    else 
    {
	return 0;
    }
}

#define CONTROL_PC { }
#define LOAD_OPCODE( v ) { v = code[ pc++ ]; CONTROL_PC; }
#define LOAD_INT( v ) \
{ \
    if( sizeof( PIX_OPCODE ) >= sizeof( PIX_INT ) ) \
    { \
	v = (PIX_INT)code[ pc++ ]; CONTROL_PC; \
    } \
    if( sizeof( PIX_INT ) / sizeof( PIX_OPCODE ) == 2 ) \
    { \
	PIX_OPCODE* v_ptr = (PIX_OPCODE*)&v; \
	v_ptr[ 0 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 1 ] = code[ pc++ ]; CONTROL_PC; \
    } \
    if( sizeof( PIX_INT ) / sizeof( PIX_OPCODE ) == 4 ) \
    { \
	PIX_OPCODE* v_ptr = (PIX_OPCODE*)&v; \
	v_ptr[ 0 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 1 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 2 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 3 ] = code[ pc++ ]; CONTROL_PC; \
    } \
    if( sizeof( PIX_INT ) / sizeof( PIX_OPCODE ) == 8 ) \
    { \
	PIX_OPCODE* v_ptr = (PIX_OPCODE*)&v; \
	v_ptr[ 0 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 1 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 2 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 3 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 4 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 5 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 6 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 7 ] = code[ pc++ ]; CONTROL_PC; \
    } \
}
#define LOAD_FLOAT( v ) \
{ \
    if( sizeof( PIX_OPCODE ) >= sizeof( PIX_INT ) ) \
    { \
	v = *( (PIX_FLOAT*)&code[ pc++ ] ); CONTROL_PC; \
    } \
    if( sizeof( PIX_INT ) / sizeof( PIX_OPCODE ) == 2 ) \
    { \
	PIX_OPCODE* v_ptr = (PIX_OPCODE*)&v; \
	v_ptr[ 0 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 1 ] = code[ pc++ ]; CONTROL_PC; \
    } \
    if( sizeof( PIX_INT ) / sizeof( PIX_OPCODE ) == 4 ) \
    { \
	PIX_OPCODE* v_ptr = (PIX_OPCODE*)&v; \
	v_ptr[ 0 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 1 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 2 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 3 ] = code[ pc++ ]; CONTROL_PC; \
    } \
    if( sizeof( PIX_INT ) / sizeof( PIX_OPCODE ) == 8 ) \
    { \
	PIX_OPCODE* v_ptr = (PIX_OPCODE*)&v; \
	v_ptr[ 0 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 1 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 2 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 3 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 4 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 5 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 6 ] = code[ pc++ ]; CONTROL_PC; \
	v_ptr[ 7 ] = code[ pc++ ]; CONTROL_PC; \
    } \
}

void* pix_vm_thread_body( void* user_data )
{
    pix_vm_thread* th = (pix_vm_thread*)user_data;
    
    DPRINT( "Thread %d begin.\n", th->thread_num );
    
    pix_vm_run( th->thread_num, 0, 0, PIX_VM_CONTINUE, th->vm );

    DPRINT( "Thread %d end.\n", th->thread_num );
    
    return 0;
}

int pix_vm_create_active_thread( int thread_num, pix_vm* vm )
{
    sundog_mutex_lock( &vm->th_mutex );
    if( (unsigned)thread_num >= PIX_VM_THREADS )
    {
	for( int i = 0; i < PIX_VM_THREADS - PIX_VM_SYSTEM_THREADS; i++ )
	{
	    if( vm->th[ i ] == 0 )
	    {
		thread_num = i;
		break;
	    }
	}
	if( (unsigned)thread_num >= PIX_VM_THREADS )
	{
	    //No thread
	    sundog_mutex_unlock( &vm->th_mutex );
	    return -1;
	}
    }
    if( vm->th[ thread_num ] == 0 )
    {
	vm->th[ thread_num ] = (pix_vm_thread*)MEM_NEW2( sizeof( pix_vm_thread ) );
	mem_zero( vm->th[ thread_num ] );
    }
    vm->th[ thread_num ]->active = 1;
    sundog_mutex_unlock( &vm->th_mutex );
    return thread_num;
}

int pix_vm_get_thread_retval( int thread_num, pix_vm* vm, PIX_VAL* retval, char* retval_type )
{
    if( (unsigned)thread_num < PIX_VM_THREADS )
    {
	pix_vm_thread* th = vm->th[ thread_num ];
	if( th->sp < PIX_VM_STACK_SIZE )
        {
            *retval = th->stack[ th->sp ];
            *retval_type = th->stack_types[ th->sp ];
            return 0;
        }
        else
        {
    	    retval[ 0 ].i = 0;
    	    *retval_type = 0;
    	    return -1;
        }
    }
}

int pix_vm_run( 
    int thread_num, 
    bool open_new_thread, 
    pix_vm_function* fun, 
    pix_vm_run_mode mode, 
    pix_vm* vm )
{
    pix_vm_thread* th;
    
    thread_num = pix_vm_create_active_thread( thread_num, vm );
    if( thread_num < 0 ) return thread_num;
    th = vm->th[ thread_num ];
    
    if( mode == PIX_VM_CONTINUE )
    {
    }
    else
    {
	//Reset thread:
	th->sp = PIX_VM_STACK_SIZE;
	//Push the parameters:
	if( fun && fun->p_num )
	{
	    for( int pnum = fun->p_num - 1; pnum >= 0; pnum-- )
	    {
		th->sp--;
		th->stack[ th->sp ] = fun->p[ pnum ];
		th->stack_types[ th->sp ] = fun->p_types[ pnum ];
	    }
	    th->sp--;
	    th->fp = th->sp;
	    th->stack[ th->sp ].i = fun->p_num;
	    th->stack_types[ th->sp ] = 0;
	}
	else 
	{
	    //No parameters:
	    th->sp--;
	    th->fp = th->sp;
	    th->stack[ th->sp ].i = 0;
	    th->stack_types[ th->sp ] = 0;
	}
	//Push previous FP:
	th->sp--;
	th->stack[ th->sp ].i = PIX_VM_STACK_SIZE - 1;
	//Push previous PC:
	th->sp--;
	th->stack[ th->sp ].i = (PIX_INT)vm->halt_addr;
	//Set function pointer:
	switch( mode )
	{
	    case PIX_VM_CALL_FUNCTION: th->pc = fun->addr; break;
	    case PIX_VM_CALL_MAIN: th->pc = 1; break;
	}
    }
    
    if( open_new_thread )
    {
	//Open new thread for code execution:
	th->thread_num = thread_num;
	th->vm = vm;
	th->thread_open = 1;
	sundog_thread_create( &th->th, &pix_vm_thread_body, (void*)th, 0 );
	return thread_num;
    }
    else
    {
	//Execute code in the current thread (blocking mode):
	size_t pc = th->pc;
	size_t sp = th->sp;
	size_t fp = th->fp;
	PIX_OPCODE* code = vm->code;
	size_t code_size = vm->code_size;
	PIX_VAL* vars = vm->vars;
	char* var_types = vm->var_types;
	PIX_VAL* stack = th->stack;
	char* stack_types = th->stack_types;
	PIX_OPCODE val, val2;
	PIX_INT val_i;
	PIX_FLOAT val_f;
	size_t sp2;
	if( pc >= code_size )
	{
	    dprint( "Error. Incorrect PC (program counter) value %u\n", (unsigned int)pc );
	}
	else
	while( th->active )
	{
#ifdef VM_DEBUGGER
	    dprint( "PC:%d; SP:%d; FP:%d. q - exit.\n", (int)pc, (int)sp, (int)fp );
	    int key = getchar();
	    if( key == 'q' ) { th->active = 0; break; }
#endif
	    PIX_OPCODE c = code[ pc ];
	    pc++;
	    switch( (pix_vm_opcode)( c & PIX_OPCODE_MASK ) )
	    {
		case OPCODE_NOP:
		    break;
		
		case OPCODE_HALT:
		    th->active = 0;
		    break;
		
		case OPCODE_PUSH_I:
		    LOAD_INT( val_i );
		    sp--;
		    stack_types[ sp ] = 0;
		    stack[ sp ].i = val_i;
		    break;
		case OPCODE_PUSH_i:
		    sp--;
		    stack_types[ sp ] = 0;
		    stack[ sp ].i = (signed)c >> PIX_OPCODE_BITS;
		    break;
                case OPCODE_PUSH_F:
		    LOAD_FLOAT( val_f );
		    sp--;
		    stack_types[ sp ] = 1;
		    stack[ sp ].f = val_f;
		    break;
		case OPCODE_PUSH_v:
		    val = c >> PIX_OPCODE_BITS;
		    sp--;
		    stack_types[ sp ] = var_types[ val ];
		    stack[ sp ] = vars[ val ];
		    break;
		    
		case OPCODE_GO:
		    {
			size_t addr;
			if( stack_types[ sp ] == 0 )
			    addr = (size_t)stack[ sp ].i;
			else
			    addr = (size_t)stack[ sp ].f;
			if( IS_ADDRESS_CORRECT( addr ) )
			    pc = addr & PIX_INT_ADDRESS_MASK;
			else
			    dprint( "Pixilang VM Error. %u: OPCODE_GO. Address %u is incorrect\n", (unsigned int)pc, (unsigned int)stack[ sp ].i );
		    }
		    sp++;
		    break;
		case OPCODE_JMP_i:
		    pc += ( (signed)c >> PIX_OPCODE_BITS ) - 1;
		    break;
		case OPCODE_JMP_I:
		    LOAD_INT( val_i );
		    pc += val_i - ( 1 + sizeof( PIX_INT ) / sizeof( PIX_OPCODE ) );
		    break;
		case OPCODE_JMP_IF_FALSE_i:
		    if( ( stack_types[ sp ] == 0 && stack[ sp ].i == 0 ) ||
		        ( stack_types[ sp ] == 1 && stack[ sp ].f == 0 ) )
		    {
			pc += ( (signed)c >> PIX_OPCODE_BITS ) - 1;
		    }
		    sp++;
		    break;
		case OPCODE_JMP_IF_FALSE_I:
		    LOAD_INT( val_i );
		    if( ( stack_types[ sp ] == 0 && stack[ sp ].i == 0 ) ||
		        ( stack_types[ sp ] == 1 && stack[ sp ].f == 0 ) )
		    {
			pc += val_i - ( 1 + sizeof( PIX_INT ) / sizeof( PIX_OPCODE ) );
		    }
		    sp++;
		    break;
		    
		case OPCODE_SAVE_TO_VAR_v:
		    val = c >> PIX_OPCODE_BITS;
		    var_types[ val ] = stack_types[ sp ];
		    vars[ val ] = stack[ sp ];
		    sp++;
		    break;

		case OPCODE_SAVE_TO_PROP_I:
		    {
			LOAD_INT( val_i );
			PIX_CID cnum;
			int prop_hash;
			UTF8_CHAR* prop_name;
	                cnum = (PIX_CID)stack[ sp + 1 ].i;
	            	prop_name = pix_vm_get_variable_name( vm, val_i );
	            	if( prop_name )
	            	{
	            	    prop_hash = (int)vm->vars[ val_i ].i;
	            	    pix_vm_set_container_property( cnum, prop_name + 1, prop_hash, stack_types[ sp ], stack[ sp ], vm );
	            	}
	                sp += 2;
		    }
		    break;
		case OPCODE_LOAD_FROM_PROP_I:
		    {
			LOAD_INT( val_i );
			PIX_CID cnum;
			size_t prop_var;
			int prop_hash;
			UTF8_CHAR* prop_name;
	                cnum = (PIX_CID)stack[ sp ].i;
	            	prop_name = pix_vm_get_variable_name( vm, val_i );
	            	bool loaded = 0;
	            	if( prop_name )
	            	{
	            	    prop_hash = (int)vm->vars[ val_i ].i;
	            	    pix_sym* sym = pix_vm_get_container_property( cnum, prop_name + 1, prop_hash, vm );
	            	    if( sym )
	            	    {
	            		stack[ sp ] = sym->val;
	            		if( sym->type == SYMTYPE_NUM_I )
	            		    stack_types[ sp ] = 0;
	            		else
	            		    stack_types[ sp ] = 1;
	            		loaded = 1;
	            	    }
	            	}
	            	if( loaded == 0 )
	            	{
	            	    stack[ sp ].i = 0;
	            	    stack_types[ sp ] = 0;
			}
		    }
		    break;
		    
		case OPCODE_SAVE_TO_MEM:
		    {
			PIX_CID cnum;
	                cnum = (PIX_CID)stack[ sp + 2 ].i;
			if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS && vm->c[ cnum ] )
			{
			    pix_vm_container* cont = vm->c[ cnum ];
	                    PIX_INT offset;
			    if( stack_types[ sp + 1 ] == 0 )
				offset = stack[ sp + 1 ].i;
			    else
				offset = (PIX_INT)stack[ sp + 1 ].f;
			    if( (unsigned)offset < cont->size )
			    {
				if( stack_types[ sp ] == 0 )
				{
				    //Integer value:
				    switch( cont->type )
				    {
					case PIX_CONTAINER_TYPE_INT8: ((signed char*)cont->data)[ offset ] = (signed char)stack[ sp ].i; break;
					case PIX_CONTAINER_TYPE_INT16: ((signed short*)cont->data)[ offset ] = (signed short)stack[ sp ].i; break;
					case PIX_CONTAINER_TYPE_INT32: ((signed int*)cont->data)[ offset ] = (signed int)stack[ sp ].i; break;
#ifdef PIX_INT64_ENABLED
					case PIX_CONTAINER_TYPE_INT64: ((long long*)cont->data)[ offset ] = (long long)stack[ sp ].i; break;
#endif
					case PIX_CONTAINER_TYPE_FLOAT32: ((float*)cont->data)[ offset ] = (float)stack[ sp ].i; break;
#ifdef PIX_FLOAT64_ENABLED
					case PIX_CONTAINER_TYPE_FLOAT64: ((double*)cont->data)[ offset ] = (double)stack[ sp ].i; break;
#endif
					case PIX_CONTAINER_TYPE_DYNAMIC: 
					    cont->data_types[ offset ] = stack_types[ sp ];
					    ((PIX_VAL*)cont->data)[ offset ] = stack[ sp ];
					    break;
				    }
				}
				else 
				{
				    //Floating point value:
				    switch( cont->type )
				    {
					case PIX_CONTAINER_TYPE_INT8: ((signed char*)cont->data)[ offset ] = (signed char)stack[ sp ].f; break;
					case PIX_CONTAINER_TYPE_INT16: ((signed short*)cont->data)[ offset ] = (signed short)stack[ sp ].f; break;
					case PIX_CONTAINER_TYPE_INT32: ((signed int*)cont->data)[ offset ] = (signed int)stack[ sp ].f; break;
#ifdef PIX_INT64_ENABLED
					case PIX_CONTAINER_TYPE_INT64: ((long long*)cont->data)[ offset ] = (long long)stack[ sp ].f; break;
#endif
					case PIX_CONTAINER_TYPE_FLOAT32: ((float*)cont->data)[ offset ] = (float)stack[ sp ].f; break;
#ifdef PIX_FLOAT64_ENABLED
					case PIX_CONTAINER_TYPE_FLOAT64: ((double*)cont->data)[ offset ] = (double)stack[ sp ].f; break;
#endif
					case PIX_CONTAINER_TYPE_DYNAMIC: 
					    cont->data_types[ offset ] = stack_types[ sp ];
					    ((PIX_VAL*)cont->data)[ offset ] = stack[ sp ];
					    break;
				    }
				}
			    }
			}
		    }
		    sp += 3;
		    break;
		case OPCODE_SAVE_TO_MEM_2D:
		    {
			PIX_CID cnum;
	                cnum = (PIX_CID)stack[ sp + 3 ].i;
			if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS && vm->c[ cnum ] )
			{
			    pix_vm_container* cont = vm->c[ cnum ];
	                    PIX_INT offset;
			    if( stack_types[ sp + 2 ] == 0 )
				offset = stack[ sp + 2 ].i;
			    else
				offset = (PIX_INT)stack[ sp + 2 ].f;
			    if( stack_types[ sp + 1 ] == 0 )
				offset += stack[ sp + 1 ].i * cont->xsize;
			    else
				offset += (PIX_INT)stack[ sp + 1 ].f * cont->xsize;
			    if( (unsigned)offset < cont->size )
			    {
				if( stack_types[ sp ] == 0 )
				{
				    //Integer value:
				    switch( cont->type )
				    {
					case PIX_CONTAINER_TYPE_INT8: ((signed char*)cont->data)[ offset ] = (signed char)stack[ sp ].i; break;
					case PIX_CONTAINER_TYPE_INT16: ((signed short*)cont->data)[ offset ] = (signed short)stack[ sp ].i; break;
					case PIX_CONTAINER_TYPE_INT32: ((signed int*)cont->data)[ offset ] = (signed int)stack[ sp ].i; break;
#ifdef PIX_INT64_ENABLED
					case PIX_CONTAINER_TYPE_INT64: ((long long*)cont->data)[ offset ] = (long long)stack[ sp ].i; break;
#endif
					case PIX_CONTAINER_TYPE_FLOAT32: ((float*)cont->data)[ offset ] = (float)stack[ sp ].i; break;
#ifdef PIX_FLOAT64_ENABLED
					case PIX_CONTAINER_TYPE_FLOAT64: ((double*)cont->data)[ offset ] = (double)stack[ sp ].i; break;
#endif
					case PIX_CONTAINER_TYPE_DYNAMIC: 
					    cont->data_types[ offset ] = stack_types[ sp ];
					    ((PIX_VAL*)cont->data)[ offset ] = stack[ sp ];
					    break;
				    }
				}
				else 
				{
				    //Floating point value:
				    switch( cont->type )
				    {
					case PIX_CONTAINER_TYPE_INT8: ((signed char*)cont->data)[ offset ] = (signed char)stack[ sp ].f; break;
					case PIX_CONTAINER_TYPE_INT16: ((signed short*)cont->data)[ offset ] = (signed short)stack[ sp ].f; break;
					case PIX_CONTAINER_TYPE_INT32: ((signed int*)cont->data)[ offset ] = (signed int)stack[ sp ].f; break;
#ifdef PIX_INT64_ENABLED
					case PIX_CONTAINER_TYPE_INT64: ((long long*)cont->data)[ offset ] = (long long)stack[ sp ].f; break;
#endif
					case PIX_CONTAINER_TYPE_FLOAT32: ((float*)cont->data)[ offset ] = (float)stack[ sp ].f; break;
#ifdef PIX_FLOAT64_ENABLED
					case PIX_CONTAINER_TYPE_FLOAT64: ((double*)cont->data)[ offset ] = (double)stack[ sp ].f; break;
#endif
					case PIX_CONTAINER_TYPE_DYNAMIC: 
					    cont->data_types[ offset ] = stack_types[ sp ];
					    ((PIX_VAL*)cont->data)[ offset ] = stack[ sp ];
					    break;
				    }
				}
			    }
			}
		    }
		    sp += 4;
		    break;
		case OPCODE_LOAD_FROM_MEM:
		    {
			sp++;
			PIX_CID cnum;
			cnum = (PIX_CID)stack[ sp ].i;
			if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS && vm->c[ cnum ] )
			{
			    pix_vm_container* cont = vm->c[ cnum ];
			    PIX_INT offset;
			    if( stack_types[ sp - 1 ] == 0 )
				offset = stack[ sp - 1 ].i;
			    else
				offset = (PIX_INT)stack[ sp - 1 ].f;
			    if( (unsigned)offset < cont->size )
			    {
				switch( cont->type )
				{
				    case PIX_CONTAINER_TYPE_INT8: stack_types[ sp ] = 0; stack[ sp ].i = (PIX_INT)( ((signed char*)cont->data)[ offset ] ); break;
				    case PIX_CONTAINER_TYPE_INT16: stack_types[ sp ] = 0; stack[ sp ].i = (PIX_INT)( ((signed short*)cont->data)[ offset ] ); break;
				    case PIX_CONTAINER_TYPE_INT32: stack_types[ sp ] = 0; stack[ sp ].i = (PIX_INT)( ((signed int*)cont->data)[ offset ] ); break;
#ifdef PIX_INT64_ENABLED
				    case PIX_CONTAINER_TYPE_INT64: stack_types[ sp ] = 0; stack[ sp ].i = (PIX_INT)( ((long long*)cont->data)[ offset ] ); break;
#endif
				    case PIX_CONTAINER_TYPE_FLOAT32: stack_types[ sp ] = 1; stack[ sp ].f = (PIX_FLOAT)( ((float*)cont->data)[ offset ] ); break;
#ifdef PIX_FLOAT64_ENABLED
				    case PIX_CONTAINER_TYPE_FLOAT64: stack_types[ sp ] = 1; stack[ sp ].f = (PIX_FLOAT)( ((double*)cont->data)[ offset ] ); break;
#endif
				    case PIX_CONTAINER_TYPE_DYNAMIC: 
					stack_types[ sp ] = cont->data_types[ offset ]; 
					stack[ sp ] = ((PIX_VAL*)cont->data)[ offset ];
					break;
				}
			    }
			    else
			    {
				stack_types[ sp ] = 0;
				stack[ sp ].i = 0;
			    }
			}
			else
			{
			    stack_types[ sp ] = 0;
			    stack[ sp ].i = 0;
			}
	   	    }
		    break;
		case OPCODE_LOAD_FROM_MEM_2D:
		    {
			sp += 2;
			PIX_CID cnum;
			cnum = (PIX_CID)stack[ sp ].i;
			if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS && vm->c[ cnum ] )
			{
			    pix_vm_container* cont = vm->c[ cnum ];
			    PIX_INT offset;
			    if( stack_types[ sp - 1 ] == 0 )
				offset = stack[ sp - 1 ].i;
			    else
				offset = (PIX_INT)stack[ sp - 1 ].f;
			    if( stack_types[ sp - 2 ] == 0 )
				offset += stack[ sp - 2 ].i * cont->xsize;
			    else
				offset += (PIX_INT)stack[ sp - 2 ].f * cont->xsize;
			    if( (unsigned)offset < cont->size )
			    {
				switch( cont->type )
				{
				    case PIX_CONTAINER_TYPE_INT8: stack_types[ sp ] = 0; stack[ sp ].i = (PIX_INT)( ((signed char*)cont->data)[ offset ] ); break;
				    case PIX_CONTAINER_TYPE_INT16: stack_types[ sp ] = 0; stack[ sp ].i = (PIX_INT)( ((signed short*)cont->data)[ offset ] ); break;
				    case PIX_CONTAINER_TYPE_INT32: stack_types[ sp ] = 0; stack[ sp ].i = (PIX_INT)( ((signed int*)cont->data)[ offset ] ); break;
#ifdef PIX_INT64_ENABLED
				    case PIX_CONTAINER_TYPE_INT64: stack_types[ sp ] = 0; stack[ sp ].i = (PIX_INT)( ((long long*)cont->data)[ offset ] ); break;
#endif
				    case PIX_CONTAINER_TYPE_FLOAT32: stack_types[ sp ] = 1; stack[ sp ].f = (PIX_FLOAT)( ((float*)cont->data)[ offset ] ); break;
#ifdef PIX_FLOAT64_ENABLED
				    case PIX_CONTAINER_TYPE_FLOAT64: stack_types[ sp ] = 1; stack[ sp ].f = (PIX_FLOAT)( ((double*)cont->data)[ offset ] ); break;
#endif
				    case PIX_CONTAINER_TYPE_DYNAMIC: 
					stack_types[ sp ] = cont->data_types[ offset ]; 
					stack[ sp ] = ((PIX_VAL*)cont->data)[ offset ];
					break;
				}
			    }
			    else
			    {
				stack_types[ sp ] = 0;
				stack[ sp ].i = 0;
			    }
			}
			else
			{
			    stack_types[ sp ] = 0;
			    stack[ sp ].i = 0;
			}
	   	    }
		    break;
		
		case OPCODE_SAVE_TO_STACKFRAME_i:
		    val = (signed)c >> PIX_OPCODE_BITS;
		    stack_types[ fp + (signed)val ] = stack_types[ sp ];
		    stack[ fp + (signed)val ] = stack[ sp ];
		    sp++;
		    break;
		case OPCODE_LOAD_FROM_STACKFRAME_i:
		    val = (signed)c >> PIX_OPCODE_BITS;
		    sp--;
		    stack_types[ sp ] = stack_types[ fp + (signed)val ];
		    stack[ sp ] = stack[ fp + (signed)val ];
		    break;
		
		case OPCODE_SUB:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i -= stack[ sp ].i; break;
			case 1: /*IF*/ stack_types[ sp2 ] = 1; stack[ sp2 ].f = (PIX_FLOAT)stack[ sp2 ].i - stack[ sp ].f; break;
			case 2: /*FI*/ stack[ sp2 ].f = stack[ sp2 ].f - (PIX_FLOAT)stack[ sp ].i; break;
			case 3: /*FF*/ stack[ sp2 ].f -= stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_ADD:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i += stack[ sp ].i; break;
			case 1: /*IF*/ stack_types[ sp2 ] = 1; stack[ sp2 ].f = (PIX_FLOAT)stack[ sp2 ].i + stack[ sp ].f; break;
			case 2: /*FI*/ stack[ sp2 ].f = stack[ sp2 ].f + (PIX_FLOAT)stack[ sp ].i; break;
			case 3: /*FF*/ stack[ sp2 ].f += stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_MUL:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i *= stack[ sp ].i; break;
			case 1: /*IF*/ stack_types[ sp2 ] = 1; stack[ sp2 ].f = (PIX_FLOAT)stack[ sp2 ].i * stack[ sp ].f; break;
			case 2: /*FI*/ stack[ sp2 ].f = stack[ sp2 ].f * (PIX_FLOAT)stack[ sp ].i; break;
			case 3: /*FF*/ stack[ sp2 ].f *= stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_IDIV:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i /= stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = stack[ sp2 ].i / (PIX_INT)stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = (PIX_INT)stack[ sp2 ].f / stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = (PIX_INT)stack[ sp2 ].f / (PIX_INT)stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_DIV:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack_types[ sp2 ] = 1; stack[ sp2 ].f = (PIX_FLOAT)stack[ sp2 ].i / (PIX_FLOAT)stack[ sp ].i; break;
			case 1: /*IF*/ stack_types[ sp2 ] = 1; stack[ sp2 ].f = (PIX_FLOAT)stack[ sp2 ].i / stack[ sp ].f; break;
			case 2: /*FI*/ stack[ sp2 ].f = stack[ sp2 ].f / (PIX_FLOAT)stack[ sp ].i; break;
			case 3: /*FF*/ stack[ sp2 ].f /= stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_MOD:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i %= stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = stack[ sp2 ].i % (PIX_INT)stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = (PIX_INT)stack[ sp2 ].f % stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = (PIX_INT)stack[ sp2 ].f % (PIX_INT)stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_AND:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i &= stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = stack[ sp2 ].i & (PIX_INT)stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = (PIX_INT)stack[ sp2 ].f & stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = (PIX_INT)stack[ sp2 ].f & (PIX_INT)stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_OR:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i |= stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = stack[ sp2 ].i | (PIX_INT)stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = (PIX_INT)stack[ sp2 ].f | stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = (PIX_INT)stack[ sp2 ].f | (PIX_INT)stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_XOR:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i ^= stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = stack[ sp2 ].i ^ (PIX_INT)stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = (PIX_INT)stack[ sp2 ].f ^ stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = (PIX_INT)stack[ sp2 ].f ^ (PIX_INT)stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_ANDAND:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i = stack[ sp2 ].i && stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = stack[ sp2 ].i && stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f && stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f && stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_OROR:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i = stack[ sp2 ].i || stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = stack[ sp2 ].i || stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f || stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f || stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_EQ:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i = stack[ sp2 ].i == stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = (PIX_FLOAT)stack[ sp2 ].i == stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f == (PIX_FLOAT)stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f == stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_NEQ:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i = stack[ sp2 ].i != stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = (PIX_FLOAT)stack[ sp2 ].i != stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f != (PIX_FLOAT)stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f != stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_LESS:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i = stack[ sp2 ].i < stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = (PIX_FLOAT)stack[ sp2 ].i < stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f < (PIX_FLOAT)stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f < stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_LEQ:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i = stack[ sp2 ].i <= stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = (PIX_FLOAT)stack[ sp2 ].i <= stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f <= (PIX_FLOAT)stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f <= stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_GREATER:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i = stack[ sp2 ].i > stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = (PIX_FLOAT)stack[ sp2 ].i > stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f > (PIX_FLOAT)stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f > stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_GEQ:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i = stack[ sp2 ].i >= stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = (PIX_FLOAT)stack[ sp2 ].i >= stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f >= (PIX_FLOAT)stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = stack[ sp2 ].f >= stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_LSHIFT:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i = stack[ sp2 ].i << (unsigned)stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = stack[ sp2 ].i << (unsigned)(PIX_INT)stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = (PIX_INT)stack[ sp2 ].f << (unsigned)stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = (PIX_INT)stack[ sp2 ].f << (unsigned)(PIX_INT)stack[ sp ].f; break;
		    }
		    sp++;
		    break;
		case OPCODE_RSHIFT:
		    sp2 = sp + 1;
		    switch( ( stack_types[ sp2 ] << 1 ) + stack_types[ sp ] )
		    {
			case 0: /*II*/ stack[ sp2 ].i = stack[ sp2 ].i >> (unsigned)stack[ sp ].i; break;
			case 1: /*IF*/ stack[ sp2 ].i = stack[ sp2 ].i >> (unsigned)(PIX_INT)stack[ sp ].f; break;
			case 2: /*FI*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = (PIX_INT)stack[ sp2 ].f >> (unsigned)stack[ sp ].i; break;
			case 3: /*FF*/ stack_types[ sp2 ] = 0; stack[ sp2 ].i = (PIX_INT)stack[ sp2 ].f >> (unsigned)(PIX_INT)stack[ sp ].f; break;
		    }
		    sp++;
		    break;

		case OPCODE_NEG:
		    if( stack_types[ sp ] == 0 )
			stack[ sp ].i = -stack[ sp ].i;
		    else
			stack[ sp ].f = -stack[ sp ].f;
		    break;
		    
		case OPCODE_CALL_BUILTIN_FN:
		    {
			int pars_num = c >> ( PIX_OPCODE_BITS + PIX_FN_BITS );
			int fn_num = ( c >> PIX_OPCODE_BITS ) & PIX_FN_MASK;
			g_pix_fns[ fn_num ]( fn_num, pars_num, sp, th, vm );
		        sp += pars_num - 1;
		    }
		    break;
		case OPCODE_CALL_BUILTIN_FN_VOID:
		    {
			int pars_num = c >> ( PIX_OPCODE_BITS + PIX_FN_BITS );
			int fn_num = ( c >> PIX_OPCODE_BITS ) & PIX_FN_MASK;
			g_pix_fns[ fn_num ]( fn_num, pars_num, sp, th, vm );
		        sp += pars_num;
		    }
		    break;
		case OPCODE_CALL_i:
		    {
			size_t addr;
		        if( stack_types[ sp ] == 0 )
			    addr = (size_t)stack[ sp ].i;
		        else
			    addr = (size_t)stack[ sp ].f;
			if( IS_ADDRESS_CORRECT( addr ) )
                        {
		    	    //Set new PC:
			    PIX_INT old_pc = (PIX_INT)pc;
			    pc = addr & PIX_INT_ADDRESS_MASK;
		    	    //Save number of parameters:
		    	    stack_types[ sp ] = 0;
		    	    stack[ sp ].i = (PIX_INT)( c >> PIX_OPCODE_BITS );
			    //Save FP:
    			    sp--;
		    	    //Dont touch the type of stack item, because user can't access this item directly
		    	    stack[ sp ].i = (PIX_INT)fp;
			    //Save PC:
			    sp--;
		    	    stack[ sp ].i = old_pc;
		    	    //Set new FP:
		    	    fp = sp + 2;
		    	}
		    	else
		    	{
		    	    int pars = (int)( c >> PIX_OPCODE_BITS );
			    dprint( "Pixilang VM Error. %u: call function i(%d). Address %u is incorrect\n", (unsigned int)pc, pars, (unsigned int)addr );
			    sp += pars - 1;
			    stack_types[ sp ] = 0;
                            stack[ sp ].i = 0;
		    	}
		    }
		    break;
		case OPCODE_INC_SP_i:
		    sp += (signed)c >> PIX_OPCODE_BITS;
		    break;
		case OPCODE_RET_i:
		    {
			PIX_INT pars_num;
			if( stack_types[ fp ] == 0 )
			    pars_num = stack[ fp ].i;
			else
			    pars_num = (PIX_INT)stack[ fp ].f;
			sp = fp + pars_num;
			stack_types[ fp + pars_num ] = 0;
			stack[ fp + pars_num ].i = (PIX_INT)( (signed)c >> PIX_OPCODE_BITS );
			pc = stack[ fp - 2 ].i;
			fp = stack[ fp - 1 ].i;
		    }
		    break;
		case OPCODE_RET_I:
		    {
			LOAD_INT( val_i );
			PIX_INT pars_num;
			if( stack_types[ fp ] == 0 )
			    pars_num = stack[ fp ].i;
			else
			    pars_num = (PIX_INT)stack[ fp ].f;
			sp = fp + pars_num;
			stack_types[ fp + pars_num ] = 0;
			stack[ fp + pars_num ].i = val_i;
			pc = stack[ fp - 2 ].i;
			fp = stack[ fp - 1 ].i;
		    }
		    break;
		case OPCODE_RET:
		    {
			PIX_INT pars_num;
			if( stack_types[ fp ] == 0 )
			    pars_num = stack[ fp ].i;
			else
			    pars_num = (PIX_INT)stack[ fp ].f;
			stack_types[ fp + pars_num ] = stack_types[ sp ];
			stack[ fp + pars_num ] = stack[ sp ];
			sp = fp + pars_num;
			pc = stack[ fp - 2 ].i;
			fp = stack[ fp - 1 ].i;
		    }
		    break;
	    }
	}
	DPRINT( "Thread %d halted. PC: %d; SP: %d; FP: %d.\n", thread_num, (int)pc, (int)sp, (int)fp );
	//Correct values after return from the main function:
	//PC = 1 (next instruction after HALT);
	//SP = PIX_VM_STACK_SIZE - 1 (points to the returned value);
	//FP = PIX_VM_STACK_SIZE - 1;
	th->pc = pc;
	th->sp = sp;
	th->fp = fp;
    }
    
    return thread_num;
}

int pix_vm_save_code( const UTF8_CHAR* name, pix_vm* vm )
{
    int rv = 0;
    
    V3_FILE f = v3_open( name, "wb" );
    if( f )
    {
	const UTF8_CHAR* signature = "PIXICODE";
	v3_write( (void*)signature, 1, 8, f );
	for( int i = 0; i < 8; i++ ) v3_putc( 0, f );
	
	v3_write( &vm->code_size, 1, 4, f );
	v3_write( vm->code, sizeof( PIX_OPCODE ), vm->code_size, f );
	v3_write( &vm->code_ptr, 1, 4, f );
	v3_write( &vm->halt_addr, 1, 4, f );
	
	v3_write( &vm->vars_num, 1, 4, f );
	v3_write( vm->var_types, 1, vm->vars_num, f );
	for( int i = 0; i < (int)vm->vars_num; i++ )
	{
	    if( vm->var_types[ i ] == 0 )
	    {
		//Int:
		PIX_INT v = vm->vars[ i ].i;
		v3_write( &v, 1, sizeof( v ), f );
	    }
	    else
	    {
		//Float:
		PIX_FLOAT v = vm->vars[ i ].f;
		v3_write( &v, 1, sizeof( v ), f );
	    }
	    UTF8_CHAR* var_name = vm->var_names[ i ];
	    int name_len = 0;
	    if( var_name ) name_len = mem_strlen( var_name );
	    v3_write( &name_len, 1, 4, f );
	    if( name_len > 0 )
		v3_write( var_name, 1, name_len, f );
	}
	
	v3_write( &vm->screen, 1, 4, f );
	
	for( int i = 0; i < PIX_VM_FONTS; i++ )
	{
	    pix_vm_font* font = &vm->fonts[ i ];
	    v3_write( &font->font, 1, sizeof( PIX_CID ), f );
	    v3_write( &font->xchars, 1, 4, f );
	    v3_write( &font->ychars, 1, 4, f );
	    v3_write( &font->first, 1, 4, f );
	    v3_write( &font->last, 1, 4, f );
	}
	
	v3_write( &vm->event, 1, sizeof( PIX_CID ), f );
	
	v3_write( &vm->current_path, 1, sizeof( PIX_CID ), f );
	v3_write( &vm->user_path, 1, sizeof( PIX_CID ), f );
	v3_write( &vm->temp_path, 1, sizeof( PIX_CID ), f );
	v3_write( &vm->os_name, 1, sizeof( PIX_CID ), f );
	v3_write( &vm->arch_name, 1, sizeof( PIX_CID ), f );
	
	for( PIX_CID i = 0; i < PIX_VM_CONTAINERS; i++ )
	{
	    pix_vm_container* c = vm->c[ i ];
	    if( c )
	    {
		v3_write( &i, 1, sizeof( PIX_CID ), f );
		v3_write( &c->type, 1, 1, f );
		v3_write( &c->flags, 1, 4, f );
		v3_write( &c->xsize, 1, 4, f );
		v3_write( &c->ysize, 1, 4, f );
		uchar key_color[ 3 ];
		key_color[ 0 ] = red( c->key );
		key_color[ 1 ] = green( c->key );
		key_color[ 2 ] = blue( c->key );
		v3_write( &key_color, 1, 3, f );
		v3_write( &c->alpha, 1, sizeof( PIX_CID ), f );
		v3_write( c->data, g_pix_container_type_sizes[ c->type ], c->xsize * c->ysize, f );
		if( c->type == PIX_CONTAINER_TYPE_DYNAMIC )
		    v3_write( c->data_types, 1, c->xsize * c->ysize, f );
	    }
	}
	
	v3_close( f );
    }
    else
    {
	dprint( "Can't open %s for writing.\n" );
	rv = 1;
    }
    
    return rv;
}

void pix_vm_set_systeminfo_containers( pix_vm* vm )
{
    //System name (container):
    {
        size_t slen = mem_strlen( OS_NAME );
        if( vm->os_name == -1 )
    	    vm->os_name = pix_vm_new_container( -1, slen, 1, PIX_CONTAINER_TYPE_INT8, 0, 0, vm );
    	else
    	    pix_vm_resize_container( vm->os_name, slen, 1, -1, vm );
        mem_copy( pix_vm_get_container_data( vm->os_name, vm ), OS_NAME, slen );
    }
    //Architecture name (container):
    {
        const UTF8_CHAR* name = ARCH_NAME;
        size_t slen = mem_strlen( name );
        if( vm->arch_name == -1 )
    	    vm->arch_name = pix_vm_new_container( -1, slen, 1, PIX_CONTAINER_TYPE_INT8, 0, 0, vm );
    	else
    	    pix_vm_resize_container( vm->arch_name, slen, 1, -1, vm );
        mem_copy( pix_vm_get_container_data( vm->arch_name, vm ), name, slen );
    }
    //Current working path (container):
    {
        UTF8_CHAR* p = v3_make_filename( vm->base_path );
        size_t slen = mem_strlen( p );
        if( vm->current_path == -1 )
    	    vm->current_path = pix_vm_new_container( -1, slen, 1, PIX_CONTAINER_TYPE_INT8, 0, 0, vm );
    	else
    	    pix_vm_resize_container( vm->current_path, slen, 1, -1, vm );
        mem_copy( pix_vm_get_container_data( vm->current_path, vm ), p, slen );
        mem_free( p );
    }
    //User data path (container):
    {
        UTF8_CHAR* p = v3_make_filename( get_user_path() );
        size_t slen = mem_strlen( p );
        if( vm->user_path == -1 )
    	    vm->user_path = pix_vm_new_container( -1, slen, 1, PIX_CONTAINER_TYPE_INT8, 0, 0, vm );
    	else
	    pix_vm_resize_container( vm->user_path, slen, 1, -1, vm );
        mem_copy( pix_vm_get_container_data( vm->user_path, vm ), p, slen );
        mem_free( p );
    }
    //Temp path (container):
    {
        UTF8_CHAR* p = v3_make_filename( get_temp_path() );
        size_t slen = mem_strlen( p );
        if( vm->temp_path == -1 )
    	    vm->temp_path = pix_vm_new_container( -1, slen, 1, PIX_CONTAINER_TYPE_INT8, 0, 0, vm );
    	else
    	    pix_vm_resize_container( vm->temp_path, slen, 1, -1, vm );
        mem_copy( pix_vm_get_container_data( vm->temp_path, vm ), p, slen );
        mem_free( p );
    }
}

int pix_vm_load_code( const UTF8_CHAR* name, UTF8_CHAR* base_path, pix_vm* vm )
{
    int rv = 0;

    V3_FILE f = v3_open( name, "rb" );
    if( f )
    {
	char sign[ 16 ];
	v3_read( sign, 1, 16, f );
	
	v3_read( &vm->code_size, 1, 4, f );
	vm->code = (PIX_OPCODE*)MEM_NEW2( sizeof( PIX_OPCODE ) * vm->code_size );
	v3_read( vm->code, sizeof( PIX_OPCODE ), vm->code_size, f );
	v3_read( &vm->code_ptr, 1, 4, f );
	v3_read( &vm->halt_addr, 1, 4, f );
	
	v3_read( &vm->vars_num, 1, 4, f );
	pix_vm_resize_variables( vm );
	v3_read( vm->var_types, 1, vm->vars_num, f );
	for( int i = 0; i < (int)vm->vars_num; i++ )
    	{
	    if( vm->var_types[ i ] == 0 )
	    {
		//Int:
		PIX_INT v = 0;
		v3_read( &v, 1, sizeof( v ), f );
		vm->vars[ i ].i = v;
	    }
	    else
	    {
		//Float:
		PIX_FLOAT v = 0;
		v3_read( &v, 1, sizeof( v ), f );
		vm->vars[ i ].f = v;
	    }
    	    int name_len;
    	    v3_read( &name_len, 1, 4, f );
    	    if( name_len > 0 )
    	    {
    		UTF8_CHAR* var_name = (UTF8_CHAR*)MEM_NEW2( name_len + 1 );
    		v3_read( var_name, 1, name_len, f );
    		var_name[ name_len ] = 0;
    		vm->var_names[ i ] = var_name;
    	    }
	}
	
	v3_read( &vm->screen, 1, 4, f );
	
	for( int i = 0; i < PIX_VM_FONTS; i++ )
	{
	    pix_vm_font* font = &vm->fonts[ i ];
	    v3_read( &font->font, 1, sizeof( PIX_CID ), f );
	    v3_read( &font->xchars, 1, 4, f );
	    v3_read( &font->ychars, 1, 4, f );
	    v3_read( &font->first, 1, 4, f );
	    v3_read( &font->last, 1, 4, f );
	}
	
	v3_read( &vm->event, 1, sizeof( PIX_CID ), f );

	v3_read( &vm->current_path, 1, sizeof( PIX_CID ), f );
	v3_read( &vm->user_path, 1, sizeof( PIX_CID ), f );
	v3_read( &vm->temp_path, 1, sizeof( PIX_CID ), f );
	v3_read( &vm->os_name, 1, sizeof( PIX_CID ), f );
	v3_read( &vm->arch_name, 1, sizeof( PIX_CID ), f );
	
	while( 1 )
	{
	    PIX_CID cnum;
	    if( v3_read( &cnum, 1, sizeof( PIX_CID ), f ) != 4 ) break;
	    int type = 0;
	    int flags = 0;
	    int xsize = 0;
	    int ysize = 0;
	    uchar key_color[ 3 ];
	    COLOR color;
	    PIX_CID alpha = 0;
	    v3_read( &type, 1, 1, f );
	    v3_read( &flags, 1, 4, f );
	    v3_read( &xsize, 1, 4, f );
	    v3_read( &ysize, 1, 4, f );
	    v3_read( &key_color, 1, 3, f );
	    color = get_color( key_color[ 0 ], key_color[ 1 ], key_color[ 2 ] );
	    v3_read( &alpha, 1, sizeof( PIX_CID ), f );
	    size_t size = g_pix_container_type_sizes[ type ] * xsize * ysize;
	    void* data = MEM_NEW2( size );
	    v3_read( data, 1, size, f ); 
	    pix_vm_new_container( cnum, xsize, ysize, type, data, 0, vm );
	    pix_vm_container* c = vm->c[ cnum ];
	    if( c )
	    {
		pix_vm_set_container_flags( cnum, flags, vm );
		pix_vm_set_container_key_color( cnum, color, vm );
		pix_vm_set_container_alpha( cnum, alpha, vm );
		if( type == PIX_CONTAINER_TYPE_DYNAMIC )
		    v3_read( c->data_types, 1, xsize * ysize, f ); 
	    }
	}
	
	v3_close( f );
	
	//Set base VM path:
	mem_free( vm->base_path );
	vm->base_path = (UTF8_CHAR*)MEM_NEW2( mem_strlen( base_path ) + 1 );
	vm->base_path[ 0 ] = 0;
	mem_strcat( vm->base_path, base_path );
	
        //Set system info containers:
	pix_vm_set_systeminfo_containers( vm );

    }
    else
    {
	dprint( "Can't open %s for reading.\n" );
	rv = 1;
    }

    return rv;
}
