/*
    pixilang_vm_container.cpp
    This file is part of the Pixilang programming language.
    
    [ MIT license ]

    Copyright (c) 2006 - 2012, Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to 
    deal in the Software without restriction, including without limitation the 
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is 
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in 
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.
*/

#include "pixilang.h"
#include "zlib.h"

#include <math.h>

//#define SHOW_DEBUG_MESSAGES

#ifdef SHOW_DEBUG_MESSAGES
#define DPRINT( fmt, ARGS... ) dprint( fmt, ## ARGS )
#else
#define DPRINT( fmt, ARGS... ) {}
#endif

int check_container_parameters( PIX_INT& xsize, PIX_INT& ysize, int& type )
{
    if( xsize <= 0 || ysize <= 0 ) return -1;
    if( type == 32 )
    {
	if( COLORLEN == 1 ) type = PIX_CONTAINER_TYPE_INT8;
	if( COLORLEN == 2 ) type = PIX_CONTAINER_TYPE_INT16;
	if( COLORLEN == 4 ) type = PIX_CONTAINER_TYPE_INT32;
	if( COLORLEN == 8 ) type = PIX_CONTAINER_TYPE_INT64;
    }    
    if( (unsigned)type > PIX_CONTAINER_TYPE_DYNAMIC ) return -1;
    return 0;
}

PIX_CID pix_vm_new_container( PIX_CID cnum, PIX_INT xsize, PIX_INT ysize, int type, void* data, char* data_types, pix_vm* vm )
{
    PIX_CID rv = -1;
    
    if( check_container_parameters( xsize, ysize, type ) != 0 )
	return -1;
    
    sundog_mutex_lock( &vm->c_mutex );
    
    if( cnum >= 0 ) 
	rv = cnum;
    else
    {
	for( PIX_CID i = 0; i < PIX_VM_CONTAINERS; i++ )
	{
	    if( vm->c[ vm->c_counter ] == 0 )
	    {
		rv = vm->c_counter;
		break;
	    }
	    vm->c_counter = ( vm->c_counter + 1 ) & ( PIX_VM_CONTAINERS - 1 );
	}
    }
    if( rv < 0 )
    {
	dprint( "Can't create a new container. All slots (%d) are busy.\n", PIX_VM_CONTAINERS );
    }
    else 
    {
	pix_vm_container* cont = (pix_vm_container*)MEM_NEW2( sizeof( pix_vm_container ) );
	if( cont == 0 ) 
	{
	    dprint( "Can't create a new container. Memory allocation error.\n" );
	    rv = -1;
	}
	else
	{
	    mem_zero( cont );
	    vm->c[ rv ] = cont;
	}
    }
    
    sundog_mutex_unlock( &vm->c_mutex );
    
    if( rv >= 0 )
    {
	pix_vm_container* c = vm->c[ rv ];
	c->type = (pix_container_type)type;
	c->xsize = xsize;
	c->ysize = ysize;
	c->size = (size_t)xsize * (size_t)ysize;
	c->alpha = -1;
	
	DPRINT( "New container %d: %dx%d; type:%s\n", (PIX_CID)rv, (int)xsize, (int)ysize, g_pix_container_type_names[ type ] );

	if( data )
	{
	    c->data = data;
	}
	else
	{
	    size_t data_size = xsize * ysize * g_pix_container_type_sizes[ type ];
	    if( data_size )
	    {
		c->data = MEM_NEW2( data_size );
		if( c->data == 0 )
		{
		    dprint( "Can't create a new container. Memory allocation error.\n" );
		    vm->c[ rv ] = 0;
		    rv = -1;
		}
	    }
	    else 
		c->data = 0;
	}
	if( type == PIX_CONTAINER_TYPE_DYNAMIC )
	{
	    if( data_types )
	    {
		c->data_types = data_types;
	    }
	    else
	    {
		size_t data_types_size = xsize * ysize;
		if( data_types_size )
		{
		    c->data_types = (char*)MEM_NEW2( xsize * ysize );
		    if( c->data_types == 0 )
		    {
	    		dprint( "Can't create a new container. Memory allocation error.\n" );
			vm->c[ rv ] = 0;
			rv = -1;
		    }
		}
		else
		    c->data_types = 0;
	    }
	}
    }
    
    return rv;
}

void pix_vm_remove_container( PIX_CID cnum, pix_vm* vm )
{
    if( vm->c == 0 ) return;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	if( vm->c[ cnum ] )
	{
	    if( cnum == vm->screen )
	    {
		vm->screen_ptr = 0;
		vm->screen = -1;
	    }  
	    pix_vm_container* c = vm->c[ cnum ];
	    if( !( c->flags & PIX_CONTAINER_FLAG_STATIC_DATA ) )
	    {
		mem_free( c->data );
		mem_free( c->data_types );
	    }
	    if( c->opt_data )
	    {
		pix_symtab_deinit( &c->opt_data->props );
		pix_vm_remove_container_hdata( cnum, vm );
#ifdef OPENGL
		pix_vm_remove_container_gl_data( cnum, vm );
#endif
		mem_free( c->opt_data );
	    }
	    mem_free( c );
	    sundog_mutex_lock( &vm->c_mutex );
	    vm->c[ cnum ] = 0;
	    vm->c_counter = cnum;
	    sundog_mutex_unlock( &vm->c_mutex );
	    DPRINT( "Container %d removed.\n", (int)cnum );
	}
    }
}

int pix_vm_resize_container( PIX_CID cnum, PIX_INT xsize, PIX_INT ysize, int type, pix_vm* vm )
{
    if( vm->c == 0 ) return 1;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	if( vm->c[ cnum ] )
	{
	    pix_vm_container* c = vm->c[ cnum ];
	    
	    if( xsize == -1 ) xsize = c->xsize;
	    if( ysize == -1 ) ysize = c->ysize;
	    if( type == -1 ) type = c->type;
	    
	    if( check_container_parameters( xsize, ysize, type ) == 0 )
	    {
		c->type = (pix_container_type)type;
		c->xsize = xsize;
		c->ysize = ysize;
		c->size = (size_t)xsize * (size_t)ysize;
		
		DPRINT( "Resize container %d: %dx%d; type:%s\n", (int)cnum, (int)xsize, (int)ysize, g_pix_container_type_names[ type ] );
		
		if( c->data )
		{
		    if( !( c->flags & PIX_CONTAINER_FLAG_STATIC_DATA ) )
			c->data = mem_resize( c->data, c->size * g_pix_container_type_sizes[ type ] );	
		}
		else
		    c->data = MEM_NEW2( c->size * g_pix_container_type_sizes[ type ] );
		if( c->data == 0 ) return 1;
		if( type == PIX_CONTAINER_TYPE_DYNAMIC )
		{
		    if( c->data_types )
			c->data_types = (char*)mem_resize( c->data_types, xsize * ysize );
		    else
			c->data_types = (char*)MEM_NEW2( xsize * ysize );
		    if( c->data_types == 0 ) return 1;
		}
		
		if( cnum == vm->screen )
		{
		    pix_vm_gfx_set_screen( cnum, vm );
		}
		
#ifdef OPENGL
                pix_vm_remove_container_gl_data( cnum, vm );
#endif		
	    }
	    else 
	    {
		return 2;
	    }
	}
    }
    
    return 0;
}

int pix_vm_convert_container_type( PIX_CID cnum, int type, pix_vm* vm )
{
    if( vm->c == 0 ) return 1;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	if( vm->c[ cnum ] )
	{
	    pix_vm_container* c = vm->c[ cnum ];
	    if( c->data == 0 ) return 1;
	    
	    PIX_INT xsize = 1;
	    PIX_INT ysize = 1;
	    if( check_container_parameters( xsize, ysize, type ) == 0 )
	    {
		if( c->type == type ) return 0;
		size_t old_size = c->size * g_pix_container_type_sizes[ c->type ];
		size_t new_size = c->size * g_pix_container_type_sizes[ type ];
		if( new_size > old_size )
		{
		    if( !( c->flags & PIX_CONTAINER_FLAG_STATIC_DATA ) )
			c->data = mem_resize( c->data, new_size );
		    if( c->data == 0 ) return 1;
		}
		if( type == PIX_CONTAINER_TYPE_DYNAMIC )
		{
		    if( c->data_types == 0 )
			c->data_types = (char*)MEM_NEW2( c->size );
		    if( c->data_types == 0 ) return 1;
		}
		size_t i;
		int add;
		if( old_size < new_size )
		{
		    i = c->size - 1;
		    add = -1;
		}
		else 
		{
		    i = 0;
		    add = 1;
		}
		for( size_t i2 = 0; i2 < c->size; i2++ )
		{
		    char t;
		    signed long long v0;
		    double v1;
		    switch( c->type )
		    {
			case PIX_CONTAINER_TYPE_INT8: v0 = ((signed char*)c->data)[ i ]; t = 0; break;
			case PIX_CONTAINER_TYPE_INT16: v0 = ((signed short*)c->data)[ i ]; t = 0; break;
			case PIX_CONTAINER_TYPE_INT32: v0 = ((signed int*)c->data)[ i ]; t = 0; break;
#ifdef PIX_INT64_ENABLED
			case PIX_CONTAINER_TYPE_INT64: v0 = ((signed long long*)c->data)[ i ]; t = 0; break;
#endif
			case PIX_CONTAINER_TYPE_FLOAT32: v1 = ((float*)c->data)[ i ]; t = 1; break;
#ifdef PIX_FLOAT64_ENABLED
			case PIX_CONTAINER_TYPE_FLOAT64: v1 = ((double*)c->data)[ i ]; t = 1; break;
#endif
			case PIX_CONTAINER_TYPE_DYNAMIC: 
			    if( c->data_types[ i ] == 0 )
			    {
				v0 = ((PIX_VAL*)c->data)[ i ].i; t = 0;
			    }
			    else 
			    {
				v1 = ((PIX_VAL*)c->data)[ i ].f; t = 1;
			    }
			    break;
			default:
			    break;
		    }
		    switch( type )
		    {
			case PIX_CONTAINER_TYPE_INT8: if( t == 0 ) ((signed char*)c->data)[ i ] = v0; else ((signed char*)c->data)[ i ] = v1; break;
			case PIX_CONTAINER_TYPE_INT16: if( t == 0 ) ((signed short*)c->data)[ i ] = v0; else ((signed short*)c->data)[ i ] = v1; break;
			case PIX_CONTAINER_TYPE_INT32: if( t == 0 ) ((signed int*)c->data)[ i ] = v0; else ((signed int*)c->data)[ i ] = v1; break;
#ifdef PIX_INT64_ENABLED
			case PIX_CONTAINER_TYPE_INT64: if( t == 0 ) ((signed long long*)c->data)[ i ] = v0; else ((signed long long*)c->data)[ i ] = v1; break;
#endif
			case PIX_CONTAINER_TYPE_FLOAT32: if( t == 0 ) ((float*)c->data)[ i ] = v0; else ((float*)c->data)[ i ] = v1; break;
#ifdef PIX_FLOAT64_ENABLED
			case PIX_CONTAINER_TYPE_FLOAT64: if( t == 0 ) ((double*)c->data)[ i ] = v0; else ((double*)c->data)[ i ] = v1; break;
#endif
			case PIX_CONTAINER_TYPE_DYNAMIC:
			    c->data_types[ i ] = t;
			    if( t == 0 ) 
				((PIX_VAL*)c->data)[ i ].i = v0;
			    else
				((PIX_VAL*)c->data)[ i ].f = v1;
			    break;
			default:
			    break;
		    }
		    i += add;
		}
		c->type = (pix_container_type)type;
		if( new_size < old_size )
		{
		    c->data = mem_resize( c->data, new_size );
		    if( c->data == 0 ) return 1;
		}
#ifdef OPENGL
                pix_vm_remove_container_gl_data( cnum, vm );
#endif		
	    }
	    else 
	    {
		return 2;
	    }
	}
    }
    
    return 0;
}

void pix_vm_clean_container( PIX_CID cnum, char v_type, PIX_VAL v, PIX_INT offset, PIX_INT size, pix_vm* vm )
{
    if( vm->c == 0 ) return;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( c && c->data )
	{
	    if( size < 0 ) size = c->size;
	    if( offset + size > c->size )
	    {
		size = offset + size - c->size;
	    }
	    if( size == 0 ) return;
	    if( v_type == 0 && v.i == 0 )
	    {
		mem_set( (char*)c->data + offset * g_pix_container_type_sizes[ c->type ], size * g_pix_container_type_sizes[ c->type ], 0 );
		if( c->type == PIX_CONTAINER_TYPE_DYNAMIC )
		    mem_set( (char*)c->data_types + offset, size, 0 );
	    }
	    else 
	    {
		switch( c->type )
		{
		    case PIX_CONTAINER_TYPE_INT8:
			if( v_type == 0 ) 
			    for( size_t i = offset; i < size; i++ ) { ( (signed char*)c->data )[ i ] = (signed char)v.i; }
			else 
			    for( size_t i = offset; i < size; i++ ) { ( (signed char*)c->data )[ i ] = (signed char)v.f; }
			    break;
		    case PIX_CONTAINER_TYPE_INT16:
			if( v_type == 0 ) 
			    for( size_t i = offset; i < size; i++ ) { ( (signed short*)c->data )[ i ] = (signed short)v.i; }
			else 
			    for( size_t i = offset; i < size; i++ ) { ( (signed short*)c->data )[ i ] = (signed short)v.f; }
			    break;
		    case PIX_CONTAINER_TYPE_INT32:
			if( v_type == 0 ) 
			    for( size_t i = offset; i < size; i++ ) { ( (signed int*)c->data )[ i ] = (signed int)v.i; }
			else 
			    for( size_t i = offset; i < size; i++ ) { ( (signed int*)c->data )[ i ] = (signed int)v.f; }
			    break;
#ifdef PIX_INT64_ENABLED
		    case PIX_CONTAINER_TYPE_INT64:
			if( v_type == 0 ) 
			    for( size_t i = offset; i < size; i++ ) { ( (signed long long*)c->data )[ i ] = (signed long long)v.i; }
			else 
			    for( size_t i = offset; i < size; i++ ) { ( (signed long long*)c->data )[ i ] = (signed long long)v.f; }
			    break;
#endif
		    case PIX_CONTAINER_TYPE_FLOAT32:
			if( v_type == 0 ) 
			    for( size_t i = offset; i < size; i++ ) { ( (float*)c->data )[ i ] = (float)v.i; }
			else 
			    for( size_t i = offset; i < size; i++ ) { ( (float*)c->data )[ i ] = (float)v.f; }
			    break;
#ifdef PIX_FLOAT64_ENABLED
		    case PIX_CONTAINER_TYPE_FLOAT64:
			if( v_type == 0 ) 
			    for( size_t i = offset; i < size; i++ ) { ( (double*)c->data )[ i ] = (double)v.i; }
			else 
			    for( size_t i = offset; i < size; i++ ) { ( (double*)c->data )[ i ] = (double)v.f; }
			    break;
#endif
		    case PIX_CONTAINER_TYPE_DYNAMIC:
			for( size_t i = offset; i < size; i++ ) { ( (PIX_VAL*)c->data )[ i ] = v; c->data_types[ i ] = v_type; }
			break;
		}
	    }
	}
    }
}

PIX_CID pix_vm_clone_container( PIX_CID cnum, pix_vm* vm )
{
    if( vm->c == 0 ) return -1;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	
	PIX_CID new_cnum = pix_vm_new_container( -1, c->xsize, c->ysize, c->type, 0, 0, vm );
	if( new_cnum == -1 ) return -1;
	pix_vm_container* c2 = vm->c[ new_cnum ];
	
	size_t data_size = c->xsize * c->ysize * g_pix_container_type_sizes[ c->type ];
	if( data_size )
	    mem_copy( c2->data, c->data, data_size );
	if( c->type == PIX_CONTAINER_TYPE_DYNAMIC )
	    mem_copy( c2->data_types, c->data_types, c->xsize * c->ysize );
	c2->flags = c->flags;
	c2->key = c->key;
	c2->alpha = c->alpha;
	if( c->opt_data )
	{
	    c2->opt_data = (pix_vm_container_opt_data*)MEM_NEW2( sizeof( pix_vm_container_opt_data ) );
	    if( c2->opt_data )
	    {
		mem_copy( c2->opt_data, c->opt_data, sizeof( pix_vm_container_opt_data ) );
		pix_symtab_clone( &c2->opt_data->props, &c->opt_data->props );
		if( c->opt_data->hdata )
		{
	    	    pix_vm_clone_container_hdata( new_cnum, cnum, vm );
		}
#ifdef OPENGL
		c2->opt_data->gl = 0;
#endif
	    }
	}
	
	return new_cnum;
    }
    
    return -1;
}

pix_vm_container* pix_vm_get_container( PIX_CID cnum, pix_vm* vm )
{
    if( vm->c == 0 ) return 0;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	return vm->c[ cnum ];
    }
    
    return 0;
}

void* pix_vm_get_container_data( PIX_CID cnum, pix_vm* vm )
{
    if( vm->c == 0 ) return 0;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( c )
	{
	    return c->data;
	}
    }
    
    return 0;
}

PIX_INT pix_vm_get_container_int_element( PIX_CID cnum, size_t elnum, pix_vm* vm )
{
    if( vm->c == 0 ) return 0;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( elnum >= c->size ) return 0;
	switch( c->type )
        {
            case PIX_CONTAINER_TYPE_INT8:
        	return ((signed char*)c->data)[ elnum ];
        	break;
            case PIX_CONTAINER_TYPE_INT16:
        	return ((signed short*)c->data)[ elnum ];
        	break;
            case PIX_CONTAINER_TYPE_INT32:
        	return ((signed int*)c->data)[ elnum ];
        	break;
#ifdef PIX_INT64_ENABLED        	
            case PIX_CONTAINER_TYPE_INT64:
        	return ((signed long long*)c->data)[ elnum ];
        	break;
#endif
            case PIX_CONTAINER_TYPE_FLOAT32:
        	return ((float*)c->data)[ elnum ];
        	break;
#ifdef PIX_FLOAT64_ENABLED        	
            case PIX_CONTAINER_TYPE_FLOAT64:
        	return ((double*)c->data)[ elnum ];
        	break;
#endif
	    case PIX_CONTAINER_TYPE_DYNAMIC:
		if( c->data_types[ elnum ] == 0 )
		    return ((PIX_VAL*)c->data)[ elnum ].i;
		else
		    return ((PIX_VAL*)c->data)[ elnum ].f;
		break;
        }
    }
    
    return 0;
}

PIX_FLOAT pix_vm_get_container_float_element( PIX_CID cnum, size_t elnum, pix_vm* vm )
{
    if( vm->c == 0 ) return 0;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( elnum >= c->size ) return 0;
	switch( c->type )
        {
            case PIX_CONTAINER_TYPE_INT8:
        	return ((signed char*)c->data)[ elnum ];
        	break;
            case PIX_CONTAINER_TYPE_INT16:
        	return ((signed short*)c->data)[ elnum ];
        	break;
            case PIX_CONTAINER_TYPE_INT32:
        	return ((signed int*)c->data)[ elnum ];
        	break;
#ifdef PIX_INT64_ENABLED        	
            case PIX_CONTAINER_TYPE_INT64:
        	return ((signed long long*)c->data)[ elnum ];
        	break;
#endif
            case PIX_CONTAINER_TYPE_FLOAT32:
        	return ((float*)c->data)[ elnum ];
        	break;
#ifdef PIX_FLOAT64_ENABLED        	
            case PIX_CONTAINER_TYPE_FLOAT64:
        	return ((double*)c->data)[ elnum ];
        	break;
#endif
	    case PIX_CONTAINER_TYPE_DYNAMIC:
		if( c->data_types[ elnum ] == 0 )
		    return ((PIX_VAL*)c->data)[ elnum ].i;
		else
		    return ((PIX_VAL*)c->data)[ elnum ].f;
		break;
        }
    }
    
    return 0;
}

void pix_vm_set_container_int_element( PIX_CID cnum, size_t elnum, PIX_INT val, pix_vm* vm )
{
    if( vm->c == 0 ) return;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( elnum >= c->size ) return;
	switch( c->type )
        {
            case PIX_CONTAINER_TYPE_INT8:
        	((signed char*)c->data)[ elnum ] = val;
        	break;
            case PIX_CONTAINER_TYPE_INT16:
        	((signed short*)c->data)[ elnum ] = val;
        	break;
            case PIX_CONTAINER_TYPE_INT32:
        	((signed int*)c->data)[ elnum ] = val;
        	break;
#ifdef PIX_INT64_ENABLED        	
            case PIX_CONTAINER_TYPE_INT64:
        	((signed long long*)c->data)[ elnum ] = val;
        	break;
#endif
            case PIX_CONTAINER_TYPE_FLOAT32:
        	((float*)c->data)[ elnum ] = val;
        	break;
#ifdef PIX_FLOAT64_ENABLED        	
            case PIX_CONTAINER_TYPE_FLOAT64:
        	((double*)c->data)[ elnum ] = val;
        	break;
#endif
	    case PIX_CONTAINER_TYPE_DYNAMIC:
		c->data_types[ elnum ] = 0;
		((PIX_VAL*)c->data)[ elnum ].i = val;
		break;
        }
    }
}

void pix_vm_set_container_float_element( PIX_CID cnum, size_t elnum, PIX_FLOAT val, pix_vm* vm )
{
    if( vm->c == 0 ) return;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( elnum >= c->size ) return;
	switch( c->type )
        {
            case PIX_CONTAINER_TYPE_INT8:
        	((signed char*)c->data)[ elnum ] = val;
        	break;
            case PIX_CONTAINER_TYPE_INT16:
        	((signed short*)c->data)[ elnum ] = val;
        	break;
            case PIX_CONTAINER_TYPE_INT32:
        	((signed int*)c->data)[ elnum ] = val;
        	break;
#ifdef PIX_INT64_ENABLED        	
            case PIX_CONTAINER_TYPE_INT64:
        	((signed long long*)c->data)[ elnum ] = val;
        	break;
#endif
            case PIX_CONTAINER_TYPE_FLOAT32:
        	((float*)c->data)[ elnum ] = val;
        	break;
#ifdef PIX_FLOAT64_ENABLED        	
            case PIX_CONTAINER_TYPE_FLOAT64:
        	((double*)c->data)[ elnum ] = val;
        	break;
#endif
	    case PIX_CONTAINER_TYPE_DYNAMIC:
		c->data_types[ elnum ] = 1;
		((PIX_VAL*)c->data)[ elnum ].f = val;
		break;
        }
    }
}

int pix_vm_get_container_flags( PIX_CID cnum, pix_vm* vm )
{
    if( vm->c == 0 ) return 0;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( c )
	{
	    return c->flags;
	}
    }
    
    return 0;
}

void pix_vm_set_container_flags( PIX_CID cnum, int flags, pix_vm* vm )
{
    if( vm->c == 0 ) return;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( c )
	{
	    c->flags = flags;
	}
    }
}

COLOR pix_vm_get_container_key_color( PIX_CID cnum, pix_vm* vm )
{
    if( vm->c == 0 ) return 0;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( c )
	{
	    return c->key;
	}
    }
    
    return 0;
}

void pix_vm_set_container_key_color( PIX_CID cnum, COLOR key, pix_vm* vm )
{
    if( vm->c == 0 ) return;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( c )
	{
	    c->key = key;
	    c->flags |= PIX_CONTAINER_FLAG_USES_KEY;
	}
    }
}

void pix_vm_remove_container_key_color( PIX_CID cnum, pix_vm* vm )
{
    if( vm->c == 0 ) return;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( c )
	{
	    c->flags &= ~PIX_CONTAINER_FLAG_USES_KEY;
	}
    }
}

PIX_CID pix_vm_get_container_alpha( PIX_CID cnum, pix_vm* vm )
{
    if( vm->c == 0 ) return 0;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( c )
	{
	    return c->alpha;
	}
    }
    
    return 0;
}

void* pix_vm_get_container_alpha_data( PIX_CID cnum, pix_vm* vm )
{
    if( vm->c == 0 ) return 0;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( c )
	{
	    if( (unsigned)c->alpha < (unsigned)PIX_VM_CONTAINERS )
	    {
		pix_vm_container* alpha_cont = vm->c[ c->alpha ];
		return alpha_cont->data;
	    }
	}
    }
    
    return 0;
}

void pix_vm_set_container_alpha( PIX_CID cnum, PIX_CID alpha, pix_vm* vm )
{
    if( vm->c == 0 ) return;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( c )
	{
	    c->alpha = alpha;
	}
    }
}

size_t pix_vm_get_container_strlen( PIX_CID cnum, size_t offset, pix_vm* vm )
{
    if( vm->c == 0 ) return 0;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( c )
	{
	    size_t size = c->size * g_pix_container_type_sizes[ c->type ];
	    size_t len;
	    if( offset >= size )
		len = 0;
	    else
	    {
		UTF8_CHAR* s = (UTF8_CHAR*)c->data;
		for( len = offset; len < size; len++ )
		    if( s[ len ] == 0 ) break;
		len -= offset;
	    }
	    return len;
	}
    }
    
    return 0;
}

UTF8_CHAR* pix_vm_make_cstring_from_container( PIX_CID cnum, bool* need_to_free, pix_vm* vm )
{
    *need_to_free = 0;
    
    if( vm->c == 0 ) return 0;

    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( c )
	{
	    size_t size = c->size * g_pix_container_type_sizes[ c->type ];
	    size_t str_len;
	    UTF8_CHAR* str_ptr = (char*)c->data;
	    for( str_len = 0; str_len < size; str_len++ )
		if( str_ptr[ str_len ] == 0 ) break;
	    if( str_len == size )
	    {
		UTF8_CHAR* str = (UTF8_CHAR*)MEM_NEW2( str_len + 1 );
		mem_copy( str, str_ptr, str_len );
		str[ str_len ] = 0;
		*need_to_free = 1;
		return str;
	    }
	    else
	    {
		*need_to_free = 0;
		return str_ptr;
	    }
	}
    }
    
    return 0;
}

//
// Properties
//

pix_sym* pix_vm_get_container_property( PIX_CID cnum, const UTF8_CHAR* prop_name, int prop_hash, pix_vm* vm )
{
    if( vm->c == 0 ) return 0;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( c && c->opt_data )
	{
	    return pix_symtab_lookup( prop_name, prop_hash, 0, SYMTYPE_NUM_I, 0, 0, 0, &c->opt_data->props );
	}
    }
    
    return 0;
}

PIX_INT pix_vm_get_container_property_i( PIX_CID cnum, const UTF8_CHAR* prop_name, int prop_hash, pix_vm* vm )
{
    pix_sym* s = pix_vm_get_container_property( cnum, prop_name, prop_hash, vm );
    if( s )
    {
	if( s->type == SYMTYPE_NUM_I )
	    return s->val.i;
	else
	    return (PIX_INT)s->val.f;
    }
    return 0;
}

void pix_vm_set_container_property( PIX_CID cnum, const UTF8_CHAR* prop_name, int prop_hash, char val_type, PIX_VAL val, pix_vm* vm )
{
    if( vm->c == 0 ) return;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
	pix_vm_container* c = vm->c[ cnum ];
	if( c )
	{
	    if( c->opt_data == 0 )
	    {
		c->opt_data = (pix_vm_container_opt_data*)MEM_NEW2( sizeof( pix_vm_container_opt_data ) );
		mem_zero( c->opt_data );
		if( c->opt_data == 0 ) return;
	    }
	    if( c->opt_data->props.symtab == 0 )
	    {
		pix_symtab_init( PIX_CONTAINER_SYMTAB_SIZE, &c->opt_data->props );
	    }
	    pix_sym* sym = pix_symtab_lookup( prop_name, prop_hash, 1, SYMTYPE_NUM_I, 0, 0, 0, &c->opt_data->props );
	    if( sym )
	    {
		if( val_type == 0 )
		    sym->type = SYMTYPE_NUM_I;
		else
		    sym->type = SYMTYPE_NUM_F;
		sym->val = val;
	    }
	}
    }
}

//
// Hidden data
//

void* pix_vm_get_container_hdata( PIX_CID cnum, pix_vm* vm )
{
    if( vm->c == 0 ) return 0;

    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
        pix_vm_container* c = vm->c[ cnum ];
	if( c && c->opt_data && c->opt_data->hdata )
	{
	    return c->opt_data->hdata;
	}
    }
    
    return 0;
}

int pix_vm_create_container_hdata( PIX_CID cnum, uchar hdata_type, size_t hdata_size, pix_vm* vm )
{
    int rv = -1;
    if( vm->c == 0 ) return rv;

    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
        pix_vm_container* c = vm->c[ cnum ];
        if( c )
        {
	    if( c->opt_data == 0 )
	    {
		c->opt_data = (pix_vm_container_opt_data*)MEM_NEW2( sizeof( pix_vm_container_opt_data ) );
		mem_zero( c->opt_data );
	    }
	    if( c->opt_data )
	    {
		c->opt_data->hdata = MEM_NEW2( hdata_size );
		if( c->opt_data->hdata )
		{
		    mem_zero( c->opt_data->hdata );
		    ((uchar*)c->opt_data->hdata)[ 0 ] = hdata_type;
		    rv = 0;
		}
	    }
	}
    }
    
    return rv;
}

void pix_vm_remove_container_hdata( PIX_CID cnum, pix_vm* vm )
{
    if( vm->c == 0 ) return;

    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
        pix_vm_container* c = vm->c[ cnum ];
        if( c && c->opt_data && c->opt_data->hdata )
        {
    	    int hdata_type = ((uchar*)c->opt_data->hdata)[ 0 ];
    	    switch( hdata_type )
    	    {
    		case pix_vm_container_hdata_type_anim:
		    {
			pix_vm_container_hdata_anim* hdata = (pix_vm_container_hdata_anim*)c->opt_data->hdata;
			for( int i = 0; i < hdata->frame_count; i++ )
			{
			    mem_free( hdata->frames[ i ].pixels );
			}
			mem_free( hdata->frames );
		    }
    		    break;
    		default:
    		    break;
    	    }
    	    mem_free( c->opt_data->hdata );
    	    c->opt_data->hdata = 0;
        }
    }
}

size_t pix_vm_get_container_hdata_size( PIX_CID cnum, pix_vm* vm )
{
    size_t rv = 0;
    
    if( vm->c == 0 ) return 0;

    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
        pix_vm_container* c = vm->c[ cnum ];
        if( c && c->opt_data && c->opt_data->hdata )
        {
	    rv += 1; //type
    	    int hdata_type = ((uchar*)c->opt_data->hdata)[ 0 ];
    	    switch( hdata_type )
    	    {
    		case pix_vm_container_hdata_type_anim:
		    {
			rv += 4; //frame_count
			pix_vm_container_hdata_anim* hdata = (pix_vm_container_hdata_anim*)c->opt_data->hdata;
			for( int i = 0; i < hdata->frame_count; i++ )
			{
			    rv += 1; //frame type
			    rv += 8; //xsize
			    rv += 8; //ysize
			    rv += 8; //data size
			    rv += mem_get_size( hdata->frames[ i ].pixels );
			}
		    }
    		    break;
    		default:
    		    break;
    	    }
        }
    }
    
    return rv;
}

size_t pix_vm_save_container_hdata( PIX_CID cnum, V3_FILE f, pix_vm* vm )
{
    size_t rv = 0;
    
    if( vm->c == 0 ) return 0;

    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
        pix_vm_container* c = vm->c[ cnum ];
        if( c && c->opt_data && c->opt_data->hdata )
        {
    	    int hdata_type = ((uchar*)c->opt_data->hdata)[ 0 ];
	    if( v3_putc( hdata_type, f ) == EOF ) return rv;
	    rv += 1;
    	    switch( hdata_type )
    	    {
    		case pix_vm_container_hdata_type_anim:
		    {
			pix_vm_container_hdata_anim* hdata = (pix_vm_container_hdata_anim*)c->opt_data->hdata;
			if( v3_write( &hdata->frame_count, 1, 4, f ) != 4 ) return rv;
			rv += 4;
			for( int i = 0; i < hdata->frame_count; i++ )
			{
			    pix_vm_anim_frame* frame = &hdata->frames[ i ];
			    if( v3_putc( frame->type, f ) == EOF ) return rv;
			    rv += 1;
			    uint64 v;
			    size_t size = mem_get_size( frame->pixels );
			    v = frame->xsize; if( v3_write( &v, 1, sizeof( uint64 ), f ) != sizeof( uint64 ) ) return rv;
			    v = frame->ysize; if( v3_write( &v, 1, sizeof( uint64 ), f ) != sizeof( uint64 ) ) return rv;
			    v = size; if( v3_write( &v, 1, sizeof( uint64 ), f ) != sizeof( uint64 ) ) return rv;
			    rv += 8;
			    rv += 8;
			    rv += 8;
			    if( v3_write( frame->pixels, 1, size, f ) != size ) return rv;
			    rv += size;
			}
		    }
    		    break;
    		default:
    		    break;
    	    }
        }
    }
    
    return rv;
}

size_t pix_vm_load_container_hdata( PIX_CID cnum, V3_FILE f, pix_vm* vm )
{
    size_t rv = 0;

    if( vm->c == 0 ) return 0;
    
    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
        pix_vm_container* c = vm->c[ cnum ];
        if( c == 0 ) return 0;
            
	int hdata_type = v3_getc( f );
	if( hdata_type < 0 ) return rv;
	rv += 1;
	switch( hdata_type )
	{
    	    case pix_vm_container_hdata_type_anim:
        	{
	    	    if( pix_vm_create_container_hdata( cnum, hdata_type, sizeof( pix_vm_container_hdata_anim ), vm ) ) return rv;
	    	    if( c->opt_data == 0 ) return rv;
	    	    if( c->opt_data->hdata == 0 ) return rv;
	    	    pix_vm_container_hdata_anim* hdata = (pix_vm_container_hdata_anim*)c->opt_data->hdata;
	    	    if( v3_read( &hdata->frame_count, 1, 4, f ) != 4 ) return rv;
	    	    rv += 4;
	    	    hdata->frames = (pix_vm_anim_frame*)MEM_NEW2( sizeof( pix_vm_anim_frame ) * hdata->frame_count );
	    	    if( hdata->frames == 0 ) return rv;
	    	    mem_zero( hdata->frames );
	    	    for( int i = 0; i < hdata->frame_count; i++ )
            	    {
            		pix_vm_anim_frame* frame = &hdata->frames[ i ];
            		int frame_type = v3_getc( f );
            		if( frame_type < 0 ) return rv;
            		frame->type = (pix_container_type)frame_type;
            		rv += 1;
            		uint64 xsize = 0;
            		uint64 ysize = 0;
            		uint64 size = 0;
            		if( v3_read( &xsize, 1, sizeof( uint64 ), f ) != sizeof( uint64 ) ) return rv;
            		if( v3_read( &ysize, 1, sizeof( uint64 ), f ) != sizeof( uint64 ) ) return rv;
            		if( v3_read( &size, 1, sizeof( uint64 ), f ) != sizeof( uint64 ) ) return rv;
            		frame->xsize = (PIX_INT)xsize;
            		frame->ysize = (PIX_INT)ysize;
            		rv += 8;
            		rv += 8;
            		rv += 8;
            		frame->pixels = (COLORPTR)MEM_NEW2( size );
            		if( frame->pixels == 0 ) return rv;
            		if( v3_read( frame->pixels, 1, size, f ) != size ) return rv;
            		rv += size;
        	    }
		}
		break;
	    default:
		break;
	}
    }
    
    return rv;
}

int pix_vm_clone_container_hdata( PIX_CID new_cnum, PIX_CID old_cnum, pix_vm* vm )
{
    int rv = -1;
    if( vm->c == 0 ) return rv;

    if( (unsigned)new_cnum < (unsigned)PIX_VM_CONTAINERS && (unsigned)old_cnum < (unsigned)PIX_VM_CONTAINERS )
    {
        pix_vm_container* c2 = vm->c[ new_cnum ];
        pix_vm_container* c1 = vm->c[ old_cnum ];
        if( c1 && c1->opt_data && c1->opt_data->hdata && c2 )
        {
    	    if( c2->opt_data == 0 )
    	    {
    		c2->opt_data = (pix_vm_container_opt_data*)MEM_NEW2( sizeof( pix_vm_container_opt_data ) );
    		if( c2->opt_data == 0 ) return -1;
    		mem_zero( c2->opt_data );
    	    }
    	    int hdata_type = ((uchar*)c1->opt_data->hdata)[ 0 ];
    	    size_t hdata_size = mem_get_size( c1->opt_data->hdata );
    	    c2->opt_data->hdata = MEM_NEW2( hdata_size );
    	    if( c2->opt_data->hdata == 0 ) return -1;
    	    mem_copy( c2->opt_data->hdata, c1->opt_data->hdata, hdata_size );
    	    switch( hdata_type )
    	    {
    		case pix_vm_container_hdata_type_anim:
		    {
			pix_vm_container_hdata_anim* hdata1 = (pix_vm_container_hdata_anim*)c1->opt_data->hdata;
			pix_vm_container_hdata_anim* hdata2 = (pix_vm_container_hdata_anim*)c2->opt_data->hdata;
			hdata2->frames = (pix_vm_anim_frame*)MEM_NEW2( sizeof( pix_vm_anim_frame ) * hdata1->frame_count );
			if( hdata2->frames == 0 ) return -1;
			mem_zero( hdata2->frames );
			for( int i = 0; i < hdata1->frame_count; i++ )
			{
			    mem_copy( &hdata2->frames[ i ], &hdata1->frames[ i ], sizeof( pix_vm_anim_frame ) );
			    size_t frame_size = mem_get_size( hdata1->frames[ i ].pixels );
			    hdata2->frames[ i ].pixels = (COLORPTR)MEM_NEW2( frame_size );
			    if( hdata2->frames[ i ].pixels )
			    {
				mem_copy( hdata2->frames[ i ].pixels, hdata1->frames[ i ].pixels, frame_size );
			    }
			}
		    }
    		    break;
    		default:
    		    break;
    	    }
	    rv = 0;
        }
    }
    
    return rv;
}

PIX_INT pix_vm_container_get_cur_frame( PIX_CID cnum, pix_vm* vm )
{
    pix_vm_container* c = pix_vm_get_container( cnum, vm );
    if( c && c->opt_data && c->opt_data->hdata )
    {
	PIX_INT repeat = pix_vm_get_container_property_i( cnum, "repeat", -1, vm );
	PIX_INT cur_frame = pix_vm_get_container_property_i( cnum, "frame", -1, vm );
	if( cur_frame < 0 ) cur_frame = 0;
	switch( ((uchar*)c->opt_data->hdata)[ 0 ] )
	{
	    case pix_vm_container_hdata_type_anim:
		{
		    pix_vm_container_hdata_anim* hdata = (pix_vm_container_hdata_anim*)c->opt_data->hdata;
		    if( repeat >= 0 )
		    {
			if( cur_frame >= ( 1 + repeat ) * hdata->frame_count )
			    cur_frame = ( 1 + repeat ) * hdata->frame_count - 1;
		    }
		    cur_frame %= hdata->frame_count;
		}
		break;
	}
	return cur_frame;
    }
    return 0;
}

int pix_vm_container_hdata_get_frame_count( PIX_CID cnum, pix_vm* vm )
{
    int rv = 0;
    if( vm->c == 0 ) return rv;
    
    pix_vm_container* c = pix_vm_get_container( cnum, vm );
    if( c && c->opt_data && c->opt_data->hdata )
    {
	switch( ((uchar*)c->opt_data->hdata)[ 0 ] )
	{
	    case pix_vm_container_hdata_type_anim:
		{
		    pix_vm_container_hdata_anim* hdata = (pix_vm_container_hdata_anim*)c->opt_data->hdata;
		    if( hdata->frames )
		    {
			rv = hdata->frame_count;
		    }
		}
		break;
	}
    }
    
    return rv;
}

int pix_vm_container_hdata_get_frame_size( PIX_CID cnum, int cur_frame, pix_container_type* type, int* xsize, int* ysize, pix_vm* vm )
{
    int rv = -1;
    if( vm->c == 0 ) return rv;
    
    pix_vm_container* c = pix_vm_get_container( cnum, vm );
    if( c && c->opt_data && c->opt_data->hdata )
    {
	switch( ((uchar*)c->opt_data->hdata)[ 0 ] )
	{
	    case pix_vm_container_hdata_type_anim:
		{
		    pix_vm_container_hdata_anim* hdata = (pix_vm_container_hdata_anim*)c->opt_data->hdata;
		    if( hdata->frames )
		    {
			pix_vm_anim_frame* f = &hdata->frames[ cur_frame ];
			if( type ) *type = f->type;
			if( xsize ) *xsize = f->xsize;
			if( ysize ) *ysize = f->ysize;
			rv = 0;
		    }
		}
		break;
	}
    }
    
    return rv;
}

int pix_vm_container_hdata_unpack_frame_to_buf( PIX_CID cnum, int cur_frame, COLORPTR buf, pix_vm* vm )
{
    int rv = -1;
    if( vm->c == 0 ) return -1;
    if( buf == 0 ) return -1;
    
    pix_vm_container* c = pix_vm_get_container( cnum, vm );
    if( c && c->opt_data && c->opt_data->hdata )
    {
#ifdef OPENGL
	pix_vm_remove_container_gl_data( cnum, vm );
#endif	
	switch( ((uchar*)c->opt_data->hdata)[ 0 ] )
	{
	    case pix_vm_container_hdata_type_anim:
		{
		    pix_vm_container_hdata_anim* hdata = (pix_vm_container_hdata_anim*)c->opt_data->hdata;
		    if( hdata->frames )
		    {
			pix_vm_anim_frame* f = &hdata->frames[ cur_frame ];
			
			z_stream strm;
			mem_set( &strm, sizeof( z_stream ), 0 );
			int zlib_err;
			strm.zalloc = Z_NULL;
			strm.zfree = Z_NULL;
			strm.opaque = Z_NULL;
			strm.avail_in = 0;
			strm.next_in = Z_NULL;
			zlib_err = inflateInit( &strm );
			if( zlib_err != Z_OK )
			{
			    dprint( "zlib inflateInit error %d\n", zlib_err );
			    break;
			}
			
			size_t in_p = 0;
			size_t out_p = 0;
			size_t zlib_chunk_size = 16384;
			size_t in_size = mem_get_size( f->pixels );
			size_t out_size = f->xsize * f->ysize * g_pix_container_type_sizes[ f->type ];
			uchar* in = (uchar*)f->pixels;
			uchar* out = (uchar*)buf;
			if( out )
			{
			    in_p = 0;
			    while( in_p < in_size )
			    {
				size_t size = in_size - in_p;
				if( size > zlib_chunk_size )
				    size = zlib_chunk_size;
				strm.avail_in = size;
				strm.next_in = in + in_p;
				do {
				    size_t avail = out_size - out_p;
				    strm.avail_out = avail;
				    strm.next_out = out + out_p;
				    zlib_err = inflate( &strm, Z_NO_FLUSH );
				    if( zlib_err < 0 ) 
				    {
					dprint( "zlib inflate error %d\n", zlib_err );
					break;
				    }
				    size_t size2 = avail - strm.avail_out;
				    out_p += size2;
				    if( out_p >= out_size ) break;
				} while( strm.avail_out == 0 );
				in_p += size;
			    }
			    rv = 0;
			}
			
			inflateEnd( &strm );
		    }
		}
		break;
	}
    }
    
    return rv;
}

int pix_vm_container_hdata_unpack_frame( PIX_CID cnum, pix_vm* vm )
{
    int rv = -1;
    pix_container_type type;
    int xsize, ysize;
    int cur_frame = pix_vm_container_get_cur_frame( cnum, vm );
    rv = pix_vm_container_hdata_get_frame_size( cnum, cur_frame, &type, &xsize, &ysize, vm );
    if( rv < 0 ) return rv;
    pix_vm_resize_container( cnum, xsize, ysize, type, vm );
    COLORPTR cdata = (COLORPTR)pix_vm_get_container_data( cnum, vm );
    if( cdata )
    {
	rv = pix_vm_container_hdata_unpack_frame_to_buf( cnum, cur_frame, cdata, vm );
    }
    else rv = -1;
    return rv;
}

int pix_vm_container_hdata_pack_frame_from_buf( PIX_CID cnum, int cur_frame, COLORPTR buf, pix_container_type type, int xsize, int ysize, pix_vm* vm )
{
    int rv = -1;
    if( vm->c == 0 ) return -1;
    if( buf == 0 ) return -1;
    
    pix_vm_container* c = pix_vm_get_container( cnum, vm );
    if( c && c->opt_data && c->opt_data->hdata )
    {
	PIX_INT repeat = pix_vm_get_container_property_i( cnum, "repeat", -1, vm );
	switch( ((uchar*)c->opt_data->hdata)[ 0 ] )
	{
	    case pix_vm_container_hdata_type_anim:
		{
		    pix_vm_container_hdata_anim* hdata = (pix_vm_container_hdata_anim*)c->opt_data->hdata;
		    if( hdata->frames )
		    {
			pix_vm_anim_frame* f = &hdata->frames[ cur_frame ];
			
			z_stream strm;
			mem_set( &strm, sizeof( z_stream ), 0 );
			int zlib_err;
			strm.zalloc = Z_NULL;
			strm.zfree = Z_NULL;
			strm.opaque = Z_NULL;
			zlib_err = deflateInit( &strm, 2 );
			if( zlib_err != Z_OK )
			{
			    dprint( "zlib deflateInit error %d\n", zlib_err );
			    break;
			}
			
			size_t frame_size = xsize * ysize * g_pix_container_type_sizes[ type ];
			size_t in_p = 0;
			size_t out_p = 0;
			size_t zlib_chunk_size = 16384;
			size_t out_size = zlib_chunk_size;
			uchar* out = (uchar*)MEM_NEW2( out_size );
			if( out )
			{
			    in_p = 0;
			    while( in_p < frame_size )
			    {
				int flush = Z_NO_FLUSH;
				size_t size = frame_size - in_p;
				if( size > zlib_chunk_size )
				    size = zlib_chunk_size;
				else
				    flush = Z_FINISH;
				strm.avail_in = size;
				strm.next_in = (uchar*)buf + in_p;
				do {
				    size_t avail = out_size - out_p;
				    strm.avail_out = avail;
				    strm.next_out = out + out_p;
				    zlib_err = deflate( &strm, flush );
				    if( zlib_err < 0 ) 
				    {
					dprint( "zlib deflate error %d\n", zlib_err );
					break;
				    }
				    size_t size2 = avail - strm.avail_out;
				    out_p += size2;
				    if( out_p >= out_size )
				    {
					out_size += zlib_chunk_size;
					out = (uchar*)mem_resize( out, out_size );
					if( out == 0 ) break;
				    }
				} while( strm.avail_out == 0 );
				in_p += size;
			    }
			    out = (uchar*)mem_resize( out, out_p );
			    if( out )
			    {
				mem_free( f->pixels );
				f->type = type;
				f->xsize = xsize;
				f->ysize = ysize;
				f->pixels = (COLORPTR)out;
				rv = 0;
			    }
			}
			
			deflateEnd( &strm );
		    }
		}
		break;
	}
    }
    
    return rv;
}

int pix_vm_container_hdata_pack_frame( PIX_CID cnum, pix_vm* vm )
{
    int rv = -1;
    pix_vm_container* c = pix_vm_get_container( cnum, vm );
    if( c && c->data )
    {
	int cur_frame = pix_vm_container_get_cur_frame( cnum, vm );
	rv = pix_vm_container_hdata_pack_frame_from_buf( cnum, cur_frame, (COLORPTR)c->data, c->type, c->xsize, c->ysize, vm );
    }
    return rv;
}

int pix_vm_container_hdata_clone_frame( PIX_CID cnum, pix_vm* vm )
{	
    int rv = -1;
    if( vm->c == 0 ) return rv;
    
    pix_vm_container* c = pix_vm_get_container( cnum, vm );
    if( c && c->opt_data && c->opt_data->hdata )
    {
	int cur_frame = pix_vm_container_get_cur_frame( cnum, vm );
	switch( ((uchar*)c->opt_data->hdata)[ 0 ] )
	{
	    case pix_vm_container_hdata_type_anim:
		{
		    pix_vm_container_hdata_anim* hdata = (pix_vm_container_hdata_anim*)c->opt_data->hdata;
		    hdata->frames = (pix_vm_anim_frame*)mem_resize( hdata->frames, mem_get_size( hdata->frames ) + sizeof( pix_vm_anim_frame ) );
		    if( hdata->frames )
		    {
			hdata->frame_count++;
			PIX_VAL prop_val;
			prop_val.i = hdata->frame_count; pix_vm_set_container_property( cnum, "frames", -1, 0, prop_val, vm );
			for( int i = hdata->frame_count - 1; i > cur_frame; i-- )
			{
			    mem_copy( &hdata->frames[ i ], &hdata->frames[ i - 1 ], sizeof( pix_vm_anim_frame ) );
			}
			pix_vm_anim_frame* f = &hdata->frames[ cur_frame ];
			pix_vm_anim_frame* f2 = &hdata->frames[ cur_frame + 1 ];
			f->pixels = (COLORPTR)MEM_NEW2( mem_get_size( f2->pixels ) );
			if( f->pixels )
			{
			    mem_copy( f->pixels, f2->pixels, mem_get_size( f2->pixels ) );
			    rv = 0;
			}
		    }
		}
		break;
	}
    }
    
    return rv;
}

int pix_vm_container_hdata_remove_frame( PIX_CID cnum, pix_vm* vm )
{	
    int rv = -1;
    if( vm->c == 0 ) return rv;
    
    pix_vm_container* c = pix_vm_get_container( cnum, vm );
    if( c && c->opt_data && c->opt_data->hdata )
    {
	int cur_frame = pix_vm_container_get_cur_frame( cnum, vm );
	switch( ((uchar*)c->opt_data->hdata)[ 0 ] )
	{
	    case pix_vm_container_hdata_type_anim:
		{
		    pix_vm_container_hdata_anim* hdata = (pix_vm_container_hdata_anim*)c->opt_data->hdata;
		    if( hdata->frames && hdata->frame_count > 1 )
		    {
			pix_vm_anim_frame* f = &hdata->frames[ cur_frame ];
			mem_free( f->pixels );
			for( int i = cur_frame; i < hdata->frame_count; i++ )
			{
			    mem_copy( &hdata->frames[ i ], &hdata->frames[ i + 1 ], sizeof( pix_vm_anim_frame ) );
			}
			hdata->frame_count--;
			PIX_VAL prop_val;
			prop_val.i = hdata->frame_count; pix_vm_set_container_property( cnum, "frames", -1, 0, prop_val, vm );
			hdata->frames = (pix_vm_anim_frame*)mem_resize( hdata->frames, hdata->frame_count * sizeof( pix_vm_anim_frame ) );
			if( hdata->frames )
			    rv = 0;
		    }
		}
		break;
	}
    }
    
    return rv;
}

int pix_vm_container_hdata_autoplay_control( PIX_CID cnum, pix_vm* vm )
{
    int rv = -1;
    if( vm->c == 0 ) return rv;
    
    PIX_VAL prop_val;

    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
        pix_vm_container* c = vm->c[ cnum ];
        if( c && c->opt_data && c->opt_data->hdata )
        {
	    if( pix_vm_get_container_property_i( cnum, "play", -1, vm ) )
	    {
		uint frame_count = 0;
		int hdata_type = ((uchar*)c->opt_data->hdata)[ 0 ];
		switch( hdata_type )
		{
		    case pix_vm_container_hdata_type_anim:
			{
			    pix_vm_container_hdata_anim* hdata = (pix_vm_container_hdata_anim*)c->opt_data->hdata;
			    frame_count = hdata->frame_count;
			}
			break;
		    default:
			break;
		}
		if( frame_count > 1 )
		{
		    PIX_INT start_frame = pix_vm_get_container_property_i( cnum, "start_frame", -1, vm );
		    uint start_time = pix_vm_get_container_property_i( cnum, "start_time", -1, vm );
		    uint cur_time = (uint)time_ticks_hires();
		    uint t = cur_time - start_time;
		    PIX_INT fps = pix_vm_get_container_property_i( cnum, "fps", -1, vm );
		    if( fps <= 0 ) fps = 1;
		    PIX_INT repeat = pix_vm_get_container_property_i( cnum, "repeat", -1, vm );
		    long long ff = ( t * fps ) / time_ticks_per_second_hires();
		    PIX_INT f = start_frame + ff;

		    if( repeat >= 0 )
		    {
			if( f >= ( 1 + repeat ) * frame_count )
			{
			    prop_val.i = 0; pix_vm_set_container_property( cnum, "play", -1, 0, prop_val, vm );
			    f = ( 1 + repeat ) * frame_count - 1;
			}
		    }

		    PIX_INT prev_frame = pix_vm_get_container_property_i( cnum, "frame", -1, vm );
		    if( f != prev_frame )
		    {
			prop_val.i = f; pix_vm_set_container_property( cnum, "frame", -1, 0, prop_val, vm );
			rv = f % frame_count;
		    }
		}
	    }
	}
    }
    
    return rv;
}

//
// OpenGL data
//

#ifdef OPENGL

pix_vm_container_gl_data* pix_vm_get_container_gl_data( PIX_CID cnum, pix_vm* vm )
{
    if( vm->c == 0 ) return 0;

    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
        pix_vm_container* c = vm->c[ cnum ];
	if( c && c->opt_data && c->opt_data->gl )
	{
	    return c->opt_data->gl;
	}
    }
    
    return 0;
}

void clear_fbox( uchar* img, int img_xsize, int pixel_size, int x, int y, int xsize, int ysize )
{
    img_xsize *= pixel_size;
    x *= pixel_size;
    xsize *= pixel_size;
    uchar* ptr = img + y * img_xsize + x;
    for( int y = 0; y < ysize; y++ )
    {
	for( int x = 0; x < xsize; x++ )
	{
	    *ptr = 0;
	    ptr++;
	}
	ptr += img_xsize - xsize;
    }
}

int pix_vm_create_container_gl_data( PIX_CID cnum, pix_vm* vm )
{
    int rv = -1;
    if( vm->c == 0 ) return rv;

    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
        pix_vm_container* c = vm->c[ cnum ];
        if( c )
        {
	    if( c->opt_data == 0 )
	    {
		c->opt_data = (pix_vm_container_opt_data*)MEM_NEW2( sizeof( pix_vm_container_opt_data ) );
		mem_zero( c->opt_data );
	    }
	    if( c->opt_data )
	    {
		if( c->opt_data->gl ) return -1; //Already created
		c->opt_data->gl = (pix_vm_container_gl_data*)MEM_NEW2( sizeof( pix_vm_container_gl_data ) );
		if( c->opt_data->gl )
		{
		    pix_vm_empty_gl_trash( vm );
		
		    pix_vm_container_gl_data* gl = c->opt_data->gl;
		    
		    int size_changed = 0;
		    gl->xsize = round_to_power_of_two( c->xsize );
		    gl->ysize = round_to_power_of_two( c->ysize );
		    if( gl->xsize != c->xsize )
			size_changed |= 1;
		    if( gl->ysize != c->ysize )
			size_changed |= 2;

		    uchar* temp_pixels = 0;
		    uchar* pixels = (uchar*)c->data;
		    int base_pixel_size = g_pix_container_type_sizes[ c->type ];
		    int pixel_size = base_pixel_size;
		    bool alpha_ch = 0;
		    uchar* alpha_data = 0;
		    if( ( c->flags & PIX_CONTAINER_FLAG_USES_KEY ) || pix_vm_get_container_alpha( cnum, vm ) >= 0 )
		    {
			alpha_ch = 1;
			alpha_data = (uchar*)pix_vm_get_container_alpha_data( cnum, vm );
			pixel_size = 4;
			if( base_pixel_size == 2 )
			{
			    if( !( c->flags & PIX_CONTAINER_FLAG_GL_NICEST ) )
				pixel_size = 2;
			}
		    }

		    if( alpha_ch || size_changed )
		    {
			temp_pixels = (uchar*)MEM_NEW2( gl->xsize * gl->ysize * pixel_size );
			if( size_changed )
			{
			    //Make empty border:
			    if( size_changed & 1 )
			    {
				clear_fbox( temp_pixels, gl->xsize, pixel_size, c->xsize, 0, 1, c->ysize );
				if( c->xsize + 1 < gl->xsize )
				    clear_fbox( temp_pixels, gl->xsize, pixel_size, gl->xsize - 1, 0, 1, gl->ysize );
			    }
			    if( size_changed & 2 )
			    {
				clear_fbox( temp_pixels, gl->xsize, pixel_size, 0, c->ysize, c->xsize, 1 );
				if( c->ysize + 1 < gl->ysize )
				    clear_fbox( temp_pixels, gl->xsize, pixel_size, 0, gl->ysize - 1, gl->xsize, 1 );
			    }
			    if( size_changed == 3 )
				clear_fbox( temp_pixels, gl->xsize, pixel_size, c->xsize, c->ysize, 1, 1 );
			}
			for( int y = 0; y < c->ysize; y++ )
			{
			    switch( base_pixel_size )
			    {
				case 1:
				    mem_copy( temp_pixels + y * gl->xsize, pixels + y * c->xsize, c->xsize );
				    break;
				case 2:
				    {
					uint16* src = (uint16*)( pixels + y * c->xsize * 2 );
					uchar* alpha_src = 0;
					if( alpha_data )
					    alpha_src = alpha_data + y * c->xsize;
					switch( pixel_size )
					{
					    case 2:
						//From 16bit to 16bit:
						if( alpha_ch )
						{
						    uint16* dest = (uint16*)( temp_pixels + y * gl->xsize * 2 );
						    if( alpha_src )
							for( int x = 0; x < c->xsize; x++ )
							{
							    //to RGBA 4444
							    uint16 p = *src;
							    uchar r = ( ( p >> 11 ) << 3 ) & 0xF8; r >>= 4;
							    uchar g = ( ( p >> 5 ) << 2 ) & 0xFC; g >>= 4;
							    uchar b = ( p << 3 ) & 0xF8; b >>= 4;
							    if( ( c->flags & PIX_CONTAINER_FLAG_USES_KEY ) && (unsigned)c->key == p )
								*dest = ( r << 12 ) | ( g << 8 ) | ( b << 4 );
							    else
							    {
							        *dest = ( r << 12 ) | ( g << 8 ) | ( b << 4 ) | ( *alpha_src >> 4 );
							    }
							    alpha_src++;
							    src++;
							    dest++;
							}
						    else
							for( int x = 0; x < c->xsize; x++ )
							{
							    //to RGBA 5551
							    uint16 p = *src;
							    uchar r = ( ( p >> 11 ) << 3 ) & 0xF8; r >>= 3;
							    uchar g = ( ( p >> 5 ) << 2 ) & 0xFC; g >>= 3;
							    uchar b = ( p << 3 ) & 0xF8; b >>= 3;
							    if( ( c->flags & PIX_CONTAINER_FLAG_USES_KEY ) && (unsigned)c->key == p )
								*dest = ( r << 11 ) | ( g << 6 ) | ( b << 1 );
							    else
							    {
							        *dest = ( r << 11 ) | ( g << 6 ) | ( b << 1 ) | 1;
							    }
							    alpha_src++;
							    src++;
							    dest++;
							}
						}
						else
						{
						    mem_copy( temp_pixels + y * gl->xsize * 2, src, c->xsize * 2 );
						}
						break;
					    case 4:
						//From 16bit to 32bit:
						{
						    uint* dest = (uint*)( temp_pixels + y * gl->xsize * 4 );
						    if( alpha_src )
							for( int x = 0; x < c->xsize; x++ )
							{
							    uint16 p = *src;
							    uchar r = ( ( p >> 11 ) << 3 ) & 0xF8; if( r ) r |= 7;
							    uchar g = ( ( p >> 5 ) << 2 ) & 0xFC; if( g ) g |= 3;
							    uchar b = ( p << 3 ) & 0xF8; if( b ) b |= 7;
							    if( ( c->flags & PIX_CONTAINER_FLAG_USES_KEY ) && (unsigned)c->key == p )
								*dest = r | ( g << 8 ) | ( b << 16 );
							    else
							    {
							        *dest = r | ( g << 8 ) | ( b << 16 ) | ( *alpha_src << 24 );
							    }
							    alpha_src++;
							    src++;
							    dest++;
							}
						    else
							for( int x = 0; x < c->xsize; x++ )
							{
							    uint16 p = *src;
							    uchar r = ( ( p >> 11 ) << 3 ) & 0xF8; if( r ) r |= 7;
							    uchar g = ( ( p >> 5 ) << 2 ) & 0xFC; if( g ) g |= 3;
							    uchar b = ( p << 3 ) & 0xF8; if( b ) b |= 7;
							    if( ( c->flags & PIX_CONTAINER_FLAG_USES_KEY ) && (unsigned)c->key == p )
								*dest = r | ( g << 8 ) | ( b << 16 );
							    else
							        *dest = r | ( g << 8 ) | ( b << 16 ) | ( 255 << 24 );
							    src++;
							    dest++;
							}
						}
						break;
					    default:
						break;
					}
				    }
				    break;
				case 4:
				    {
					uint* src = (uint*)( pixels + y * c->xsize * 4 );
					switch( pixel_size )
					{
					    case 4:
						//From 32bit to 32bit:
						if( ( c->flags & PIX_CONTAINER_FLAG_USES_KEY ) || alpha_ch )
						{
						    uchar* alpha_src = 0;
						    if( alpha_data )
							alpha_src = alpha_data + y * c->xsize;
						    uint* dest = (uint*)( temp_pixels + y * gl->xsize * 4 );
						    if( alpha_src )
							for( int x = 0; x < c->xsize; x++ )
							{
							    uint p = *src;
							    if( ( c->flags & PIX_CONTAINER_FLAG_USES_KEY ) && (unsigned)c->key == p )
								*dest = p & 0x00FFFFFF;
							    else
							    {
							        *dest = ( p & 0x00FFFFFF ) | ( *alpha_src << 24 );
							    }
							    alpha_src++;
							    src++;
							    dest++;
							}
						    else
							for( int x = 0; x < c->xsize; x++ )
							{
							    uint p = *src;
							    if( ( c->flags & PIX_CONTAINER_FLAG_USES_KEY ) && (unsigned)c->key == p )
								*dest = p & 0x00FFFFFF;
							    else
							        *dest = ( p & 0x00FFFFFF ) | ( 255 << 24 );
							    src++;
							    dest++;
							}
						}
						else
						{
						    mem_copy( temp_pixels + y * gl->xsize * 4, src, c->xsize * 4 );
						}
						break;
					    default:
						break;
					}
				    }
				    break;
				default:
				    break;
			    }
			}
			pixels = temp_pixels;
		    }
		    
		    gl_lock();
		    
		    glGenTextures( 1, &gl->texture_id );
		    glEnable( GL_TEXTURE_2D );
    		    glBindTexture( GL_TEXTURE_2D, gl->texture_id );

		    int internal_format;
		    int format;
		    int type;
		    bool no_texture = 0;
		    switch( pixel_size )
		    {
			case 1:
			    internal_format = GL_ALPHA;
			    format = GL_ALPHA;
			    type = GL_UNSIGNED_BYTE;
			    break;
			case 2:
			    internal_format = GL_RGB;
			    format = GL_RGB;
			    type = GL_UNSIGNED_SHORT_5_6_5;
			    if( alpha_ch )
			    {
				internal_format = GL_RGBA;
				format = GL_RGBA;
				if( alpha_data )
				    type = GL_UNSIGNED_SHORT_4_4_4_4;
				else
				    type = GL_UNSIGNED_SHORT_5_5_5_1;
			    }
			    else
			    {
				internal_format = GL_RGB;
				format = GL_RGB;
				type = GL_UNSIGNED_SHORT_5_6_5;
			    }
			    break;
			case 4:
			    if( alpha_ch )
				internal_format = GL_RGBA;
			    else
				internal_format = GL_RGB;			    
			    format = GL_RGBA;
			    type = GL_UNSIGNED_BYTE;
			    break;
			default:
			    no_texture = 1;
			    break;
		    }
		    if( no_texture == 0 )
		    {
			glTexImage2D(
            		    GL_TEXTURE_2D,
            		    0,
		    	    internal_format,
            		    gl->xsize, gl->ysize,
		    	    0,
            		    format,
		            type,
            		    pixels );
            		rv = 0;

    			if( c->flags & PIX_CONTAINER_FLAG_GL_MIN_LINEAR )
    			    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    			else
    			    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    			if( c->flags & PIX_CONTAINER_FLAG_GL_MAG_LINEAR )
        		    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        		else
        		    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );    
    			if( c->flags & PIX_CONTAINER_FLAG_GL_NO_XREPEAT )
        		    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    			if( c->flags & PIX_CONTAINER_FLAG_GL_NO_YREPEAT )
        		    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
            	    }
		    
		    glDisable( GL_TEXTURE_2D );
		    gl_unlock();
		    
		    mem_free( temp_pixels );
		}
	    }
	}
    }
    
    return rv;
}

void pix_vm_remove_container_gl_data( PIX_CID cnum, pix_vm* vm )
{
    if( vm->c == 0 ) return;

    if( (unsigned)cnum < (unsigned)PIX_VM_CONTAINERS )
    {
        pix_vm_container* c = vm->c[ cnum ];
        if( c && c->opt_data && c->opt_data->gl )
        {
	    sundog_mutex_lock( &vm->gl_mutex );
	    if( vm->gl_unused_textures == 0 )
	    {
		vm->gl_unused_textures = (uint*)MEM_NEW2( 32 * sizeof( uint ) );
		vm->gl_unused_textures_count = 0;
	    }
	    vm->gl_unused_textures_count++;
	    if( vm->gl_unused_textures_count > mem_get_size( vm->gl_unused_textures ) / sizeof( uint ) )
	    {
		vm->gl_unused_textures = (uint*)MEM_NEW2( mem_get_size( vm->gl_unused_textures ) + 32 * sizeof( uint ) );
	    }
	    vm->gl_unused_textures[ vm->gl_unused_textures_count - 1 ] = c->opt_data->gl->texture_id;
	    sundog_mutex_unlock( &vm->gl_mutex );
	    
            mem_free( c->opt_data->gl );
            c->opt_data->gl = 0;
        }
    }
}

//Call this function only in the thread with OpenGL context!
void pix_vm_empty_gl_trash( pix_vm* vm )
{
    if( vm->gl_unused_textures_count > 0 )
    {
	sundog_mutex_lock( &vm->gl_mutex );
	if( vm->gl_unused_textures )
	{
	    gl_lock();
	    glDeleteTextures( vm->gl_unused_textures_count, vm->gl_unused_textures );	    
	    gl_unlock();
	    vm->gl_unused_textures_count = 0;
	}
	sundog_mutex_unlock( &vm->gl_mutex );
    }
}

#endif
