#ifndef __V3NUS_FS__
#define __V3NUS_FS__

//V3NUS file system. Device dependent.
//Examples of file names for different systems:
//PalmOS:
// A:/prog.txt		- prog.txt database on the internal RAM
// B:/prog.txt		- prog.txt file on the external FLASH card
//Linux:
// /prog.txt		- prog.txt file in the root directory 
//Windows:
// C:/prog.txt		- prog.txt on disk C
//Any OS:
// 0:/prog.txt		- prog.txt file on virtual disk 0. This virtual disk is the TAR file. 
//			  Example of using:
//			  g_virt_disk0 = v3_open( "arhive.tar", "rb" );
//			  v3_open( "0:/some_file.txt", "rb" );
//			  ...
// 1:/prog.txt		- ... in the current working directory (iOS:docs; WinCE:/; Android:external card; ...)
// 2:/prog.txt		- ... in the user directory for configs and caches (Linux: /home/username/; Android: internal files directory)
// 3:/prog.txt		- ... in the temp directory

#if defined(WIN) || defined(WINCE)
    #include <windows.h>
    #include <shlobj.h>
#endif

#ifdef UNIX
    #include <dirent.h> //for file find
#endif

#define MAX_DISKS	    16
#define DISKNAME_SIZE	    4
#ifdef PALMOS
    #define MAX_DIR_LEN	    1024
#else
    #define MAX_DIR_LEN	    2048
#endif
extern int disks; //number of disks available for user
void get_disks( void ); //get info about local disks
UTF8_CHAR* get_disk_name( int number ); //get disk name
int get_current_disk( void ); //get number of the current disk
UTF8_CHAR* get_current_path( void ); //get current working path (example: "C:/mydir/")
UTF8_CHAR* get_user_path( void ); //get home user path
UTF8_CHAR* get_temp_path( void );

//***********************************************************
//PalmOS functions: *****************************************
//***********************************************************

#ifdef PALMOS

int open_vfs( void ); //Return 1 if successful. vfs_volume_number = number of first finded VFS volume (flash-card)
int isBusy( DmOpenRef db, int index ); //Is record busy?
void recordsBusyReset( int f );
void get_records_info( DmOpenRef db, uint16 f_num );
void vfs2buffer( uint f );
void buffer2vfs( uint f );
int next_record( uint f );
uint get_record_rest( FILE *fpp );

#endif

//***********************************************************
//***********************************************************
//***********************************************************

//***********************************************************
//Main multiplatform functions:******************************
//***********************************************************

#ifdef PALMOS
    #define MAX_V3_DESCRIPTORS	    32
    #define V3_FOPEN_MAX	    32
#else
#if CPUMARK < 10
    #define MAX_V3_DESCRIPTORS	    64
#else
    #define MAX_V3_DESCRIPTORS	    256
#endif
    #define V3_FOPEN_MAX	    ( MAX_V3_DESCRIPTORS - 3 )
#endif

//Std files:
#define V3_STDIN		    ( MAX_V3_DESCRIPTORS - 0 )
#define V3_STDOUT		    ( MAX_V3_DESCRIPTORS - 1 )
#define V3_STDERR		    ( MAX_V3_DESCRIPTORS - 2 )

//V3_FILE types:
#define V3_FILE_STD		    0
#define V3_FILE_ON_VIRTUAL_DISK	    1
#define V3_FILE_IN_MEMORY	    2

struct V3_FILE_STRUCT
{
    UTF8_CHAR*	    filename;
    void*	    f;
    int		    type;
    char*	    virt_file_data;
    size_t	    virt_file_ptr;
    size_t	    virt_file_size;
};

#define V3_FILE uint

extern V3_FILE g_virt_disk0;

void v3_init( void );
void v3_deinit( void );
UTF8_CHAR* v3_make_filename( const UTF8_CHAR* filename );
const UTF8_CHAR* v3_get_filename_without_dir( const UTF8_CHAR* filename );
void* v3_get_data( V3_FILE f );
V3_FILE v3_open_in_memory( void* data, size_t size );
V3_FILE v3_open( const UTF8_CHAR* filename, const UTF8_CHAR* filemode );
int v3_close( V3_FILE f );
void v3_rewind( V3_FILE f );
int v3_getc( V3_FILE f );
size_t v3_tell( V3_FILE f );
int v3_seek( V3_FILE f, size_t offset, int access );
int v3_eof( V3_FILE f );
int v3_flush( V3_FILE f );
size_t v3_read( void* ptr, size_t el_size, size_t elements, V3_FILE f );
size_t v3_write( const void* ptr, size_t el_size, size_t elements, V3_FILE f );
int v3_putc( int val, V3_FILE f );
int v3_remove( const UTF8_CHAR* filename );
int v3_rename( const UTF8_CHAR* old_name, const UTF8_CHAR* new_name );
int v3_mkdir( const UTF8_CHAR* pathname, uint mode );
size_t v3_get_file_size( const UTF8_CHAR* filename );
int v3_copy_file( const UTF8_CHAR* dest, const UTF8_CHAR* src );

//***********************************************************
//***********************************************************
//***********************************************************

//FIND FILE FUNCTIONS:

//type in find_struct:
enum 
{
    TYPE_FILE = 0,
    TYPE_DIR
};

struct find_struct
{ //structure for file searching functions
    const UTF8_CHAR* start_dir; //Example: "c:/mydir/" "d:/"
    const UTF8_CHAR* mask; //Example: "xm/mod/it" (or NULL for all files)
    bool start_dir_malloc;
    
    UTF8_CHAR name[ MAX_DIR_LEN ]; //Found file's name
    UTF8_CHAR temp_name[ MAX_DIR_LEN ];
    char type; //Found file's type: 0 - file; 1 - dir

#if defined(PALMOS)
    uint16 card_id;
    LocalID db_id;
    DmSearchStateType search_info;
    FileRef dir_ref; //VFS: reference to the start_dir
    uint dir_iterator; //VFS: dir iterator
    FileInfoType file_info; //VFS: file info
#endif
#if defined(WIN)
    WIN32_FIND_DATAW find_data;
#endif
#if defined(WINCE)
    WIN32_FIND_DATA find_data;
#endif
#if defined(WIN) | defined(WINCE)
    HANDLE find_handle;
    UTF8_CHAR win_mask[ MAX_DIR_LEN ]; //Example: "*.xm *.mod *.it"
    UTF8_CHAR win_start_dir[ MAX_DIR_LEN ]; //Example: "mydir\*.xm"
#endif
#ifdef UNIX
    DIR* dir;
    struct dirent* current_file;
    UTF8_CHAR new_start_dir[ MAX_DIR_LEN ];
#endif
};

int find_first( find_struct* ); //Return values: 0 - no files
int find_next( find_struct* ); //Return values: 0 - no files
void find_close( find_struct* );

#endif //__V3NUS_FS__
