/*
    SunDog: sound.cpp. Sound and MIDI in/out
    This file is part of the SunDog engine.
    Copyright (C) 2002 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

//#define OUTPUT_TO_FILE		1	/*output file: audio.raw*/

#include "../../core/core.h"
#include "../../core/debug.h"
#include "../../filesystem/v3nus_fs.h"
#include "../../utils/utils.h"
#include "../../memory/memory.h"
#include "../sound.h"

#ifdef OUTPUT_TO_FILE
FILE* g_audio_output = 0;
#endif

sound_struct g_snd; //Main sound structure;
bool g_snd_initialized = 0;

#ifndef NOMIDI
ticks_hr_t g_last_midi_in_activity = 0;
ticks_hr_t g_last_midi_out_activity = 0;
sundog_mutex g_sundog_midi_mutex;
int g_sundog_midi_open_counter = 0;
#endif

//#define SHOW_DEBUG_MESSAGES

#ifndef NOSOUND

#include "sound_common.h"

#ifdef LINUX
#ifdef ANDROID
    #include "sound_android.h"
#else
    #include "sound_linux.h"
    #include "sound_linux_alsa_midi.h"
#endif
#endif

#ifdef PALMOS
    #include "sound_palmos.h"
#endif

#ifdef OSX
    #include "sound_osx.h"
    #include "sound_coremidi.h"
#endif

#ifdef IPHONE
    #include "sound_ios.h"
    #include "sound_coremidi.h"
#endif

#if defined(WIN) || defined(WINCE)
    #include "sound_win.h"
    #include "sound_win_mmsystem_midi.h"
#endif

#endif

//
// Sound Stream
//

void user_controlled_audio_callback( void* buffer, int frames, int latency, ticks_hr_t output_time )
{
    g_snd.out_buffer = buffer;
    g_snd.out_frames = frames;
    g_snd.out_time = output_time;
    g_snd.latency = latency;
    main_sound_output_callback( &g_snd, 0 );
}

int sound_stream_init( int mode, int freq, int channels, int flags )
{
#ifndef NOSOUND

    if( mode < 0 )
    {
	//Default mode:
	mode = SOUND_MODE_INT16;
#ifdef OSX
	mode = SOUND_MODE_FLOAT32;
#endif
    }
    if( freq < 0 )
    {
	//Default frequency:
	freq = profile_get_int_value( KEY_FREQ, 44100, 0 );
    }
    if( channels < 0 )
    {
	//Default number of channels:
	channels = 2;
    }
    if( freq < 44100 )
    {
	dprint( "WARNING. Wrong sampling frequency %d (must be >= 44100). Using 44100.\n", freq );
	freq = 44100;
    }
#ifdef ONLY44100
    if( freq != 44100 ) 
    {
	dprint( "WARNING. Sampling frequency must be 44100 for this device. Using 44100.\n" );
	freq = 44100;
    }
#endif
    
#ifdef OUTPUT_TO_FILE
    g_audio_output = fopen( "audio.raw", "wb" );
#endif
    
    mem_set( &g_snd, sizeof( g_snd ), 0 );
    g_snd.mode = mode;
    g_snd.flags = flags;
    g_snd.freq = freq;
    g_snd.channels = channels;

    sundog_mutex_init( &g_snd.mutex, 0 );
    sundog_mutex_init( &g_snd.in_mutex, 0 );

    int res;

    if( flags & SOUND_STREAM_FLAG_USER_CONTROLLED )
    {
    }
    else
    {
	res = device_sound_stream_init();
	if( res )
	    return res;
    }

    int frame = 1;
    if( mode == SOUND_MODE_INT16 ) frame = 2 * channels;
    if( mode == SOUND_MODE_FLOAT32 ) frame = 4 * channels;
#ifdef PALMOS
    g_snd.slot_buffer_size = 1024 * 4;
#else
    g_snd.slot_buffer_size = 1024 * 8;
#endif
    g_snd.slot_buffer = MEM_NEW( HEAP_DYNAMIC, g_snd.slot_buffer_size * frame );
    
    g_snd_initialized = 1;

#endif    
    
    return 0;
}

void sound_stream_set_slot_callback( int slot, void* callback, void* user_data )
{
#ifndef NOSOUND
    
#ifdef SHOW_DEBUG_MESSAGES
    dprint( "SET SOUND SLOT CALLBACK %d\n", slot );
#endif
    if( g_snd_initialized == 0 ) return;
    g_snd.slots[ slot ].callback = callback;
    g_snd.slots[ slot ].user_data = user_data;
    
#endif
}

void sound_stream_remove_slot_callback( int slot )
{
#ifndef NOSOUND

#ifdef SHOW_DEBUG_MESSAGES
    dprint( "REMOVE SOUND SLOT CALLBACK %d\n", slot );
#endif
    if( g_snd_initialized == 0 ) return;
    sound_stream_stop( slot );
    g_snd.slots[ slot ].callback = 0;
    
#endif
}

void sound_stream_create_scope_buffer( int slot, int frames, int channels, int flags )
{
#ifndef NOSOUND 
    
    if( g_snd_initialized == 0 ) return;
    if( sundog_mutex_lock( &g_snd.mutex ) ) return;
    
    if( channels <= 0 )
	channels = g_snd.channels;
    g_snd.slots[ slot ].scope_buffer = (SOUND_SCOPE_TYPE**)MEM_NEW( HEAP_DYNAMIC, channels * sizeof( SOUND_SCOPE_TYPE* ) );
    for( int c = 0; c < channels; c++ )
    {
        g_snd.slots[ slot ].scope_buffer[ c ] = (SOUND_SCOPE_TYPE*)MEM_NEW( HEAP_DYNAMIC, frames * sizeof( SOUND_SCOPE_TYPE ) );
    }
    g_snd.slots[ slot ].scope_buffer_frames = frames;
    g_snd.slots[ slot ].scope_buffer_channels = channels;
    g_snd.slots[ slot ].scope_buffer_flags = flags;

    sundog_mutex_unlock( &g_snd.mutex );

#endif
}

void sound_stream_remove_scope_buffer( int slot )
{
#ifndef NOSOUND 
    
    if( g_snd_initialized == 0 ) return;
    if( sundog_mutex_lock( &g_snd.mutex ) ) return;
    
    if( g_snd.slots[ slot ].scope_buffer )
    {
	for( int c = 0; c < g_snd.slots[ slot ].scope_buffer_channels; c++ )
	{
	    if( g_snd.slots[ slot ].scope_buffer[ c ] )
		mem_free( g_snd.slots[ slot ].scope_buffer[ c ] );
	}
	mem_free( g_snd.slots[ slot ].scope_buffer );
	g_snd.slots[ slot ].scope_buffer = 0;
    }
    
    sundog_mutex_unlock( &g_snd.mutex );
    
#endif
}

SOUND_SCOPE_TYPE* sound_stream_get_scope_buffer( int slot, int channel )
{
    if( g_snd_initialized == 0 ) return 0;
    return g_snd.slots[ slot ].scope_buffer[ channel ];
}

bool sound_stream_is_scope_buffer_locked( int slot )
{
#ifndef NOSOUND     

    if( g_snd_initialized == 0 ) return 0;
    return g_snd.slots[ slot ].scope_buffer_locked;
    
#else
    
    return 0;
    
#endif
}

void sound_stream_scope_buffer_unlock( int slot )
{
#ifndef NOSOUND     
    
    if( g_snd_initialized == 0 ) return;
    g_snd.slots[ slot ].scope_buffer_locked = 0;

#endif
}

void sound_stream_lock( int slot )
{
#ifndef NOSOUND     

#ifdef SHOW_DEBUG_MESSAGES
    dprint( "LOCK %d\n", slot );
#endif
    if( g_snd_initialized == 0 ) return;
    sundog_mutex_lock( &g_snd.mutex );
#ifdef SHOW_DEBUG_MESSAGES
    dprint( "LOCK OK %d\n", slot );
#endif

#endif
}

void sound_stream_unlock( int slot )
{
#ifndef NOSOUND 
    
#ifdef SHOW_DEBUG_MESSAGES
    dprint( "UNLOCK %d\n", slot );
#endif
    if( g_snd_initialized == 0 ) return;
    sundog_mutex_unlock( &g_snd.mutex );
#ifdef SHOW_DEBUG_MESSAGES
    dprint( "UNLOCK OK %d\n", slot );
#endif

#endif
}

void sound_stream_play( int slot )
{
#ifndef NOSOUND 
    
#ifdef SHOW_DEBUG_MESSAGES
    dprint( "PLAY %d\n", slot );
#endif
    if( g_snd_initialized == 0 ) return;
    if( g_snd.slots[ slot ].callback )
    {
	if( g_snd.slots[ slot ].status != 1 )
	{
	    if( g_snd.flags & SOUND_STREAM_FLAG_ONE_THREAD )
		g_snd.slots[ slot ].status = 1;
	    else
	    {
		volatile int event = g_snd.slots[ slot ].event;
		event++;
		event &= 0x0FFFF;
		event += 0x10000;
		g_snd.slots[ slot ].event = event;
		while( g_snd.slots[ slot ].event_answer != event ) { HANDLE_THREAD_EVENTS; time_sleep( 5 ); }
	    }
	}
    }
#ifdef SHOW_DEBUG_MESSAGES
    dprint( "PLAY OK %d\n", slot );
#endif
    
#endif
}

void sound_stream_stop( int slot )
{
#ifndef NOSOUND 
    
#ifdef SHOW_DEBUG_MESSAGES
    dprint( "STOP %d\n", slot );
#endif
    if( g_snd_initialized == 0 ) return;
    if( g_snd.slots[ slot ].callback )
    {
	if( g_snd.slots[ slot ].status != 0 )
	{
	    if( g_snd.flags & SOUND_STREAM_FLAG_ONE_THREAD )
		g_snd.slots[ slot ].status = 0;
	    else
	    {
		volatile int event = g_snd.slots[ slot ].event;
		event++;
		event &= 0x0FFFF;
		event += 0x20000;
		g_snd.slots[ slot ].event = event;
		while( g_snd.slots[ slot ].event_answer != event ) { HANDLE_THREAD_EVENTS; time_sleep( 5 ); }
	    }
	}
    }
#ifdef SHOW_DEBUG_MESSAGES
    dprint( "STOP OK %d\n", slot );
#endif
    
#endif
}

//Call this function in the main app thread, where the sound stream is not locked.
//Or use sound_stream_input_request()
void sound_stream_input( bool enable )
{
#ifndef NOSOUND
    
    if( g_snd_initialized == 0 ) return;
    if( enable ) 
	g_snd.in_enabled++;
    else
	g_snd.in_enabled--;
    if( g_snd.flags & SOUND_STREAM_FLAG_USER_CONTROLLED )
    {
    }
    else
    {
	if( g_snd.in_enabled == 0 )
	    device_sound_stream_input( 0 );
	if( enable && g_snd.in_enabled == 1 )
	    device_sound_stream_input( 1 );
    }
        
#endif    
}

void sound_stream_input_request( bool enable )
{
#ifndef NOSOUND
    
    if( g_snd_initialized == 0 ) return;
    if( enable ) 
	g_snd.in_request++;
    else
	g_snd.in_request--;
        
#endif    
}

void sound_stream_handle_input_requests( void )
{
#ifndef NOSOUND
    
    if( g_snd_initialized == 0 ) return;
    if( g_snd.in_request_answer < g_snd.in_request )
    {
	if( g_snd.in_request_answer == 0 )
	    sound_stream_input( 1 );
    }
    if( g_snd.in_request_answer > g_snd.in_request )
    {
	if( g_snd.in_request_answer > 0 && g_snd.in_request <= 0 )
	    sound_stream_input( 0 );
    }
    g_snd.in_request_answer = g_snd.in_request;
        
#endif    
}

const UTF8_CHAR* sound_stream_get_driver_name( void )
{
#ifndef NOSOUND
    
    if( g_snd_initialized == 0 ) return "";
    if( g_snd.flags & SOUND_STREAM_FLAG_USER_CONTROLLED )
	return "unknown";
    else
	return device_sound_stream_get_driver_name();
    
#else

    return "nosound";
    
#endif
}

const UTF8_CHAR* sound_stream_get_driver_info( void )
{
#ifndef NOSOUND
    
    if( g_snd_initialized == 0 ) return "";
    if( g_snd.flags & SOUND_STREAM_FLAG_USER_CONTROLLED )
	return "Unknown (User Controlled)";
    else
	return device_sound_stream_get_driver_info();
    
#else

    return "No sound";
    
#endif
}

int sound_stream_get_drivers( UTF8_CHAR*** names, UTF8_CHAR*** infos )
{
#ifndef NOSOUND

    return device_sound_stream_get_drivers( names, infos );

#else

    return 0;
    
#endif
}

int sound_stream_get_devices( const UTF8_CHAR* driver, UTF8_CHAR*** names, UTF8_CHAR*** infos, bool input )
{
#ifndef NOSOUND

    return device_sound_stream_get_devices( driver, names, infos, input );

#else

    return 0;
    
#endif
}

void sound_stream_close( void )
{
#ifndef NOSOUND 
    
    if( g_snd_initialized == 0 ) return;
    dprint( "SOUND: sound_stream_close() begin\n" );

    if( g_snd.flags & SOUND_STREAM_FLAG_USER_CONTROLLED )
    {
    }
    else
    {
	device_sound_stream_close();
    }

    if( g_snd.slot_buffer )
	mem_free( g_snd.slot_buffer );
	
    for( int c = 0; c < SOUND_SLOTS; c++ )
    {
	sound_stream_remove_scope_buffer( c );
    }

#ifdef OUTPUT_TO_FILE
    if( g_audio_output ) fclose( g_audio_output );
#endif
    
    sundog_mutex_destroy( &g_snd.mutex );
    sundog_mutex_destroy( &g_snd.in_mutex );
    
    dprint( "SOUND: sound_stream_close() end\n" );
    g_snd_initialized = 0;

#endif
}

//
// MIDI
//

int sundog_midi_init( void )
{
    int rv = 0;

#ifndef NOMIDI
    if( g_sundog_midi_open_counter == 0 )
    {
	sundog_mutex_init( &g_sundog_midi_mutex, 0 );
    }
    g_sundog_midi_open_counter++;
#endif

    return rv;
}

int sundog_midi_deinit( void )
{
    int rv = 0;

#ifndef NOMIDI
    g_sundog_midi_open_counter--;
    if( g_sundog_midi_open_counter == 0 )
    {
	sundog_mutex_destroy( &g_sundog_midi_mutex );
    }
#endif

    return rv;    
}

int sundog_midi_client_open( sundog_midi_client* c, const UTF8_CHAR* name )
{
    int rv = 0;

#ifndef NOMIDI
    
    sundog_mutex_lock( &g_sundog_midi_mutex );
    
    mem_set( c, sizeof( sundog_midi_client ), 0 );
    c->name = (UTF8_CHAR*)MEM_NEW2( mem_strlen( name ) + 1 );
    c->name[ 0 ] = 0;
    mem_strcat( c->name, name );
    
    rv = device_midi_client_open( c, name );

    sundog_mutex_unlock( &g_sundog_midi_mutex );

#endif
    
    return rv;
}

int sundog_midi_client_close( sundog_midi_client* c )
{
    int rv = 0;

#ifndef NOMIDI
    
    sundog_mutex_lock( &g_sundog_midi_mutex );
    
    if( c->ports )
    {
	int ports_count = mem_get_size( c->ports ) / sizeof( sundog_midi_port );
	for( int i = 0; i < ports_count; i++ )
	{
	    sundog_midi_client_close_port( c, i );
	}
	mem_free( c->ports );
	c->ports = 0;
    }

    rv = device_midi_client_close( c );
    
    mem_free( c->name );
    c->name = 0;
    
    sundog_mutex_unlock( &g_sundog_midi_mutex );
    
#endif

    return rv;
}

int sundog_midi_client_get_devices( sundog_midi_client* c, UTF8_CHAR*** devices, int flags )
{
    int rv = 0;
    
#ifndef NOMIDI
    
    sundog_mutex_lock( &g_sundog_midi_mutex );
    
    rv = device_midi_client_get_devices( c, devices, flags );
    
    sundog_mutex_unlock( &g_sundog_midi_mutex );
    
#endif
    
    return rv;
}

int sundog_midi_client_open_port( sundog_midi_client* c, const UTF8_CHAR* port_name, const UTF8_CHAR* dev_name, int flags )
{
    int rv = -1;
    
#ifndef NOMIDI

    if( port_name == 0 ) return -1;
    if( dev_name == 0 ) return -1;
    
    sundog_mutex_lock( &g_sundog_midi_mutex );
    
    while( 1 )
    {
	if( c->ports == 0 )
	{
	    c->ports = (sundog_midi_port*)MEM_NEW2( sizeof( sundog_midi_port ) );
	    mem_zero( c->ports );
	    rv = 0;
	}
	else
	{
	    for( int i = 0; i < (int)( mem_get_size( c->ports ) / sizeof( sundog_midi_port ) ); i++ )
	    {
		if( c->ports[ i ].active == 0 )
		{
		    rv = i;
		    break;
		}
	    }
	    if( rv < 0 )
	    {
		rv = (int)( mem_get_size( c->ports ) / sizeof( sundog_midi_port ) );
		c->ports = (sundog_midi_port*)mem_resize( c->ports, mem_get_size( c->ports ) + sizeof( sundog_midi_port ) );
		mem_set( &c->ports[ rv ], sizeof( sundog_midi_port ), 0 );
	    }
	}
	
	sundog_midi_port* port = &c->ports[ rv ];
	
	port->flags = flags;
	
	if( device_midi_client_open_port( c, rv, port_name, dev_name, flags ) )
	{
	    port->flags |= MIDI_NO_DEVICE;
	    port->device_specific = 0;
	}
	else
	{
	    port->flags &= ~MIDI_NO_DEVICE;
	}
	
	mem_free( port->port_name );
	port->port_name = (UTF8_CHAR*)MEM_NEW2( mem_strlen( port_name ) + 1 );
	port->port_name[ 0 ] = 0;
	mem_strcat( port->port_name, port_name );

	mem_free( port->dev_name );
	port->dev_name = (UTF8_CHAR*)MEM_NEW2( mem_strlen( dev_name ) + 1 );
	port->dev_name[ 0 ] = 0;
	mem_strcat( port->dev_name, dev_name );
	
	port->active = 1;
	
	break;
    }
    
    sundog_mutex_unlock( &g_sundog_midi_mutex );
    
#endif
    
    return rv;
}

int sundog_midi_client_reopen_port( sundog_midi_client* c, int pnum )
{
    int rv = 0;
    
#ifndef NOMIDI
    
    sundog_mutex_lock( &g_sundog_midi_mutex );
    
    while( 1 )
    {
	if( c->ports == 0 )
	{
	    rv = -1;
	    break;
	}
	
	if( (unsigned)pnum >= mem_get_size( c->ports ) / sizeof( sundog_midi_port ) ) 
	{
	    rv = -1;
	    break;
	}
	
	sundog_midi_port* port = &c->ports[ pnum ];
	
	if( port->active == 0 )
	{
	    rv = -1;
	    break;
	}
	
	if( ( port->flags & MIDI_NO_DEVICE ) == 0 )
	    device_midi_client_close_port( c, pnum );
	if( device_midi_client_open_port( c, pnum, port->port_name, port->dev_name, port->flags ) )
	{
	    port->flags |= MIDI_NO_DEVICE;
	    port->device_specific = 0;
	}
	else
	{
	    port->flags &= ~MIDI_NO_DEVICE;
	}
	
	break;
    }
    
    sundog_mutex_unlock( &g_sundog_midi_mutex );
    
#endif
    
    return rv;
}

int sundog_midi_client_close_port( sundog_midi_client* c, int pnum )
{
    int rv = 0;
    
#ifndef NOMIDI
    
    sundog_mutex_lock( &g_sundog_midi_mutex );
    
    while( 1 )
    {
	if( c->ports == 0 )
	{
	    rv = -1;
	    break;
	}
	
	if( (unsigned)pnum >= mem_get_size( c->ports ) / sizeof( sundog_midi_port ) ) 
	{
	    rv = -1;
	    break;
	}
	
	sundog_midi_port* port = &c->ports[ pnum ];
	
	if( port->active == 0 )
	{
	    rv = -1;
	    break;
	}

	if( ( port->flags & MIDI_NO_DEVICE ) == 0 )
	    rv = device_midi_client_close_port( c, pnum );
	
	mem_free( port->port_name );
	mem_free( port->dev_name );
	port->port_name = 0;
	port->dev_name = 0;
	
	port->active = 0;
	
	break;
    }
    
    sundog_mutex_unlock( &g_sundog_midi_mutex );

#endif
    
    return rv;
}

sundog_midi_event* sundog_midi_client_get_event( sundog_midi_client* c, int pnum )
{
    sundog_midi_event* evt = 0;

#ifndef NOMIDI

    sundog_mutex_lock( &g_sundog_midi_mutex );
    
    if( ( c->ports != 0 ) && ( (unsigned)pnum < mem_get_size( c->ports ) / sizeof( sundog_midi_port ) ) )
    {
	evt = device_midi_client_get_event( c, pnum );
    }
    
    sundog_mutex_unlock( &g_sundog_midi_mutex );
    
#endif
    
    return evt;
}

int sundog_midi_client_next_event( sundog_midi_client* c, int pnum )
{
    int rv = 0;

#ifndef NOMIDI
    
    sundog_mutex_lock( &g_sundog_midi_mutex );
    
    if( ( c->ports != 0 ) && ( (unsigned)pnum < mem_get_size( c->ports ) / sizeof( sundog_midi_port ) ) )
    {
	rv = device_midi_client_next_event( c, pnum );
    }
    
    sundog_mutex_unlock( &g_sundog_midi_mutex );
    
#endif
    
    return rv;
}

int sundog_midi_client_send_event( sundog_midi_client* c, int pnum, sundog_midi_event* evt )
{
    int rv = 0;

#ifndef NOMIDI

    sundog_mutex_lock( &g_sundog_midi_mutex );
    
    while( 1 )
    {
	if( c->ports == 0 ) { rv = -1; break; }
	if( (unsigned)pnum >= mem_get_size( c->ports ) / sizeof( sundog_midi_port ) ) { rv = -1; break; }
	if( evt == 0 || evt->size == 0 || evt->data == 0 ) { rv = -1; break; }
	
	rv = device_midi_client_send_event( c, pnum, evt );
	
	break;
    }
    
    sundog_mutex_unlock( &g_sundog_midi_mutex );
    
#endif
    
    return rv;
}
