/*
    image.cpp. Functions for working with images
    This file is part of the SunDog engine.
    Copyright (C) 2002 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

#include "../../core/core.h"
#include "../../core/debug.h"
#include "../../utils/utils.h"
#include "../wmanager.h"

sundog_image* new_image( 
    int xsize, 
    int ysize, 
    void* src,
    int src_x,
    int src_y,
    int src_xsize, //whole source image size
    int src_ysize,
    int flags )
{
    sundog_image* img = 0;
    
    img = (sundog_image*)MEM_NEW2( sizeof( sundog_image ) );
    if( img )
    {
	img->xsize = xsize;
	img->ysize = ysize;
	img->flags = flags;
	img->backcolor = COLORMASK;
	img->color = COLORMASK;
	img->transp = 255;
#if defined(OPENGL) && !defined(FRAMEBUFFER)
	img->int_xsize = round_to_power_of_two( xsize );
	img->int_ysize = round_to_power_of_two( ysize );
#else
	img->int_xsize = xsize;
	img->int_ysize = ysize;
#endif
	if( flags & IMAGE_STATIC_SOURCE )
	{
	    img->data = src;
	}
	else
	{
	    if( src_xsize - src_x < xsize ) xsize -= ( xsize - ( src_xsize - src_x ) );
	    if( src_ysize - src_y < ysize ) ysize -= ( ysize - ( src_ysize - src_y ) );
	    if( flags & IMAGE_NATIVE_RGB )
	    {
		img->data = MEM_NEW( HEAP_DYNAMIC, img->int_xsize * img->int_ysize * COLORLEN );
		if( flags & IMAGE_CLEAN )
		    mem_zero( img->data );
		if( src )
		{
		    COLORPTR data = (COLORPTR)img->data;
	    	    COLORPTR psrc = (COLORPTR)src;
	    	    for( int y = 0; y < ysize; y++ )
	    	    {
			int ptr = y * img->int_xsize;
			int src_ptr = ( y + src_y ) * src_xsize + src_x;
			for( int x = 0; x < xsize; x++ )
			{
		    	    data[ ptr ] = psrc[ src_ptr ];
			    ptr++;
			    src_ptr++;
			}
		    }
		}
	    }
	    if( flags & IMAGE_ALPHA8 )
	    {
		img->data = MEM_NEW( HEAP_DYNAMIC, img->int_xsize * img->int_ysize );
		if( flags & IMAGE_CLEAN )
		    mem_zero( img->data );
		if( src )
		{
	    	    uchar* data = (uchar*)img->data;
	    	    uchar* csrc = (uchar*)src;
	    	    for( int y = 0; y < ysize; y++ )
	    	    {
			int ptr = y * img->int_xsize;
			int src_ptr = ( y + src_y ) * src_xsize + src_x;
	    		for( int x = 0; x < xsize; x++ )
			{
		    	    data[ ptr ] = csrc[ src_ptr ];
		    	    ptr++;
		    	    src_ptr++;
			}
		    }
		}
	    }
	}
#if defined(OPENGL) && !defined(FRAMEBUFFER)
	if( img && img->data )
	{
	    gl_lock();
	    
	    unsigned int texture_id;
	    glGenTextures( 1, &texture_id );
	    img->gl_texture_id = texture_id;
	    flush_image( img );
	    
	    gl_unlock();
	}
#endif
    }
    
    return img;
}

void flush_image( sundog_image* img )
{
#if defined(OPENGL) && !defined(FRAMEBUFFER)
    if( img && img->data )
    {
	gl_lock();
	
	void* data = img->data;
#if defined(OPENGL) && !defined(FRAMEBUFFER)
	void* temp_data = 0;
#endif
	
	if( img->flags & IMAGE_STATIC_SOURCE )
	{
#if defined(OPENGL) && !defined(FRAMEBUFFER)
	    if( img->int_xsize != img->xsize || img->int_ysize != img->ysize )
	    {
		int color_len = 1;
		if( img->flags & IMAGE_NATIVE_RGB ) color_len = COLORLEN;
		temp_data = MEM_NEW2( img->int_xsize * img->int_ysize * color_len );
		data = temp_data;
		uchar* src_ptr = (uchar*)img->data;
		for( int y = 0; y < img->ysize; y++ )
		{
		    uchar* dest_ptr = (uchar*)temp_data + y * img->int_xsize * color_len;
		    mem_copy( dest_ptr, src_ptr, img->xsize * color_len );
		    dest_ptr += img->xsize * color_len;
		    src_ptr += img->xsize * color_len;
		}
	    }
#endif
	}
	
        glEnable( GL_TEXTURE_2D );
        glBindTexture( GL_TEXTURE_2D, img->gl_texture_id );
	if( img->flags & IMAGE_INTERPOLATION )
	{
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	}
	else
	{
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	}
        if( img->flags & IMAGE_NATIVE_RGB )
        {
#ifdef COLOR32BITS
	    glTexImage2D(
    	        GL_TEXTURE_2D,
    	        0,
    	        GL_RGB,
    	        img->int_xsize, img->int_ysize,
    	        0,
    	        GL_RGBA,
    	        GL_UNSIGNED_BYTE,
    	        data );
#endif
#ifdef COLOR16BITS
	    glTexImage2D(
    	        GL_TEXTURE_2D,
    	        0,
    	        GL_RGB,
    	        img->int_xsize, img->int_ysize,
    	        0,
    	        GL_RGB,
    	        GL_UNSIGNED_SHORT_5_6_5,
    	        data );
#endif
	}
	else if( img->flags & IMAGE_ALPHA8 )
	{
	    glTexImage2D(
    	        GL_TEXTURE_2D,
    	        0,
    	        GL_ALPHA,
    	        img->int_xsize, img->int_ysize,
    	        0,
    	        GL_ALPHA,
    	        GL_UNSIGNED_BYTE,
    	        data );
	}
	glDisable( GL_TEXTURE_2D );

#if defined(OPENGL) && !defined(FRAMEBUFFER)
	mem_free( temp_data );
#endif
	
	gl_unlock();
    }
#endif
}

sundog_image* resize_image( int resize_flags, int new_xsize, int new_ysize, sundog_image* img )
{
    if( img && !( img->flags & IMAGE_STATIC_SOURCE ) && img->data )
    {
	sundog_image* new_img = new_image( new_xsize, new_ysize, img->data, 0, 0, img->int_xsize, img->int_ysize, img->flags );
	remove_image( img );
	return new_img;
    }
    return 0;
}

void remove_image( sundog_image* img )
{
    if( img )
    {
#if defined(OPENGL) && !defined(FRAMEBUFFER)
	gl_lock();
	glDeleteTextures( 1, &img->gl_texture_id );
	gl_unlock();
#endif
	if( img->data && !( img->flags & IMAGE_STATIC_SOURCE ) )
	{
	    mem_free( img->data );
	}
	mem_free( img );
    }
}

bool image_can_be_redrawn_safely( sundog_image* img )
{
    if( img )
    {
#if defined(OPENGL) && !defined(FRAMEBUFFER)
	return 1;
#endif
    }
    return 0;
}
