/*
    wm_opengl.h. Platform-dependent module : OpenGL functions
    This file is part of the SunDog engine.
    Copyright (C) 2002 - 2013 Alexander Zolotov <nightradio@gmail.com>
    www.warmplace.ru
*/

#ifndef __WINMANAGER_OPENGL__
#define __WINMANAGER_OPENGL__

#include <string.h>

//#################################
//## DEVICE DEPENDENT FUNCTIONS: ##
//#################################

void gl_lock( void )
{
#if defined(OSX)
    int err = CGLLockContext( (CGLContextObj)g_gl_context_obj );
    if( err != 0 )
    {
	dprint( "CGLLockContext() error %d\n", err );
    }
#else
    if( g_wm ) sundog_mutex_lock( &g_wm->gl_mutex );
#endif
}

void gl_unlock( void )
{
#if defined(OSX)
    int err = CGLUnlockContext( (CGLContextObj)g_gl_context_obj );
    if( err != 0 )
    {
	dprint( "CGLUnlockContext() error %d\n", err );
    }
#else
    if( g_wm ) sundog_mutex_unlock( &g_wm->gl_mutex );
#endif
}

void gl_init( window_manager* wm )
{
    gl_lock();
    
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();

    glClearColor( 0, 0, 0, 1 );
#ifdef OPENGLES
    glClearDepthf( 1 );
#else
    glClearDepth( 1 );
#endif
    glClearStencil( 0 );
    glDepthFunc( GL_LESS );
    glDisable( GL_DEPTH_TEST );
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

#if defined(FRAMEBUFFER)
    glGenTextures( 1, &wm->gl_fb_texture_id );
    wm->fb_xsize = round_to_power_of_two( wm->screen_xsize );
    wm->fb_ysize = round_to_power_of_two( wm->screen_ysize );
    glEnable( GL_TEXTURE_2D );
    glBindTexture( GL_TEXTURE_2D, wm->gl_fb_texture_id );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
#ifdef COLOR8BITS
    glTexImage2D( 
	GL_TEXTURE_2D,
        0,
        GL_LUMINANCE,
        wm->fb_xsize, wm->fb_ysize,
        0,
        GL_LUMINANCE,
        GL_UNSIGNED_BYTE,
        0 );
#endif
#ifdef COLOR16BITS
    glTexImage2D( 
	GL_TEXTURE_2D,
        0,
        GL_RGB,
	wm->fb_xsize, wm->fb_ysize,
        0,
	GL_RGB,
        GL_UNSIGNED_SHORT_5_6_5,
        0 );
#endif
#ifdef COLOR32BITS
    glTexImage2D( 
	GL_TEXTURE_2D,
        0,
        GL_RGBA,
        wm->fb_xsize, wm->fb_ysize,
        0,
        GL_RGBA,
        GL_UNSIGNED_BYTE,
        0 );
#endif
    int err = glGetError();
    if( err > 0 ) printf( "glTexImage2D() err: %x\n", err );
#endif

    wm->screen_xsize /= wm->screen_zoom;
    wm->screen_ysize /= wm->screen_zoom;

#ifdef WIN
    wm->wglSwapIntervalEXT = 0;
    const char* ext = (const char*)glGetString( GL_EXTENSIONS );
    if( ext )
    {
	if( mem_strstr( ext, "WGL_EXT_swap_control" ) )
	    wm->wglSwapIntervalEXT = (BOOL(WINAPI*)(int))wglGetProcAddress( "wglSwapIntervalEXT" );
    }
#endif
#if defined(LINUX) && !defined(OPENGLES)
    wm->glXSwapIntervalEXT = 0;
    wm->glXSwapIntervalSGI = 0;
    wm->glXSwapIntervalMESA = 0;
    const char* ext = glXQueryExtensionsString( g_wm->dpy, 0 );
    if( ext )
    {
	if( mem_strstr( ext, "GLX_EXT_swap_control" ) )
	    wm->glXSwapIntervalEXT = (int(*)(Display*,GLXDrawable,int))glXGetProcAddress( (const GLubyte*)"glXSwapIntervalEXT" );
	if( mem_strstr( ext, "GLX_SGI_swap_control" ) )
	    wm->glXSwapIntervalSGI = (void(*)(GLint))glXGetProcAddress( (const GLubyte*)"glXSwapIntervalSGI" );
	if( mem_strstr( ext, "GLX_MESA_swap_control" ) )
	    wm->glXSwapIntervalMESA = (void(*)(GLint))glXGetProcAddress( (const GLubyte*)"glXSwapIntervalMESA" );
    }
#endif
    
    gl_unlock();
}

void gl_deinit( window_manager* wm )
{
    gl_lock();
    
#ifdef FRAMEBUFFER
    glDeleteTextures( 1, &wm->gl_fb_texture_id );
#endif
    
    gl_unlock();
}

void gl_resize( window_manager* wm )
{
    //Set viewport to cover the window:
    
    gl_lock();
    
    if( wm->screen_zoom > 1 )
    {
        wm->gl_xscale = wm->screen_zoom;
        wm->gl_yscale = wm->screen_zoom;
    }
    else
    {
        wm->gl_xscale = 1;
        wm->gl_yscale = 1;
    }

    glViewport( 0, 0, wm->real_window_width, wm->real_window_height );    
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
#ifdef FRAMEBUFFER
#ifdef OPENGLES
    glOrthof( -1, 1, -1, 1, -100000, 100000 );
#else
    glOrtho( -1, 1, -1, 1, -100000, 100000 );
#endif
#else //FRAMEBUFFER
#ifdef OPENGLES
    glOrthof( 0, wm->real_window_width, wm->real_window_height, 0, -100000, 100000 );
#else
    glOrtho( 0, wm->real_window_width, wm->real_window_height, 0, -100000, 100000 );
#endif
    switch( wm->screen_angle )
    {
	case 0:
	    break;
	case 1:
	    glTranslatef( 0, wm->real_window_height, 0 );
	    glRotatef( -90, 0, 0, 1 );
	    break;
	case 2:
	    glTranslatef( wm->real_window_width, wm->real_window_height, 0 );
	    glRotatef( -90 * 2, 0, 0, 1 );
	    break;
	case 3:
	    glTranslatef( wm->real_window_width, 0, 0 );
	    glRotatef( -90 * 3, 0, 0, 1 );
	    break;
    }
    glTranslatef( 0.375, 0.375, 0 );
    if( wm->gl_xscale != 1 && wm->gl_yscale != 1 )
    {
	float xs, ys;
	xs = wm->gl_xscale;
	ys = wm->gl_yscale;
	glScalef( xs, ys, 1 );
    }
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
#endif

    glMatrixMode( GL_MODELVIEW );
    
    gl_unlock();
}

#if defined(OPENGL) && !defined(FRAMEBUFFER)

void device_draw_line( int x1, int y1, int x2, int y2, COLOR color )
{
    int transp = g_wm->cur_transparency;
    if( transp > 255 ) transp = 255;
    glColor4ub( red( color ), green( color ), blue( color ), (uchar)transp );
    int16 v[ 4 ];
    v[ 0 ] = (int16)x1; v[ 1 ] = (int16)y1;
    v[ 2 ] = (int16)x2; v[ 3 ] = (int16)y2;
    glEnableClientState( GL_VERTEX_ARRAY );
    glVertexPointer( 2, GL_SHORT, 0, v );
    glDrawArrays( GL_LINES, 0, 2 );
    glDrawArrays( GL_POINTS, 1, 1 );
}

void device_draw_line_with_transp( int x1, int y1, int t1, int x2, int y2, int t2, COLOR color )
{
    int transp = g_wm->cur_transparency;
    if( transp > 255 ) transp = 255;
    t1 *= transp;
    t1 /= 256;
    if( t1 > 255 ) t1 = 255;
    t2 *= transp;
    t2 /= 256;
    if( t2 > 255 ) t2 = 255;
    int16 v[ 4 ];
    uchar c[ 4 * 2 ];
    v[ 0 ] = (int16)x1; v[ 1 ] = (int16)y1;
    v[ 2 ] = (int16)x2; v[ 3 ] = (int16)y2;
    c[ 0 ] = red( color );
    c[ 1 ] = green( color );
    c[ 2 ] = blue( color );
    c[ 3 ] = (uchar)t1;
    c[ 4 ] = c[ 0 ];
    c[ 5 ] = c[ 1 ];
    c[ 6 ] = c[ 2 ];
    c[ 7 ] = (uchar)t2;
    glEnableClientState( GL_VERTEX_ARRAY );
    glEnableClientState( GL_COLOR_ARRAY );
    glVertexPointer( 2, GL_SHORT, 0, v );
    glColorPointer( 4, GL_UNSIGNED_BYTE, 0, c );
    glDrawArrays( GL_LINES, 0, 2 );
    glDrawArrays( GL_POINTS, 1, 1 );
    glDisableClientState( GL_COLOR_ARRAY );
}

void device_draw_box( int x, int y, int xsize, int ysize, COLOR color )
{
    int transp = g_wm->cur_transparency; 
    if( transp > 255 ) transp = 255;
    glColor4ub( red( color ), green( color ), blue( color ), (uchar)transp );
    int16 v[ 8 ];
    v[ 0 ] = (int16)x; v[ 1 ] = (int16)y;
    v[ 2 ] = (int16)x + (int16)xsize; v[ 3 ] = (int16)y;
    v[ 4 ] = (int16)x; v[ 5 ] = (int16)y + (int16)ysize;
    v[ 6 ] = (int16)x + (int16)xsize; v[ 7 ] = (int16)y + (int16)ysize;
    glEnableClientState( GL_VERTEX_ARRAY );
    glVertexPointer( 2, GL_SHORT, 0, v );
    glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );
}

void device_draw_polygon( sundog_polygon* p )
{
    bool one_color = ( g_wm->cur_flags & WBD_FLAG_ONE_COLOR ) != 0;
    bool one_transp = ( g_wm->cur_flags & WBD_FLAG_ONE_TRANSP ) != 0;
    glColor4ub( 255, 255, 255, 255 );

    int16 v[ 2 * 3 ];
    uchar c[ 4 * 3 ];
    v[ 0 ] = p->v[ 0 ].x;
    v[ 1 ] = p->v[ 0 ].y;
    COLOR color = p->v[ 0 ].c;
    c[ 0 ] = red( color );
    c[ 1 ] = green( color );
    c[ 2 ] = blue( color );
    c[ 3 ] = p->v[ 0 ].t;
    glEnableClientState( GL_VERTEX_ARRAY );
    glEnableClientState( GL_COLOR_ARRAY );
    for( int i = 2; i < p->vnum; i++ )
    {
	v[ 2 ] = p->v[ i - 1 ].x;
	v[ 3 ] = p->v[ i - 1 ].y;
	v[ 4 ] = p->v[ i ].x;
	v[ 5 ] = p->v[ i ].y;
	if( one_color )
	{
	    c[ 4 ] = c[ 0 ];
	    c[ 5 ] = c[ 1 ];
	    c[ 6 ] = c[ 2 ];
	    c[ 8 ] = c[ 0 ];
	    c[ 9 ] = c[ 1 ];
	    c[ 10 ] = c[ 2 ];
	}
	else 
	{
	    color = p->v[ i - 1 ].c;
	    c[ 4 ] = red( color );
	    c[ 5 ] = green( color );
	    c[ 6 ] = blue( color );
	    color = p->v[ i ].c;
	    c[ 8 ] = red( color );
	    c[ 9 ] = green( color );
	    c[ 10 ] = blue( color );
	}
	if( one_transp )
	{
	    c[ 7 ] = c[ 3 ];
	    c[ 11 ] = c[ 3 ];
	}
	else 
	{
	    c[ 7 ] = p->v[ i - 1 ].t;
	    c[ 11 ] = p->v[ i ].t;
	}
	glVertexPointer( 2, GL_SHORT, 0, v );
	glColorPointer( 4, GL_UNSIGNED_BYTE, 0, c );
	glDrawArrays( GL_TRIANGLES, 0, 3 );
    }
    glDisableClientState( GL_COLOR_ARRAY );
}

void device_draw_bitmap( 
    int dest_x, int dest_y, 
    int dest_xs, int dest_ys,
    int src_x, int src_y,
    sundog_image* data )
{
    int16 v[ 8 ];
    v[ 0 ] = (int16)dest_x; v[ 1 ] = (int16)dest_y;
    v[ 2 ] = (int16)dest_x + (int16)dest_xs; v[ 3 ] = (int16)dest_y;
    v[ 4 ] = (int16)dest_x; v[ 5 ] = (int16)dest_y + (int16)dest_ys;
    v[ 6 ] = (int16)dest_x + (int16)dest_xs; v[ 7 ] = (int16)dest_y + (int16)dest_ys;

    float t[ 8 ];
    float tx = (float)src_x / (float)data->int_xsize;
    float ty = (float)src_y / (float)data->int_ysize;
    float tx2 = (float)( src_x + dest_xs ) / (float)data->int_xsize;
    float ty2 = (float)( src_y + dest_ys ) / (float)data->int_ysize;
    t[ 0 ] = tx; t[ 1 ] = ty;
    t[ 2 ] = tx2; t[ 3 ] = ty;
    t[ 4 ] = tx; t[ 5 ] = ty2;
    t[ 6 ] = tx2; t[ 7 ] = ty2;

    glEnable( GL_TEXTURE_2D );
    glEnableClientState( GL_VERTEX_ARRAY );
    glEnableClientState( GL_TEXTURE_COORD_ARRAY );

    glVertexPointer( 2, GL_SHORT, 0, v );
    glTexCoordPointer( 2, GL_FLOAT, 0, t );
    uchar transp = data->transp;
    glColor4ub( red( data->color ), green( data->color ), blue( data->color ), transp );
    glBindTexture( GL_TEXTURE_2D, data->gl_texture_id );

    glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );

    glDisableClientState( GL_TEXTURE_COORD_ARRAY );
    glDisable( GL_TEXTURE_2D );
}

void device_draw_bitmap_scaled( 
    int dest_x, int dest_y, 
    int dest_xs, int dest_ys,
    float src_x, float src_y,
    float src_xs, float src_ys,
    sundog_image* data )
{
    int16 v[ 8 ];
    v[ 0 ] = (int16)dest_x; v[ 1 ] = (int16)dest_y;
    v[ 2 ] = (int16)dest_x + (int16)dest_xs; v[ 3 ] = (int16)dest_y;
    v[ 4 ] = (int16)dest_x; v[ 5 ] = (int16)dest_y + (int16)dest_ys;
    v[ 6 ] = (int16)dest_x + (int16)dest_xs; v[ 7 ] = (int16)dest_y + (int16)dest_ys;

    float t[ 8 ];
    float tx = (float)src_x / (float)data->int_xsize;
    float ty = (float)src_y / (float)data->int_ysize;
    float tx2 = (float)( src_x + src_xs ) / (float)data->int_xsize;
    float ty2 = (float)( src_y + src_ys ) / (float)data->int_ysize;
    t[ 0 ] = tx; t[ 1 ] = ty;
    t[ 2 ] = tx2; t[ 3 ] = ty;
    t[ 4 ] = tx; t[ 5 ] = ty2;
    t[ 6 ] = tx2; t[ 7 ] = ty2;

    glEnable( GL_TEXTURE_2D );
    glEnableClientState( GL_VERTEX_ARRAY );
    glEnableClientState( GL_TEXTURE_COORD_ARRAY );

    glVertexPointer( 2, GL_SHORT, 0, v );
    glTexCoordPointer( 2, GL_FLOAT, 0, t );
    uchar transp = data->transp;
    glColor4ub( red( data->color ), green( data->color ), blue( data->color ), transp );
    glBindTexture( GL_TEXTURE_2D, data->gl_texture_id );

    glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );

    glDisableClientState( GL_TEXTURE_COORD_ARRAY );
    glDisable( GL_TEXTURE_2D );
}

void device_vsync( bool vsync )
{
#ifdef WIN
    if( g_wm->wglSwapIntervalEXT )
	g_wm->wglSwapIntervalEXT( vsync );
#endif
#ifdef X11
#ifndef OPENGLES
    while( 1 )
    {
	if( g_wm->glXSwapIntervalEXT )
	{
	    g_wm->glXSwapIntervalEXT( g_wm->dpy, g_wm->win, vsync );
	    break;
	}
	if( g_wm->glXSwapIntervalSGI )
	{
	    g_wm->glXSwapIntervalSGI( vsync );
	    break;
	}
	if( g_wm->glXSwapIntervalMESA )
	{
	    g_wm->glXSwapIntervalMESA( vsync );
	    break;
	}
	break;
    }
#endif
#endif    
#ifdef OSX
    GLint p = vsync;
    CGLSetParameter( CGLGetCurrentContext(), kCGLCPSwapInterval, &p );
#endif
}

void device_redraw_framebuffer( void ) 
{
    if( g_wm->screen_changed == 0 ) return;
    g_wm->screen_changed = 0;
    
#ifdef WIN
    glFlush();
    SwapBuffers( g_wm->hdc );
#endif

#ifdef X11
#ifdef OPENGLES
    glFlush();
    eglSwapBuffers( egl_display, egl_surface );
#else
    XSync( g_wm->dpy, 0 );
    glFlush();
    glXSwapBuffers( g_wm->dpy, g_wm->win );
#endif
#endif

#ifdef IPHONE
    iphone_sundog_screen_redraw();
#endif

#ifdef OSX
    static bool gl_first_redraw = 1;
    if( gl_first_redraw )
    {
        glFlush();
        gl_first_redraw = 0;
    }
    osx_sundog_screen_redraw();
#endif

#ifdef ANDROID
    android_sundog_screen_redraw();
#endif
}	

#endif

//#################################
//#################################
//#################################

#endif
